//------------------------------------------------------------------------------------------------------
//
//  BlueMatikTest.c
//
//  For testing low-level ToothPIC routines for controlling it's piggy-backed BlueMatik module 
//  needs to conect to ToothPICTestWin windows console app 
//
//  (c) FlexiPanel Ltd
//
//-------------------------------------------------------------------------------------------------------
 
#include "..\ToothPIC.h"
#include <p18f6720.h>

#define BUFFLEN 	64 
Bool AwaitingDisconnect = False;
Bool AwaitingLinkQ = False;
Bool AwaitingSignal = False;
Bool AwaitingEnquiryEnd = False;
unsigned rom char pBogusBTAddr[] = { 0xCD, 0xCD, 0xCD, 0xCD, 0xCD, 0xCD };
unsigned rom char pEnquiryLen[] = "A"; // 10 seconds in (hex)

#define NAMEMAXLEN 21
#define MAXDEV 10
unsigned char NumDevices;
unsigned char pDescoveredDeviceName[MAXDEV][NAMEMAXLEN];

void main( void )
{
	unsigned char pBuff[BUFFLEN];
	unsigned char pTestComputerAddr[BTADDRLEN];
	unsigned char i;
	unsigned short isCount;
	unsigned short isInVal = 0xCDCD;

	// if no BlueMatik, flash rapidly
	while ((ToothPICSemaphores&TPSF_BMTEXISTS)==0)
	{
		LedRed = ~LedRed;
		msDelay(50);
	}

	// indicate init OK with green light
	LedGreen = LedGreenOn;

	// test incoming connection, no authentication
	BMTCommand( BMTC_Slave, 0, 0 );			
	AwaitBMTConnect( 0 );

	// Should succeed!
	if (!ConnectSuccess) 	Breakpoint (51);

	// send text "Slave" to tell ToothPICTestWin we've connected
	BMTTransmit( (unsigned rom char *) "Slave", 0, 5, 0 );		

	// await "Master" confirmation message from ToothPICTestWin
	BMTReceive( pBuff, 6, 0 );
	if (*pBuff!='M' || pBuff[1]!='a' || pBuff[2]!='s' || pBuff[3]!='t' || pBuff[4]!='e' || pBuff[5]!='r')  
	{
		// using breakpoint code which isn't confused with Error or BlueMatik Error / Warning
		Breakpoint (52);
	}

	// store bluetooth address of remote computer for later
	for (i=0; i<BTADDRLEN; i++) pTestComputerAddr[i] = pRemoteBTAddr[i];

	// send text "Disco" telling ToothPICTestWin we are about to disconnect
	BMTTransmit( (unsigned rom char *) "Disco", 0, 5, 0 );		

	// wait till buffer empty 
	while (BMTTxNChar) {;}

	// disconnect
	BMTCommand( BMTC_Disco, 0, 0 );			
	AwaitBMTOK;

	// toggle LED at end of test
	LedGreen = ~LedGreen;

	// set security
	BMTCommand( BMTC_Security, 0, Sec_Auth );			
	AwaitBMTOK;

	// connect as master
	BMTCommand( BMTC_Master , pTestComputerAddr, 0 );			
	AwaitBMTConnect( 250 );

	// Should succeed!
	if (!ConnectSuccess) 	Breakpoint (55);

	// send text AuthMaster to ToothPICTestWin
	BMTTransmit( (unsigned rom char *) "AuthMaster", 0, 10, 0 );		

	// await "Slave" response from ToothPICTestWin
	BMTReceive( pBuff, 5, 0 );
	if (*pBuff!='S' || pBuff[1]!='l' || pBuff[2]!='a' || pBuff[3]!='v' || pBuff[4]!='e' )  
	{
		// using breakpoint code which isn't confused with Error or BlueMatik Error / Warning
		Breakpoint (56);
	}

	// send text "Disco" and disconnect
	BMTTransmit( (unsigned rom char *) "Disco", 0, 5, 0 );		

	// wait till buffer empty 
	while (BMTTxNChar) {;}

	// disconnect
	BMTCommand( BMTC_Disco, 0, 0 );			
	AwaitBMTOK;

	// toggle LED at end of test
	LedGreen = ~LedGreen;

	// check that failure to connect to non-existent device causes no problems	
	BMTCommand( BMTC_Master , 0, pBogusBTAddr );		
	AwaitBMTConnect( 50 );

	// Shouldn't succeed!
	if (ConnectSuccess) 	
	{
		LedRed = ~LedRed;
		msDelay(50);
	}

	// toggle LED at end of test
	LedGreen = ~LedGreen;

	// device enquiry
	// send device enquiry command; no need for AwaitBMTOK because using AwaitingEnquiryEnd Semaphore
	AwaitingEnquiryEnd = True;
	NumDevices = 0;
	i = 10;  // duration in seconds
	BMTCommand( BMTC_Inquire, 0, pEnquiryLen );			

	// device found and end of enquiry handled in BMTEvent callback; wait for it to complete
	while (AwaitingEnquiryEnd) {;}
	
	// toggle LED at end of test
	LedGreen = ~LedGreen;

	// connect as secure master
	BMTCommand( BMTC_Master , pTestComputerAddr, 0 );			
	AwaitBMTConnect( 50 );

	// Should succeed!
	if (!ConnectSuccess) 	Breakpoint (57);

	// send text PairMaster
	BMTTransmit( (unsigned rom char *) "PairMaster", 0, 10, 0 );		

	// toggle LED at end of test
	LedGreen = ~LedGreen;

	// send number of devices then device names
	BMTTransmit( 0, &NumDevices, 1, 0 );		
	for (i=0; i<NumDevices; i++) BMTTransmit( 0, &pDescoveredDeviceName[i][0] , NAMEMAXLEN, 0);

	// bulk transfer out test
	// transmit 6250 words + 1 word to stop stopwatch
	for (isCount = 6251; isCount!=0; isCount--)  	
	{
		// transmits count value
		BMTTransmit( 0, (unsigned char *)&isCount, 2, 0 );	
	}
	
	// toggle LED at end of test
	LedGreen = ~LedGreen;

	// bulk transfer in test
	// receive 62500 words + 1 word to stop stopwatch
	for (isCount = 6251; isCount!=0; isCount--)  	
	{
		// receive count value
		BMTReceive( (unsigned char *)&isInVal, 2, 0 );	
		if (isInVal!=isCount)
		{
			Breakpoint (58);
		}
	}
	// send text 'OK' to stop stopwatch
	BMTTransmit( (unsigned rom char *) "OK", 0, 2, 0 );		
	
	// toggle LED at end of test
	LedGreen = ~LedGreen;

	// send signal strength command; no need for AwaitBMTOK because using AwaitingSignal Semaphore
	AwaitingSignal = True;
	BMTCommand( BMTC_Signal, 0, 0 );			

	// signal strength result handled in BMTEvent callback; wait for it to complete
	while (AwaitingSignal) {;}

	// toggle LED at end of test
	LedGreen = ~LedGreen;

	// send link quality command; no need for AwaitBMTOK because using AwaitingLinkQ Semaphore
	AwaitingLinkQ = True;
	BMTCommand( BMTC_LinkQ, 0, 0 );			

	// link quality result handled in BMTEvent callback; wait for it to complete
	while (AwaitingLinkQ) {;}

	// check the remote device can disconect withoug problems
	LedGreen = LedGreenOff;
	AwaitingDisconnect = True;
	while (AwaitingDisconnect) {;}
	
	// SUCCESS
	while ( 1 ) 
	{
		// Flash smugly to indicate success
		LedGreen = LedGreenOn;
		msDelay( 250 );
		msDelay( 250 );
		msDelay( 250 );
		msDelay( 250 );
		LedGreen = LedGreenOff;
		msDelay( 100 );
	}

}

// empty High Interrupt handler
void HighInterrupt (void)
{
}

// Low Interrupt handler
void LowInterrupt (void)
{
	if (IsSWI( SWI_BMTData ) )
	{
		// ignore interrupt - we're waiting for the data in a BMTReceive call anyway
		ClearSWI( SWI_BMTData );		// Clear interrupt flag
		return;
	}
	if (IsSWI( SWI_Tick ) )
	{
		ClearSWI( SWI_Tick );		// Clear interrupt flag
		return;
	}
}


// error handler - show status with breakpoint
void ErrorStatus( unsigned char ErrNum )
{
	Breakpoint( ErrNum );
}

// BMTEvent handler 
void BMTEvent( unsigned char EventID, void *pData1, void *pData2 )
{
	// show status with breakpoint
	if ( EventID==BMTE_Error || EventID==BMTE_Warning )
	{
		Breakpoint( ((unsigned char*)pData1)[0] );	
	}

	// Disconnect message from remote device disconnect test - raise semaphore
	if ( EventID==BMTE_Disco ) 
	{
		AwaitingDisconnect = False;
	}

	// Process Link quality response
	if ( EventID==BMTE_LinkQ )
	{
		// send result
		BMTTransmit( (unsigned rom char *) "LinkQ", 0, 5, 0 );		
		BMTTransmit( 0, pData1, 1, 0 );		
		AwaitingLinkQ = False;
	}

	// Process Signal strength response
	if ( EventID==BMTE_Signal )
	{
		// send result
		BMTTransmit( (unsigned rom char *) "Signal", 0, 6, 0 );		
		BMTTransmit( 0, pData1, 1, 0 );		
		AwaitingSignal = False;
	}

	// Process Device Found response
	if ( EventID==BMTE_Found )
	{	
		if (NumDevices < MAXDEV)
		{
			// store name
			unsigned char i;
			for (i=0; i<NAMEMAXLEN-1; i++) pDescoveredDeviceName[NumDevices][i] = ((unsigned char *)pData2)[i];
			pDescoveredDeviceName[NAMEMAXLEN-1][i] = 0;
			NumDevices++;
		}
	}

	// Process end of inquiry response
	if ( EventID==BMTE_OK )
	{
		if (AwaitingEnquiryEnd) AwaitingEnquiryEnd = 0;
	}
}



// empty FxPEvent handler
void FxPEvent( unsigned char EventID, void *pData )
{

}

