//------------------------------------------------------------------------------------------------------
//
//  DARC-I.c  
//
//  Application for driving a DARC module remotely via Bluetooth
//
//  (c) FlexiPanel Ltd
//
//-------------------------------------------------------------------------------------------------------

#include "..\ToothPIC.h"
#include <p18f6720.h>

#define SHORTMESSLEN	('0'-1) 					/* max short message length */
#define LONGMESSLEN		(SHORTMESSLEN*2+2) 					/* max long message length ascii hex followed by <CR><LF> */

// data in ToothPIC.c
extern rom unsigned short int DeviceClass; 
extern rom unsigned char pLocalName[];    
extern rom char pszPIN[];

// reserve RAM for user data
#pragma udata TP_UIRAM000			// located at 0x100 - 0x1FF
unsigned char pFxPRAM000[0x100];  	
#pragma udata TP_UIRAM100			// located at 0x200 - 0x1FF
unsigned char pFxPRAM100[0x100];
#pragma udata TP_UIRAM200			// located at 0x300 - 0x1FF
unsigned char pFxPRAM200[0x100];
#pragma udata TP_UIRAM300			// located at 0x400 - 0x8FF
unsigned char pFxPRAM300[0x100];
#pragma udata TP_UIRAM400			// located at 0x500 - 0x8FF
unsigned char pFxPRAM400[0x100];
#pragma udata TP_UIRAM500			// located at 0x600 - 0x8FF
unsigned char pFxPRAM500[0x100];
#pragma udata TP_UIRAM600			// located at 0x700 - 0x8FF
unsigned char pFxPRAM600[0x100];
#pragma udata TP_UIRAM700			// located at 0x800 - 0x8FF
unsigned char pFxPRAM700[0x100];
#pragma udata TP_UIRAM800			// located at 0x900 - 0x9FF
unsigned char pFxPRAM800[0x100];
#pragma udata

// reserve rom for user data - maximum available is 0x00F000 bytes
#pragma romdata Section_010000 = 0x010000
rom unsigned char p010000[0x4000]; 
#pragma romdata Section_014000 = 0x014000
rom unsigned char p014000[0x4000];
#pragma romdata Section_018000 = 0x018000
rom unsigned char p018000[0x4000];
#pragma romdata Section_01C000 = 0x01C000
rom unsigned char p01C000[0x3000];
#pragma romdata 


// Parallel, analog and CCP I/O functions
unsigned char NumAnalog;
unsigned char ParallelA;
unsigned char ParallelB;
unsigned char ParallelC;
unsigned char CCPFunc;
#define CCP1Func	0x01
#define CCP2Func	0x02
#define CCP3Func	0x04
#define CCP4Func	0x08
#define CCP5Func	0x01

// settings definitions
#define DarcS_None		0x00  /* default */
#define DarcS_Auth		0x01  
#define DarcS_AuthEncr	0x02  
#define DarcF_None		0x00  /* default */
#define DarcE_OnErrFlash	0x01  
#define DarcE_OnErrReset	0x02  
#define DarcE_OnErrResp	0x03  
#define DarcI_None		0x00
#define DarcI_I2C100kHz	0x01
#define DarcI_I2C400kHz	0x02

// permanent settings stored in Flash rom below 0x010000
#pragma romdata ROMDarcData=0x00FF80
rom unsigned char ROMDarc_Security = DarcS_None;			// Security level, see DarcS_ defines
rom unsigned char ROMDarc_InitMsg = True;					// Send initialization response
rom unsigned char ROMDarc_ASCIIMessages = True;				// reply messages are in ascii		
rom unsigned char ROMDarc_OnInternaError = DarcE_OnErrResp;	// what to do if an internal error occurs
rom unsigned char ROMDarc_I2CSetup = DarcI_None;			// I2C memory type
#pragma romdata 

// temporary copy of permanent settings 
unsigned char Darc_Security = DarcS_None;			
unsigned char Darc_FlowControl = DarcF_None;		
unsigned char Darc_InitMsg = True;				
unsigned char Darc_ASCIIMessages = True;				
unsigned char Darc_OnInternaError = DarcE_OnErrResp;
unsigned char Darc_I2CSetup = DarcI_None;				

// text strings
#define ResOKLen 1
rom unsigned char pResOK[] = { 0x01 };
#define ResInitLen 7
rom unsigned char pResInit[] = "DARC-I ";
extern rom unsigned char pVersionStr[];  // defined in ToothPIC.c
#define ResErrOverFlowLen 2
rom unsigned char pResErrOverFlowFull[] = { 0x02, 0xF4 };
rom unsigned char pResErrNotUnderstood[] = { 0x02, 0xF1 };
#define ResErrLen 2
rom unsigned char pResErrPingFail[] = { 0x02, 0xF2 };
#define ResErrInternalLen 1
rom unsigned char pResErrInternal[] = { 0x02 };
#define ResDateTimeLen 1
rom unsigned char pResDateTime[] = { 0x03 };
#define ResIOValLen 1
rom unsigned char pResIOVal[] = { 0x04 };
#define ResGotBytesLen 1
rom unsigned char pResGotBytes[] = { 0x05 };
#define ResGotSerialLen 1
rom unsigned char pResGotSerial[] = { 0x06 };

#define HTRxBuffLen 256
#define HTRxBuff pFxPRAM800

volatile unsigned char HTRxStart;			// Receive buffer startpoint
volatile unsigned char HTRxNext;		// Receive buffer endpoint
unsigned char PauseVal;
#define HTRxNChar 			( (unsigned char) (HTRxNext - HTRxStart ) )
#define HTRxSpace			( (unsigned char) ( 0xFF - HTRxNChar ) )
#define HTRxEmpty 			( HTRxNext == HTRxStart )
#define HTClear				(0)
#define HTStop				(1)
#define FC_MAX  220 // number of chars in buffer before flow control triggered

#define HTRxLoc( i )		( HTRxBuff + ( HTRxStart + ( (unsigned char) i ) ) )
#define HTRxCh				( HTRxBuff + HTRxStart )

// flow control pins
#define RTSPin	INT0Pin
#define DirRTS	DirINT0
#define CTSPin	INT1Pin
#define DirCTS	DirINT1

// function prototypes
void SendResponseBytes( rom unsigned char * pROMStr, unsigned char * pRAMStr, unsigned char nChar, unsigned char flags );
#define SRB_FIRST 	0x01
#define SRB_LAST	0x02
#define SRB_NOLENGTHPREFIX	0x04
unsigned char SetHex( unsigned char * pBuff, unsigned char ChASCII );
unsigned char ExtractBinaryVal( unsigned char * pASCII );
void ConfigDarcCmd( void );
void SendSerial( void );
void SerialPortSetup( unsigned char Speed );
void ConfigIOCmd( void );
void GetIOCmd( void );
void SetIOCmd( void );
void GetMemCmd( void );
void SetMemCmd( void );
void SendOKResponse(void);
void SendDateTimeResponse( void );
void InternalError( unsigned char ErrorID );
void SendIOValResponse( unsigned char iProperty, unsigned char iValue );
void SendADResponse( unsigned char iProperty );
void StreamFrameCmd( unsigned char bStream );
void SetLogInput( unsigned char Input, unsigned char OnOff );
unsigned char GetRecord( void );
void FindNextRecordLoc( unsigned char nBytes );
void GetADVal( unsigned char iProperty );

// Semaphores
unsigned char DarcSF;
#define DarcSF_CommandError 	0x01
#define DarcSF_InSlaveMode	 	0x02
#define DarcSF_SendingMessage 	0x04
#define DarcSF_Streaming	 	0x08  // as opposed to Framing
#define DarcSF_Framing	 		0x10  // as opposed to Streaming
#define DarcSF_ProcessingStrFrm	0x20  // processing a stream or frame
#define DarcSF_StreamError		0x40  // error during streaming
#define DarcSF_ResetMePlease	0x80  	// new for V3.0.00005 - reset BlueMatik on general error
#define InSlaveMode	( DarcSF & DarcSF_InSlaveMode )
#define IsCmdErr	( DarcSF & DarcSF_CommandError )
#define SetCmdErr	{ DarcSF |= DarcSF_CommandError; }
#define ClearCmdErr	{ DarcSF &= ~DarcSF_CommandError; }

// Stream, frame and log data
unsigned short StreamRate;
unsigned char LogRate;
unsigned char FLMemStart;
unsigned char FLMemEnd;
unsigned char FLMemType;
unsigned short Countdown;
unsigned short NumFrameLeft;
ADD FLMemPtr;  // where next record will be written
unsigned char LogPar;
#define LPar_A	0x01
#define LPar_B	0x02
#define LPar_C	0x04
unsigned char LogAN07;
#define LAN07_AN0 0x01
#define LAN07_AN1 0x02
#define LAN07_AN2 0x04
#define LAN07_AN3 0x08
#define LAN07_AN4 0x10
#define LAN07_AN5 0x20
#define LAN07_AN6 0x40
#define LAN07_AN7 0x80
unsigned char LogAN811;
#define LAN811_AN8 0x01
#define LAN811_AN9 0x02
#define LAN811_AN10 0x04
#define LAN811_AN11 0x08
unsigned char LogCCP;
#define LCCP_CCP1 0x01
#define LCCP_CCP2 0x02
#define LCCP_CCP3 0x04
#define LCCP_CCP4 0x08
#define LCCP_CCP5 0x10
unsigned char LogAUX;
#define LAUX_INT0 	0x01
#define LAUX_INT1	0x02
#define LAUX_SCL	0x04
#define LAUX_SDA	0x08
#define LAUX_SDO	0x10
#define LAUX_TxD	0x20
#define LAUX_RxD	0x40

// maximum record length = 39 bytes = 2 * 12 (analog) + 15 (digital)
unsigned char record[ 12 * 2 + 15 ];


// main program loop
void main( void )
{
	// if no BlueMatik, flash red led rapidly
	while ((ToothPICSemaphores&TPSF_BMTEXISTS)==0)
	{
		LedRed = ~LedRed;
		msDelay(50);
	}

	// initialize static variables
	DarcSF = 0;
	ParallelA = 0;
	ParallelB = 0;
	ParallelC = 0;
	CCPFunc = 0;
	NumAnalog = 0;
	StreamRate = 5000;  // 1 second 
	LogRate = 0x00;		// off
	FLMemStart = 0x02; 	// RAM
	FLMemEnd = 0x02;  	// RAM
	FLMemPtr = 0x00;	// start of RAM
	LogPar = 0; // none
	LogAN07 = 0; // none
	LogAN811 = 0; // none
	LogCCP = 0; // none
	LogAUX = 0; // none

	// read permanently stored values 
	GetBytes( STR_ROM00, (ADD)&ROMDarc_Security, &Darc_Security, 1 );
	GetBytes( STR_ROM00, (ADD)&ROMDarc_InitMsg, &Darc_InitMsg, 1 );
	GetBytes( STR_ROM00, (ADD)&ROMDarc_ASCIIMessages, &Darc_ASCIIMessages, 1 );
	GetBytes( STR_ROM00, (ADD)&ROMDarc_OnInternaError, &Darc_OnInternaError, 1 );
	GetBytes( STR_ROM00, (ADD)&ROMDarc_I2CSetup, &Darc_I2CSetup, 1 );

	// initialize I2C memory
	switch (Darc_I2CSetup)
	{
		case DarcI_None:   break;
		case DarcI_I2C100kHz:  I2CMemInit100kHz;   break;
		case DarcI_I2C400kHz:  I2CMemInit400kHz;   break;
	}	

	// initialize security
	switch (Darc_Security)
	{
		case DarcS_None:  BMTCommand( BMTC_Security, 0, Sec_None );   break;
		case DarcS_Auth:  BMTCommand( BMTC_Security, 0, Sec_Auth );   break;
		case DarcS_AuthEncr:  BMTCommand( BMTC_Security, 0, Sec_AuthEncr );   break;
	}	
	AwaitBMTOK;						    		

	// LED test completes initialization
	LedGreen = LedGreenOn;
	msDelay( 250 );
	LedGreen = LedGreenOff;
	LedRed = LedRedOn;
	msDelay( 250 );
	LedRed = LedRedOff;

	// main loop
 	while ( 1 ) 
	{
		// if not in slave mode, go into slave mode
		if (!InSlaveMode && !(DarcSF & DarcSF_ResetMePlease) ) 
		{
			BMTCommand( BMTC_Slave, 0, 0 );  
			AwaitBMTOK;		
			DarcSF |= DarcSF_InSlaveMode;    		
		}
		// new for V3.0.00005 - reset BlueMatik on general error
		if (DarcSF & DarcSF_ResetMePlease)
		{
			BMTCommand( BMTC_Reset, 0, 0 );  
			DarcSF &= ~DarcSF_ResetMePlease;

			BMTCommand( BMTC_Slave, 0, 0 );  
			AwaitBMTOK;		
			DarcSF |= DarcSF_InSlaveMode;    		
		}
		if (HTRxNext!=HTRxStart)
		{
			// got data - transmit it
			unsigned char TmpNext; 

			// transmit if data not still streaming in and buffer not half full
			// the buffer not half full bit ensures don't overrun when no flow control is used
			TmpNext = HTRxNext;
			msDelay( PauseVal );
			if (TmpNext == HTRxNext || HTRxNChar>127)
			{
				DisableLowInts;
				TmpNext = HTRxNext;
	
				SendResponseBytes( pResGotSerial, 0, ResGotSerialLen, SRB_FIRST );
	
				while (HTRxStart != TmpNext)
				{
					unsigned char NextHTRxStartVal = HTRxStart + 1;
					SendResponseBytes( 0, HTRxBuff + HTRxStart, 1, (NextHTRxStartVal == TmpNext) ? SRB_LAST : 0 );
					HTRxStart = NextHTRxStartVal;
				}
				EnableLowInts;
				RTSPin = HTClear;
			}
		}

	}  // end of while ( 1 ) 

}

// HighInterrupt handler - nothing needs to be done
void HighInterrupt (void)
{
	if (UARTRxIntFlag)
	{
		// note PIR3bits.TX2IF cleared by hardware
		if (UARTOverrunErr || UARTFrameErr) 
		{
			unsigned char ch;
			ch = UARTChIn;  // clears UARTRxIntFlag

			// turn on red led to indicate an error
			UARTEnable = 0;
			UARTEnable = 1;
			return;
		}

		// store character and advance buffer
		*( HTRxBuff + HTRxNext ) = UARTChIn;
		HTRxNext++;

		// If flow control, tell RTS to stop if getting near full
		if ( BMTRxNChar > FC_MAX )  	RTSPin = HTStop;

		return;
	}
}

// LowInterrupt handler
void LowInterrupt (void)
{
	// timer for framing and streaming
	if (PIE3bits.TMR4IE && PIR3bits.TMR4IF)
	{
		unsigned char nBytes;
		
		// timer interrupt
		PIR3bits.TMR4IF = 0;  // allow interrupt to trigger again

		// decrement Countdown counter
		Countdown --;
		if (Countdown) return;

		// reset clock
		Countdown = StreamRate;

		if ( DarcSF & DarcSF_ProcessingStrFrm )
		{
			// haven't finished processing the last frame
			PIE3bits.TMR4IE = 0;
			T0CON = 0; // turn timer off
			// generate error
			DarcSF |= DarcSF_StreamError;
			if ( DarcSF & DarcSF_Framing )
			{
				SendResponseBytes( pResErrOverFlowFull, 0, ResErrLen, SRB_FIRST | SRB_LAST );
				return;
			}
		}
		// raise flag to indicate must log
		DarcSF |= DarcSF_ProcessingStrFrm;

		// take log
		nBytes = GetRecord( );

		if (DarcSF & DarcSF_Framing)
		{
			// log data
			SetBytes( FLMemType, FLMemPtr, 0, (unsigned char *) record, nBytes );

			// locate next record location
			FindNextRecordLoc( nBytes );
			
			// decrement NumFrameLeft and stop if have enough data
			NumFrameLeft --;
			if (!NumFrameLeft)
			{
				PIE3bits.TMR4IE = 0;
				T0CON = 0; // turn timer off
				SendResponseBytes( pResOK, 0, ResOKLen, SRB_FIRST | SRB_LAST );
				return;
			}
		}
		else
		{
			// transmit result if streaming hasn't been cancelled
			DisableLowInts;
			DisableHighInts;
			if ( DarcSF & DarcSF_Streaming ) 	SendResponseBytes( 0, record, nBytes, SRB_FIRST | SRB_LAST | SRB_NOLENGTHPREFIX );
			EnableHighInts;
			EnableLowInts;
		}

		// clear interrupt processing flag
		DarcSF &= ~DarcSF_ProcessingStrFrm;
		return;
	}

	if (IsSWI( SWI_Tick ) )
	{
		// decide whether it is time for a log event or not 
		unsigned char DoLog = False;
		if (LogRate == 0x01)	DoLog = True;
		if (LogRate == 0x02 && (RealTimeClock.sec & 0x01) )  	DoLog = True;
		if (LogRate == 0x03 && (RealTimeClock.sec % 5) )  		DoLog = True;
		if (LogRate == 0x04 && (RealTimeClock.sec % 10) )  	DoLog = True;
		if (LogRate == 0x05 && (RealTimeClock.sec % 15) )  	DoLog = True;
		if (LogRate == 0x06 && (RealTimeClock.sec % 30) )  	DoLog = True;
		if (!RealTimeClock.sec)
		{
			if (LogRate == 0x07)	DoLog = True;
			if (LogRate == 0x08 && (RealTimeClock.min & 0x01) )  	DoLog = True;
			if (LogRate == 0x09 && (RealTimeClock.min % 5) )  		DoLog = True;
			if (LogRate == 0x0A && (RealTimeClock.min % 10) )  	DoLog = True;
			if (LogRate == 0x0B && (RealTimeClock.min % 15) )  	DoLog = True;
			if (LogRate == 0x0C && (RealTimeClock.min % 30) )  	DoLog = True;
			if (!RealTimeClock.min)
			{
				if (LogRate == 0x0D)	DoLog = True;
				if (LogRate == 0x0E && (RealTimeClock.hour & 0x01) )  	DoLog = True;
				if (LogRate == 0x0F && (RealTimeClock.hour % 3) )  		DoLog = True;
				if (LogRate == 0x10 && (RealTimeClock.hour % 6) )  	DoLog = True;
				if (LogRate >= 0x40  && LogRate <0x58 && RealTimeClock.hour == LogRate - 0x40)  	DoLog = True;
			}
		}

		if (DoLog)
		{
			unsigned char nBytes;

			// record time
			SetBytes( FLMemType, FLMemPtr, 0, (unsigned char *) &RealTimeClock, sizeof(RealTimeClock) );

			// log data
			nBytes = GetRecord( );
			SetBytes( FLMemType, FLMemPtr+sizeof(RealTimeClock), 0, (unsigned char *) record, nBytes );

			// locate next record location
			FindNextRecordLoc( nBytes + sizeof(RealTimeClock) );
		}

		ClearSWI( SWI_Tick );		// Clear clock interrupt flag
		return;
	}

	if (IsSWI(SWI_BMTData))  // remote host data received
	{
		unsigned char CommandLen;
		unsigned char bIsASCII;

		// if streaming, end stream
		if (DarcSF & DarcSF_Streaming)
		{
			DarcSF &= ~DarcSF_Streaming;
			PIE2bits.TMR3IE = 0;
			T3CON = 0; // turn timer off

			// send error or OK as appropriate
			if (DarcSF & DarcSF_StreamError)
				SendResponseBytes( pResErrOverFlowFull, 0, ResErrLen, SRB_FIRST | SRB_LAST );
			else
				SendResponseBytes( pResOK, 0, ResOKLen, SRB_FIRST | SRB_LAST );

			// reset input buffer
			BMTRxEnd = 0;
			BMTRxStart = 0;
			ClearSWI( SWI_BMTData );

			return;
		}


		// decide if ASCII
		bIsASCII = False;
		if ( (BMTRxBuff[0] >='0' && BMTRxBuff[0] <='9') || (BMTRxBuff[0] >='a' && BMTRxBuff[0] <='f') || (BMTRxBuff[0] >='A' && BMTRxBuff[0] <='F') ) bIsASCII = True;

		// if looks like ASCII command and this is the first byte of command length byte, do no further processing
		if (BMTRxEnd==1 && bIsASCII ) return;

		// determine command length
		if (bIsASCII) // calculate length for text commands
		{
			CommandLen = ExtractBinaryVal( (unsigned char *) BMTRxBuff );
			CommandLen = CommandLen * 2 + 2;  // correct number of characetrs including <CR><LF>
			if ( CommandLen > LONGMESSLEN ) SetCmdErr;
		}
		else
		{
			CommandLen = BMTRxBuff[0];  // length for binary commands
			if ( CommandLen > SHORTMESSLEN ) SetCmdErr;
		}

		if ( BMTRxEnd >= CommandLen || IsCmdErr) // complete command has been read, or error occured
		{
			SetSWI( SWI_SWI1 ); // tell low priority interrrupt to process command
		}

		ClearSWI( SWI_BMTData );
	}

	if (IsSWI( SWI_SWI1 ) )			// command to process
	{
		// translate to binary if required
		if (!IsCmdErr && BMTRxEnd>1 && BMTRxBuff[0] >='0') 
		{
			unsigned char iCh;

			// convert length byte to binary
			BMTRxBuff[0] = ExtractBinaryVal( (unsigned char *) BMTRxBuff );
			
			// if length byte is zero, reset
			if (!IsCmdErr && !BMTRxBuff[0]) Reset();

			// generate error if final characters are not <CR><LF> or if too long
			if (!IsCmdErr && (BMTRxBuff[BMTRxEnd-2] != '\r' || BMTRxBuff[BMTRxEnd-1] != '\n' || BMTRxBuff[0] > LONGMESSLEN )) SetCmdErr;
	
			// convert remaining bytes to binary
			for (iCh = 1; !IsCmdErr && iCh < BMTRxBuff[0]; iCh++)
			{
				BMTRxBuff[iCh] = ExtractBinaryVal( (unsigned char *) BMTRxBuff+(iCh<<1) );
			}
			BMTRxEnd = BMTRxBuff[0];
		}

		// process command
		if (!IsCmdErr)
		{ 
			switch ( BMTRxBuff[1] )
			{
				case 0x00: // disconnect and reset
					BMTCommand( BMTC_Disco, 0, 0 );  
					msDelay( 250 );
					msDelay( 250 );  // wait a second for BlueMatik to disconnect
					msDelay( 250 );
					msDelay( 250 );
					Reset();

				case 0x01:	// Config Peripheral Mode command
					ConfigDarcCmd();
					break;

				case 0x02:	// Config IO command
					ConfigIOCmd();
					break;

				case 0x03:	// Set IO command
					SetIOCmd();
					break;

				case 0x04:	// Get IO command
					GetIOCmd();
					break;

				case 0x05:  // get memory
					GetMemCmd();
					break;

				case 0x06:  // set memory
					SetMemCmd();
					break;

				case 0x07:  // stream
					StreamFrameCmd( True );
					break;

				case 0x08:  // frame
					StreamFrameCmd( False );
					break;

				case 0x09:  // frame
					SendSerial( );
					break;

				default:  // not understood
					SetCmdErr;
					break;
			}
		}

		if (IsCmdErr)
		{
			// didn't understand command - error
			SendResponseBytes( pResErrNotUnderstood, 0, ResErrLen, SRB_FIRST | SRB_LAST );
			ClearCmdErr;
		}

		// reset incoming command buffer
		BMTRxStart = 0;
		BMTRxEnd = 0;

		ClearSWI( SWI_SWI1 );		// Clear software interrupt
		return;
	}


}


// Error event handler 
void ErrorStatus( unsigned char ErrNum )
{
	InternalError( ErrNum );
}

// BlueMatik Event handler 
void BMTEvent( unsigned char EventID, void *pData1, void *pData2 )
{
	switch (EventID)
	{
		case BMTE_Error:
		case BMTE_Warning:
			if (((unsigned char*)pData1)[0] == 0x0F) // general error
			{
				DarcSF |= DarcSF_ResetMePlease;
				return;
			}

			InternalError( ((unsigned char*)pData1)[0] );
			return;
		case BMTE_Connect:
			// transmit greetings message
			if ( Darc_InitMsg )  
			{
				SendResponseBytes( pResInit, 0, ResInitLen, SRB_FIRST );
				SendResponseBytes( pVersionStr, 0, VERSTRLEN-1, SRB_LAST );
			}
			// reset BMT receive buffer
			BMTRxStart = 0;
			BMTRxEnd = 0;
			return;
		case BMTE_Disco:
			// set semaphore to re-enter slave mode
			DarcSF &= ~DarcSF_InSlaveMode;    	
			return;	
	}
	
}

// FxPEvent handler 
void FxPEvent( unsigned char EventID, void *pData )
{

}

void InternalError( unsigned char ErrorID )
{
	switch (Darc_OnInternaError)
	{
		case DarcE_OnErrFlash:
			Breakpoint( ErrorID );
			Reset();

		case DarcE_OnErrReset:
			Reset();

		case DarcE_OnErrResp:
			SendResponseBytes( pResErrInternal, 0, ResErrInternalLen, SRB_FIRST | SRB_LAST );
	}
}

void SendResponseBytes( rom unsigned char * pROMStr, unsigned char * pRAMStr, unsigned char nChar, unsigned char flags )
{
	static unsigned char MyByteLenCount = 0;  // including byte length counter
	static unsigned char MyTxBuffPtr = 0;
	unsigned char i;

	if (!IsConnected) return;

	// error if curently sending a message and this is the start of a new one
	if ( (flags & SRB_FIRST) && (DarcSF & DarcSF_SendingMessage))  Breakpoint( 222 );

	// raise sending message semaphore
	DarcSF |= DarcSF_SendingMessage;

	// if first part of message, reset TxBuffer leaving space for length byte if necessary
	if ( flags & SRB_FIRST )
	{
		if ( flags & SRB_NOLENGTHPREFIX )
		{ 
			MyByteLenCount = 0;  
			BMTTxStart = 0;
			BMTTxEnd = 0;
			if (Darc_ASCIIMessages)
				MyTxBuffPtr = 2;
			else
				MyTxBuffPtr = 0;
		}
		else
		{
			MyByteLenCount = 1;  // including byte length counter
			BMTTxStart = 0;
			BMTTxEnd = 0;
			if (Darc_ASCIIMessages)
				MyTxBuffPtr = 4;
			else
				MyTxBuffPtr = 1;
		}
	}

	MyByteLenCount += nChar;

	if (Darc_ASCIIMessages)
	{
		// if required, add <CR><LF> 
		if (flags&SRB_FIRST)  
		{
			BMTTxBuff[0] = 0x0D;
			BMTTxBuff[1] = 0x0A;
		}

		// if ascii messages, convert to ascii and copy to tx buffer
		for (i=0; i< nChar; i++)
		{
			SetHex( (unsigned char *) BMTTxBuff+MyTxBuffPtr, ((pROMStr) ? pROMStr[i] : pRAMStr[i]) );
			MyTxBuffPtr += 2;
		}
		// if required, add <CR><LF> and update length counter
		if ( flags & SRB_LAST )  
		{
			BMTTxBuff[MyTxBuffPtr] = 0x0D;
			MyTxBuffPtr ++;			
			BMTTxBuff[MyTxBuffPtr] = 0x0A;
			MyTxBuffPtr ++;		
			if ( ( flags & SRB_NOLENGTHPREFIX )==0 )
			{
				SetHex( (unsigned char *) &BMTTxBuff[2], MyByteLenCount );
			}
		}
	}
	else
	{
		// otherwise copy directly
		// if ascii messages, convert to ascii and copy to tx buffer
		for (i=0; i< nChar; i++)
		{
			BMTTxBuff[MyTxBuffPtr] = ((pROMStr) ? pROMStr[i] : pRAMStr[i] );
			MyTxBuffPtr ++;
		}
		// update length counter
		if ( (flags & SRB_LAST) && ( flags & SRB_NOLENGTHPREFIX )==0 )  BMTTxBuff[0] = MyByteLenCount;
	}

	// if that completes the message, transmit it
	if ( flags & SRB_LAST )  
	{
		// for version 3.0.00004, TxBuff is yet implemented, so will call BMTTransmit.
		// when TxBuff is implemented, correct call would be TxAdvance
		// however, leaving the current call would still work
		BMTTransmit( 0, (unsigned char *)BMTTxBuff, MyTxBuffPtr, 0 );
	
		// lower DarcSF_SendingMessage since that was the end of the message
		DarcSF &= ~DarcSF_SendingMessage;
	}
}

// convert hexadecimal ascii to binary
unsigned char ExtractBinaryVal( unsigned char * pASCII )
{
	unsigned char ch, RetVal;

	// convert to upper case
	if (pASCII[0]>='a' && pASCII[0]<='f') pASCII[0]-=0x20;
	if (pASCII[1]>='a' && pASCII[1]<='f') pASCII[1]-=0x20;

	// check valid
	if (pASCII[0] < '0' || (pASCII[0]>'9' && pASCII[0]<'A') || pASCII[0]>'F' || 
		pASCII[1] < '0' || (pASCII[1]>'9' && pASCII[1]<'A') || pASCII[1]>'F' ) 
	{
		SetCmdErr;
		return 0xFF;
	}

	// convert to binary
	ch = pASCII[0] - '0';
	if (ch>9) ch -=7;
	RetVal = ch << 4;
	ch = pASCII[1] - '0';
	if (ch>9) ch -=7;
	RetVal += ch;

	return RetVal;
}

void SendOKResponse( void )
{
	// send OK message
	SendResponseBytes( pResOK, 0, ResOKLen, SRB_FIRST | SRB_LAST );
}

void SendDateTimeResponse( void )
{
	// send OK message
	SendResponseBytes( pResDateTime, 0, ResDateTimeLen, SRB_FIRST );
	SendResponseBytes( 0, (unsigned char *)&RealTimeClock, sizeof(RealTimeClock), SRB_LAST );
}

void ConfigDarcCmd( void )
{
	// error if no 3rd byte
	if (BMTRxEnd<3) { SetCmdErr;  return; }

	// read 3rd byte to identify the property being set
	switch (BMTRxBuff[2])
	{
		case 0x01:  // security
			if (BMTRxEnd<4 || BMTRxBuff[3] > 0x02) { SetCmdErr;  return; }
			SetBytes( STR_ROM00, (ADD)&ROMDarc_Security, 0, (unsigned char *) &BMTRxBuff[3], 1 );
			break;
		case 0x02:  // PIN
			if (BMTRxEnd<4 || BMTRxBuff[3] == 0x00 ) { SetCmdErr;  return; }
			// ensure some kind of zero terminator exists
			BMTRxBuff[19] = 0x00;
			SetBytes( STR_ROM00, (ADD)pszPIN, 0, (unsigned char *) &BMTRxBuff[3], 17 );
			break;
		case 0x03:  // Name
			if (BMTRxEnd<4 || BMTRxBuff[3] == 0x00 ) { SetCmdErr;  return; }
			// ensure some kind of zero terminator exists
			BMTRxBuff[19] = 0x00;
			SetBytes( STR_ROM00, (ADD)pLocalName, 0, (unsigned char *) &BMTRxBuff[3], 17 );
			break;
		case 0x04:  // initialization response
			if (BMTRxEnd<4) { SetCmdErr;  return; }  // any nonzero value is assumed to mean 'true'
			SetBytes( STR_ROM00, (ADD)&ROMDarc_InitMsg, 0, (unsigned char *) &BMTRxBuff[3], 1 );
			break;
		case 0x05:  // ASCII responses
			if (BMTRxEnd<4) { SetCmdErr;  return; }  // any nonzero value is assumed to mean 'true'
			SetBytes( STR_ROM00, (ADD)&ROMDarc_ASCIIMessages, 0, (unsigned char *) &BMTRxBuff[3], 1 );
			break;
		case 0x06:  // On internal error
			if (BMTRxEnd<4  || BMTRxBuff[3] == 0x00 || BMTRxBuff[3] > 0x03) { SetCmdErr;  return; }
			SetBytes( STR_ROM00, (ADD)&ROMDarc_OnInternaError, 0, (unsigned char *) &BMTRxBuff[3], 1 );
			break;
		case 0x07:  // I2C Setup
			if (BMTRxEnd<4 || BMTRxBuff[3] > 0x02) { SetCmdErr;  return; }
			SetBytes( STR_ROM00, (ADD)&ROMDarc_I2CSetup, 0, (unsigned char *) &BMTRxBuff[3], 1 );
			break;
		case 0x08:  // Device class
			if (BMTRxEnd<6) { SetCmdErr;  return; }
			// swap first and third byte
			BMTRxBuff[0] = BMTRxBuff[3];
			BMTRxBuff[3] = BMTRxBuff[5];
			BMTRxBuff[5] = BMTRxBuff[0];
			SetBytes( STR_ROM00, (ADD)&DeviceClass, 0, (unsigned char *) &BMTRxBuff[3], 3 );
			break;
		case 0x09:  // Serial port setup
			if (BMTRxEnd<4 || BMTRxBuff[3] > 0x06) { SetCmdErr;  return; }
			SerialPortSetup( BMTRxBuff[3] );
			break;
		case 0x21:  // Stream / Frame rate
			if (BMTRxEnd!=5) { SetCmdErr;  return; }
			StreamRate = BMTRxBuff[3] + 0x100 * BMTRxBuff[4];
			break;
		case 0x22:  // Log rate
			if (BMTRxEnd!=4) { SetCmdErr;  return; }
			FLMemType = FLMemStart;
			FLMemPtr = 0x0000;
			LogRate = BMTRxBuff[3];
			break;
		case 0x23:  // Stream / frame / log inputs
			if (BMTRxEnd!=5) { SetCmdErr;  return; }
			SetLogInput( BMTRxBuff[3], BMTRxBuff[4] );
			break;
		case 0x24:  // FLMemStart, FLMemEnd
			if (BMTRxEnd!=5) { SetCmdErr;  return; }
			FLMemStart = BMTRxBuff[3];
			FLMemEnd = BMTRxBuff[4];
			break;
		case 0x41:  // daylight savings state
			if (BMTRxEnd<4) { SetCmdErr;  return; }
			DSTEvent = BMTRxBuff[3];
			break;
		case 0x42:  // set time
			if (BMTRxEnd<10) { SetCmdErr;  return; }
			RealTimeClock.sec = BMTRxBuff[3];
			RealTimeClock.min = BMTRxBuff[4];
			RealTimeClock.hour = BMTRxBuff[5];
			RealTimeClock.date = BMTRxBuff[6];
			RealTimeClock.month = BMTRxBuff[8];
			RealTimeClock.year = *(( unsigned short *) (&BMTRxBuff[9]) );
			CalcDayOfWeek( RealTimeClock.dow, RealTimeClock.year, RealTimeClock.month, RealTimeClock.date ) 
			break;
		case 0x43:  // request data / time
			SendDateTimeResponse();
			return;
		default:
			SetCmdErr;  
			return; 
	}

	// generate OK Response
	SendOKResponse();

}

void SendSerial( void )
{
	unsigned char ByteLoc = 2;

	// output data
	while (ByteLoc < BMTRxEnd)
	{
			while (UARTNotTxmitting==0) {;};
			while (CTSPin == HTStop) {;};
			UARTChOut = BMTRxBuff[ByteLoc];
			ByteLoc++;
	}

	// generate OK Response
	SendOKResponse();

}

void SerialPortSetup( unsigned char Speed )
{
	unsigned char BRGVal, BRGHighVal ;
	
	switch( Speed )
	{
	case 0:	
	    RCSTA2bits.SPEN = 0; 
		PauseVal = 1;
		return;	
	case 1:
		BRGVal = SPBaudBrghZero2400;
		BRGHighVal = 0;
		PauseVal = 32;
		break;
	case 2:
		BRGVal = SPBaudBrghZero4800;
		BRGHighVal = 0;
		PauseVal = 16;
		break;
	case 3:
		BRGVal = SPBaudBrghOne9600;
		BRGHighVal = 1;
		PauseVal = 8;
		break;
	case 4:
		BRGVal = SPBaudBrghOne19200;
		BRGHighVal = 1;
		PauseVal = 4;
		break;
	case 5:
		BRGVal = SPBaudBrghOne38400;
		BRGHighVal = 1;
		PauseVal = 2;
		break;
	case 6:
		BRGVal = SPBaudBrghOne57600;
		BRGHighVal = 1;
		PauseVal = 1;
		break;
	}

	InitUART( BRGVal, BRGHighVal );
	UARTEnable = 1;
	DirRTS = 0;  // rts clear output
	DirCTS = 1;  // cts input
}


void SetLogInput( unsigned char Input, unsigned char OnOff )
{
	if (OnOff)
	{
		switch (Input)
		{
			case 0x06:  LogPar |= LPar_A;  return;
			case 0x07:  LogPar |= LPar_B;  return;
			case 0x08:  LogPar |= LPar_C;  return;
			case 0x10:  LogAN07 |= LAN07_AN0;  return;
			case 0x11:  LogAN07 |= LAN07_AN1;  return;
			case 0x12:  LogAN07 |= LAN07_AN2;  return;
			case 0x13:  LogAN07 |= LAN07_AN3;  return;
			case 0x14:  LogAN07 |= LAN07_AN4;  return;
			case 0x15:  LogAN07 |= LAN07_AN5;  return;
			case 0x16:  LogAN07 |= LAN07_AN6;  return;
			case 0x17:  LogAN07 |= LAN07_AN7;  return;
			case 0x18:  LogAN811 |= LAN811_AN8;  return;
			case 0x19:  LogAN811 |= LAN811_AN9;  return;
			case 0x1A:  LogAN811 |= LAN811_AN10;  return;
			case 0x1B:  LogAN811 |= LAN811_AN11;  return;
			case 0x1C:  LogCCP |= LCCP_CCP1;  return;
			case 0x1D:  LogCCP |= LCCP_CCP2;  return;
			case 0x1E:  LogCCP |= LCCP_CCP3;  return;
			case 0x1F:  LogCCP |= LCCP_CCP4;  return;
			case 0x20:  LogCCP |= LCCP_CCP5;  return;
			case 0x21:  LogAUX |= LAUX_INT0;  return;
			case 0x22:  LogAUX |= LAUX_INT1;  return;
			case 0x23:  LogAUX |= LAUX_SCL;  return;
			case 0x24:  LogAUX |= LAUX_SDA;  return;
			case 0x25:  LogAUX |= LAUX_SDO;  return;
			case 0x26:  LogAUX |= LAUX_TxD;  return;
			case 0x27:  LogAUX |= LAUX_RxD;  return;
			default:  SetCmdErr;  return;
		}
	}
	else
	{
		switch (Input)
		{
			case 0x06:  LogPar &= ~LPar_A;  return;
			case 0x07:  LogPar &= ~LPar_B;  return;
			case 0x08:  LogPar &= ~LPar_C;  return;
			case 0x10:  LogAN07 &= ~LAN07_AN0;  return;
			case 0x11:  LogAN07 &= ~LAN07_AN1;  return;
			case 0x12:  LogAN07 &= ~LAN07_AN2;  return;
			case 0x13:  LogAN07 &= ~LAN07_AN3;  return;
			case 0x14:  LogAN07 &= ~LAN07_AN4;  return;
			case 0x15:  LogAN07 &= ~LAN07_AN5;  return;
			case 0x16:  LogAN07 &= ~LAN07_AN6;  return;
			case 0x17:  LogAN07 &= ~LAN07_AN7;  return;
			case 0x18:  LogAN811 &= ~LAN811_AN8;  return;
			case 0x19:  LogAN811 &= ~LAN811_AN9;  return;
			case 0x1A:  LogAN811 &= ~LAN811_AN10;  return;
			case 0x1B:  LogAN811 &= ~LAN811_AN11;  return;
			case 0x1C:  LogCCP &= ~LCCP_CCP1;  return;
			case 0x1D:  LogCCP &= ~LCCP_CCP2;  return;
			case 0x1E:  LogCCP &= ~LCCP_CCP3;  return;
			case 0x1F:  LogCCP &= ~LCCP_CCP4;  return;
			case 0x20:  LogCCP &= ~LCCP_CCP5;  return;
			case 0x21:  LogAUX &= ~LAUX_INT0;  return;
			case 0x22:  LogAUX &= ~LAUX_INT1;  return;
			case 0x23:  LogAUX &= ~LAUX_SCL;  return;
			case 0x24:  LogAUX &= ~LAUX_SDA;  return;
			case 0x25:  LogAUX &= ~LAUX_SDO;  return;
			case 0x26:  LogAUX &= ~LAUX_TxD;  return;
			case 0x27:  LogAUX &= ~LAUX_RxD;  return;
			default:  SetCmdErr;  return;
		}
	}
}

void ConfigIOCmd( void )
{
	// error if no 3rd byte
	if (BMTRxEnd<3) { SetCmdErr;  return; }

	// read 3rd byte to identify the property being set
	switch (BMTRxBuff[2])
	{
		case 0x01:  // a to d channels
			NumAnalog = BMTRxBuff[3];
			if (BMTRxEnd<4 || NumAnalog > 0x0C) { SetCmdErr;  return; }
			// set number of channels
			ADCON1 = (ADCON1 & 0xF0) | (0x0F-NumAnalog);
			// tun a to d on or off according to number of channels
			if (NumAnalog)
				ADConverterOn10bit
			else
				ADConverterOff
			break;
		case 0x02:  // -ve volt ref
			if (BMTRxEnd<4) { SetCmdErr;  return; }  // any nonzero value is assumed to mean 'true'
			if (BMTRxBuff[3])
				VRefNegIsAN2
			else
				VRefNegIsVss
			break;
		case 0x03:  // -ve volt ref
			if (BMTRxEnd<4) { SetCmdErr;  return; }  // any nonzero value is assumed to mean 'true'
			if (BMTRxBuff[3])
				VRefPosIsAN3
			else
				VRefNegIsVss
			break;
		case 0x04:  // PWM time base units
			if (BMTRxEnd<4 || BMTRxBuff[3] == 0 || BMTRxBuff[3] > 0x03) { SetCmdErr;  return; }
			if (!BMTRxBuff[3])
			{  // PWM off
				T4CON = 0b00000000;  // timer 4 off
			}
			else
			{  // PWM on, set time base
				T3CONbits.T3CCP1 = 1;  // set PWM to timer 4
				T3CONbits.T3CCP2 = 1;  // set PWM to timer 4
				T4CON = 0b00000100;  // 0.2us timebase
				if (BMTRxBuff[3]==2) T4CON |= 0b00000001;  // 0.8us timebase
				if (BMTRxBuff[3]==3) T4CON |= 0b00000010;  // 3.2us timebase	
			}		
			break;
		case 0x05:  // PWM period
			if (BMTRxEnd<4) { SetCmdErr;  return; }
			PR4 = BMTRxBuff[3];  // timer period = (PR4 + 1) x (PWM time base units)
			break;
		case 0x06:  // Parallel I/O A
			if (BMTRxEnd<4 || BMTRxBuff[3] == 0x01 || (BMTRxBuff[3]>0x07 && BMTRxBuff[3]<0x12) || BMTRxBuff[3] > 0x17 ) { SetCmdErr;  return; }		
			ParallelA = BMTRxBuff[3]; 
			switch (ParallelA)
			{
				case 0x00:  break;

				case 0x07:  DirAN5 = DirOutput; // note how case statements fall through
				case 0x06:	DirAN6 = DirOutput;
				case 0x05:	DirAN7 = DirOutput;
				case 0x04:	DirAN8 = DirOutput;
				case 0x03:	DirAN9 = DirOutput;
				case 0x02:	DirAN10 = DirOutput;
							DirAN11 = DirOutput;
							break;

				case 0x17:  DirAN5 = DirInput; // note how case statements fall through
				case 0x16:	DirAN6 = DirInput;
				case 0x15:	DirAN7 = DirInput;
				case 0x14:	DirAN8 = DirInput;
				case 0x13:	DirAN9 = DirInput;
				case 0x12:	DirAN10 = DirInput;
							DirAN11 = DirInput;
							break;
			}
			break;
		case 0x07:  // Parallel I/O B
			if (BMTRxEnd<4 || BMTRxBuff[3] == 0x01 || (BMTRxBuff[3]>0x05 && BMTRxBuff[3]<0x12) || BMTRxBuff[3] > 0x15 ) { SetCmdErr;  return; }		
			ParallelB = BMTRxBuff[3]; 
			switch (ParallelB)
			{
				case 0x00:  break;

				case 0x05:	DirAN0 = DirOutput;
				case 0x04:	DirAN1 = DirOutput;
				case 0x03:	DirAN2 = DirOutput;
				case 0x02:	DirAN3 = DirOutput;
							DirAN4 = DirOutput;
							break;

				case 0x15:	DirAN0 = DirInput;
				case 0x14:	DirAN1 = DirInput;
				case 0x13:	DirAN2 = DirInput;
				case 0x12:	DirAN3 = DirInput;
							DirAN4 = DirInput;
							break;
			}
			break;
		case 0x08:  // Parallel I/O C
			if (BMTRxEnd<4 || BMTRxBuff[3] == 0x01 || (BMTRxBuff[3]>0x05 && BMTRxBuff[3]<0x12) || BMTRxBuff[3] > 0x15 ) { SetCmdErr;  return; }		
			ParallelC = BMTRxBuff[3]; 
			switch (ParallelC)
			{
				case 0x00:  break;

				case 0x05:	DirCCP3 = DirOutput;
				case 0x04:	DirTxD = DirOutput;
				case 0x03:	DirRxD = DirOutput;
				case 0x02:	DirCCP4 = DirOutput;
							DirCCP5 = DirOutput;
							break;

				case 0x15:	DirCCP3 = DirInput;
				case 0x14:	DirTxD = DirInput;
				case 0x13:	DirRxD = DirInput;
				case 0x12:	DirCCP4 = DirInput;
							DirCCP5 = DirInput;
							break;
			}
			break;
		case 0x10:  // AN0 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN0 = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x11:  // AN1 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN1 = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x12:  // AN2 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN2 = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x13:  // AN3 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN3 = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x14:  // AN4 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN4 = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x15:  // AN5 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN5 = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x16:  // AN6 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN6 = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x17:  // AN7 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN7 = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x18:  // AN8 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN8 = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x19:  // AN9 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN9 = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x1A:  // AN10 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN10 = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x1B:  // AN11 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN11 = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x1C:  // CCP1 digital pin
			if (BMTRxEnd<4 || BMTRxBuff[3]>0x02 ) { SetCmdErr;  return; }	  
			if (BMTRxBuff[3]==0x02)
			{
				CCP1CON = 0x0F;   // turn PWM on, zero output
				CCPR1L = 0; 
				CCPFunc |= CCP1Func;
				DirCCP1 = DirOutput;
			}
			else
			{
				DirCCP1 = (BMTRxBuff[3] ? DirOutput : DirInput); 
				CCPFunc &= ~CCP1Func;
			}
			break;
		case 0x1D:  // CCP2 digital pin
			if (BMTRxEnd<4 || BMTRxBuff[3]>0x02 ) { SetCmdErr;  return; }	  
			if (BMTRxBuff[3]==0x02)
			{
				CCP2CON = 0x0F;   // turn PWM on, zero output
				CCPR2L = 0; 
				CCPFunc |= CCP2Func;
				DirCCP2 = DirOutput;
			}
			else
			{
				DirCCP2 = (BMTRxBuff[3] ? DirOutput : DirInput); 
				CCPFunc &= ~CCP2Func;
			}
			break;
		case 0x1E:  // CCP3 digital pin
			if (BMTRxEnd<4 || BMTRxBuff[3]>0x02 ) { SetCmdErr;  return; }	  
			if (BMTRxBuff[3]==0x02)
			{
				CCP3CON = 0x0F;   // turn PWM on, zero output
				CCPR3L = 0; 
				CCPFunc |= CCP3Func;
				DirCCP3 = DirOutput;
			}
			else
			{
				DirCCP3 = (BMTRxBuff[3] ? DirOutput : DirInput); 
				CCPFunc &= ~CCP3Func;
			}
			break;
		case 0x1F:  // CCP4 digital pin
			if (BMTRxEnd<4 || BMTRxBuff[3]>0x02 ) { SetCmdErr;  return; }	  
			if (BMTRxBuff[3]==0x02)
			{
				CCP4CON = 0x0F;   // turn PWM on, zero output
				CCPR4L = 0; 
				CCPFunc |= CCP4Func;
				DirCCP4 = DirOutput;
			}
			else
			{
				DirCCP4 = (BMTRxBuff[3] ? DirOutput : DirInput); 
				CCPFunc &= ~CCP4Func;
			}
			break;
		case 0x20:  // CCP5 digital pin
			if (BMTRxEnd<4 || BMTRxBuff[3]>0x02 ) { SetCmdErr;  return; }	  
			if (BMTRxBuff[3]==0x02)
			{
				CCP5CON = 0x0F;   // turn PWM on, zero output
				CCPR5L = 0; 
				CCPFunc |= CCP5Func;
				DirCCP5 = DirOutput;
			}
			else
			{
				DirCCP5 = (BMTRxBuff[3] ? DirOutput : DirInput); 
				CCPFunc &= ~CCP5Func;
			}
			break;
		case 0x21:  // INT0 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirINT0 = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x22:  // INT1 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirINT1 = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x23:  // SCL digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirSCL = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x24:  // SDA digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirSDA = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x25:  // SDO digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirSDO = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x26:  // TxD digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirTxD = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x27:  // RxD digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirRxD = (BMTRxBuff[3] ? DirOutput : DirInput); 
			break;
		default:
			SetCmdErr;  
			return; 
	}

	// generate OK Response
	SendOKResponse();
}

void SetIOCmd( void )
{
	// error if no 3rd byte
	if (BMTRxEnd<3) { SetCmdErr;  return; }

	// read 3rd byte to identify the property being set
	switch (BMTRxBuff[2])
	{
		case 0x06:  // Parallel I/O A
			if ( BMTRxEnd<4 ) { SetCmdErr;  return; }		
			switch (ParallelA)
			{
				case 0x07:  SetPA7(  BMTRxBuff[3] ); break;
				case 0x06:  SetPA6(  BMTRxBuff[3] ); break;
				case 0x05:  SetPA5(  BMTRxBuff[3] ); break;
				case 0x04:  SetPA4(  BMTRxBuff[3] ); break;
				case 0x03:  SetPA3(  BMTRxBuff[3] ); break;
				case 0x02:  SetPA2(  BMTRxBuff[3] ); break;
				default: break;
			}
			break;
		case 0x07:  // Parallel I/O B
			if ( BMTRxEnd<4 ) { SetCmdErr;  return; }		
			switch (ParallelB)
			{
				case 0x05:  SetPB5(  BMTRxBuff[3] ); break;
				case 0x04:  SetPB4(  BMTRxBuff[3] ); break;
				case 0x03:  SetPB3(  BMTRxBuff[3] ); break;
				case 0x02:  SetPB2(  BMTRxBuff[3] ); break;
				default: break;
			}
			break;
		case 0x08:  // Parallel I/O C
			if ( BMTRxEnd<4 ) { SetCmdErr;  return; }		
			switch (ParallelC)
			{
				case 0x05:  SetPC5(  BMTRxBuff[3] ); break;
				case 0x04:  SetPC4(  BMTRxBuff[3] ); break;
				case 0x03:  SetPC3(  BMTRxBuff[3] ); break;
				case 0x02:  SetPC2(  BMTRxBuff[3] ); break;
				default: break;
			}
			break;
		case 0x10:  // AN0 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN0Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x11:  // AN1 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN1Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x12:  // AN2 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN2Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x13:  // AN3 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN3Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x14:  // AN4 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN4Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x15:  // AN5 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN5Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x16:  // AN6 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN6Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x17:  // AN7 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN7Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x18:  // AN8 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN8Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x19:  // AN9 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN9Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x1A:  // AN10 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN10Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x1B:  // AN11 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN11Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x1C:  // CCP1 PWM / digital pin
			if (BMTRxEnd<4) { SetCmdErr;  return; }	  
			if (CCPFunc & CCP1Func)  // if is PWM
			{
				unsigned char iVal = BMTRxBuff[4];  		  // Note, high byte sent first since more intuitive for ASCII commands; FlexiPanel Ltd is usually little-endian
				if (BMTRxEnd<5) { SetCmdErr;  return; }	  
				// lower 2 bits
				CCP1CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
				// upper 8 bits
				iVal = (iVal >> 2) + (BMTRxBuff[3] << 6);
				CCPR1L = (unsigned char)iVal;
			}
			else
				CCP1Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x1D:  // CCP2 PWM / digital pin
			if (BMTRxEnd<4) { SetCmdErr;  return; }	  
			if (CCPFunc & CCP2Func)  // if is PWM
			{
				unsigned char iVal = BMTRxBuff[4];
				if (BMTRxEnd<5) { SetCmdErr;  return; }	  
				// lower 2 bits
				CCP2CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
				// upper 8 bits
				iVal = (iVal >> 2) + (BMTRxBuff[3] << 6);
				CCPR2L = (unsigned char)iVal;
			}
			else
				CCP2Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x1E:  // CCP3 PWM / digital pin
			if (BMTRxEnd<4) { SetCmdErr;  return; }	  
			if (CCPFunc & CCP3Func)  // if is PWM
			{
				unsigned char iVal = BMTRxBuff[4];
				if (BMTRxEnd<5) { SetCmdErr;  return; }	  
				// lower 2 bits
				CCP3CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
				// upper 8 bits
				iVal = (iVal >> 2) + (BMTRxBuff[3] << 6);
				CCPR3L = (unsigned char)iVal;
			}
			else
				CCP3Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x1F:  // CCP4 PWM / digital pin
			if (BMTRxEnd<4) { SetCmdErr;  return; }	  
			if (CCPFunc & CCP4Func)  // if is PWM
			{
				unsigned char iVal = BMTRxBuff[4];
				if (BMTRxEnd<5) { SetCmdErr;  return; }	  
				// lower 2 bits
				CCP4CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
				// upper 8 bits
				iVal = (iVal >> 2) + (BMTRxBuff[3] << 6);
				CCPR4L = (unsigned char)iVal;
			}
			else
				CCP4Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x20:  // CCP5 PWM / digital pin
			if (BMTRxEnd<4) { SetCmdErr;  return; }	  
			if (CCPFunc & CCP5Func)  // if is PWM
			{
				unsigned char iVal = BMTRxBuff[4];
				if (BMTRxEnd<5) { SetCmdErr;  return; }	  
				// lower 2 bits
				CCP5CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
				// upper 8 bits
				iVal = (iVal >> 2) + (BMTRxBuff[3] << 6);
				CCPR5L = (unsigned char)iVal;
			}
			else
				CCP5Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x21:  // INT0 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			INT0Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x22:  // INT1 digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			INT1Pin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x23:  // SCL digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			SCLPin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x24:  // SDA digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			SDAPin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x25:  // SDO digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			SDOPin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x26:  // TxD digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			TxDPin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x27:  // RxD digital pin
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			RxDPin = (BMTRxBuff[3] ? 1 : 0); 
			break;
		case 0x30:  // Green LED
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			LedGreen = (BMTRxBuff[3] ? LedGreenOn : LedGreenOff); 
			break;
		case 0x31:  // Red LED
			if (BMTRxEnd<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			LedRed = (BMTRxBuff[3] ? LedRedOn : LedRedOff); 
			break;
		default:
			SetCmdErr;  
			return; 
	}

	// generate OK Response
	SendOKResponse();
}

void GetIOCmd( void )
{
	// error if no 3rd byte
	if (BMTRxEnd<3) { SetCmdErr;  return; }

	// read 3rd byte to identify the property being requested
	switch (BMTRxBuff[2])
	{
		case 0x06:  // Parallel I/O A
			switch (ParallelA)
			{
				case 0x17:  SendIOValResponse( BMTRxBuff[2], GetPA7 ); return;
				case 0x16:  SendIOValResponse( BMTRxBuff[2], GetPA6 ); return;
				case 0x15:  SendIOValResponse( BMTRxBuff[2], GetPA5 ); return;
				case 0x14:  SendIOValResponse( BMTRxBuff[2], GetPA4 ); return;
				case 0x13:  SendIOValResponse( BMTRxBuff[2], GetPA3 ); return;
				case 0x12:  SendIOValResponse( BMTRxBuff[2], GetPA2 ); return;
				default: 
					SetCmdErr;
					return;
			}
		case 0x07:  // Parallel I/O B
			switch (ParallelB)
			{
				case 0x15:  SendIOValResponse( BMTRxBuff[2], GetPB5 ); return;
				case 0x14:  SendIOValResponse( BMTRxBuff[2], GetPB4 ); return;
				case 0x13:  SendIOValResponse( BMTRxBuff[2], GetPB3 ); return;
				case 0x12:  SendIOValResponse( BMTRxBuff[2], GetPB2 ); return;
				default: 
					SetCmdErr;
					return;
			}
		case 0x0C:  // Parallel I/O C
			switch (ParallelC)
			{
				case 0x15:  SendIOValResponse( BMTRxBuff[2], GetPC5 ); return;
				case 0x14:  SendIOValResponse( BMTRxBuff[2], GetPC4 ); return;
				case 0x13:  SendIOValResponse( BMTRxBuff[2], GetPC3 ); return;
				case 0x12:  SendIOValResponse( BMTRxBuff[2], GetPC2 ); return;
				default: 
					SetCmdErr;
					return;
			}
		case 0x10:  // AN0 analog / digital pin
			if (NumAnalog>0)
				SendADResponse( BMTRxBuff[2] );
			else
				SendIOValResponse( BMTRxBuff[2], AN0Pin ); 
			return;
		case 0x11:  // AN1 analog / digital pin
			if (NumAnalog>0)
				SendADResponse( BMTRxBuff[2] );
			else
				SendIOValResponse( BMTRxBuff[2], AN1Pin ); 
			return;
		case 0x12:  // AN2 analog / digital pin
			if (NumAnalog>0)
				SendADResponse( BMTRxBuff[2] );
			else
				SendIOValResponse( BMTRxBuff[2], AN2Pin ); 
			return;
		case 0x13:  // AN3 analog / digital pin
			if (NumAnalog>0)
				SendADResponse( BMTRxBuff[2] );
			else
				SendIOValResponse( BMTRxBuff[2], AN3Pin ); 
			return;
		case 0x14:  // AN4 analog / digital pin
			if (NumAnalog>0)
				SendADResponse( BMTRxBuff[2] );
			else
				SendIOValResponse( BMTRxBuff[2], AN4Pin ); 
			return;
		case 0x15:  // AN5 analog / digital pin
			if (NumAnalog>0)
				SendADResponse( BMTRxBuff[2] );
			else
				SendIOValResponse( BMTRxBuff[2], AN5Pin ); 
			return;
		case 0x16:  // AN6 analog / digital pin
			if (NumAnalog>0)
				SendADResponse( BMTRxBuff[2] );
			else
				SendIOValResponse( BMTRxBuff[2], AN6Pin ); 
			return;
		case 0x17:  // AN7 analog / digital pin
			if (NumAnalog>0)
				SendADResponse( BMTRxBuff[2] );
			else
				SendIOValResponse( BMTRxBuff[2], AN7Pin ); 
			return;
		case 0x18:  // AN8 analog / digital pin
			if (NumAnalog>0)
				SendADResponse( BMTRxBuff[2] );
			else
				SendIOValResponse( BMTRxBuff[2], AN8Pin ); 
			return;
		case 0x19:  // AN9 analog / digital pin
			if (NumAnalog>0)
				SendADResponse( BMTRxBuff[2] );
			else
				SendIOValResponse( BMTRxBuff[2], AN9Pin ); 
			return;
		case 0x1A:  // AN10 analog / digital pin
			if (NumAnalog>0)
				SendADResponse( BMTRxBuff[2] );
			else
				SendIOValResponse( BMTRxBuff[2], AN10Pin ); 
			return;
		case 0x1B:  // AN11 analog / digital pin
			if (NumAnalog>0)
				SendADResponse( BMTRxBuff[2] );
			else
				SendIOValResponse( BMTRxBuff[2], AN11Pin ); 
			return;
		case 0x1C:  // CCP1 digital pin
			SendIOValResponse( BMTRxBuff[2], CCP1Pin ); 
			return;
		case 0x1D:  // CCP2 digital pin
			SendIOValResponse( BMTRxBuff[2], CCP2Pin ); 
			return;
		case 0x1E:  // CCP3 digital pin
			SendIOValResponse( BMTRxBuff[2], CCP3Pin ); 
			return;
		case 0x1F:  // CCP4 digital pin
			SendIOValResponse( BMTRxBuff[2], CCP4Pin ); 
			return;
		case 0x20:  // CCP5 digital pin
			SendIOValResponse( BMTRxBuff[2], CCP5Pin ); 
			return;
		case 0x21:  // INT0 digital pin
			SendIOValResponse( BMTRxBuff[2], INT0Pin );
			return;
		case 0x22:  // INT1 digital pin
			SendIOValResponse( BMTRxBuff[2], INT1Pin );
			return;
		case 0x23:  // SCL digital pin
			SendIOValResponse( BMTRxBuff[2], SCLPin );
			return;
		case 0x24:  // SDA digital pin
			SendIOValResponse( BMTRxBuff[2], SDAPin );
			return;
		case 0x25:  // SDO digital pin
			SendIOValResponse( BMTRxBuff[2], SDOPin );
			return;
		case 0x26:  // TxD digital pin
			SendIOValResponse( BMTRxBuff[2], TxDPin );
			return;
		case 0x27:  // RxD digital pin
			SendIOValResponse( BMTRxBuff[2], RxDPin );
			return;
		case 0x30:  // Green LED
			SendIOValResponse( BMTRxBuff[2], (LedGreen==LedGreenOn) ? 0x01 : 0x00 );
			break;
		case 0x31:  // Red LED
			SendIOValResponse( BMTRxBuff[2], (LedRed==LedRedOn) ? 0x01 : 0x00 );
			break;
		case 0x32:  // Pushbutton
			SendIOValResponse( BMTRxBuff[2], (PushButton==PushButtonOn) ? 0x01 : 0x00 );
			break;
		default:
			SetCmdErr;  
			return; 
	}
}


void SendIOValResponse( unsigned char iProperty, unsigned char iValue )
{
	SendResponseBytes( pResIOVal, 0, ResIOValLen, SRB_FIRST );
	SendResponseBytes( 0, (unsigned char *)&iProperty, sizeof(iProperty), 0 );
	SendResponseBytes( 0, (unsigned char *)&iValue, sizeof(iValue), SRB_LAST );
}

void GetADVal( unsigned char iChan )
{
	SetADChan( iChan );  					// select a to d channel
	CyclesDelay3p2plus3p2times( 4 ); 		// 16us delay to charge S & H cap (refer to PIC18F6720 manual)
	StartAtoD;								// start a to d conversion
	AwaitAtoDComplete;						// await end of conversion
}

void SendADResponse( unsigned char iProperty )
{
	GetADVal( iProperty - 0x10  );  	

	SendResponseBytes( pResIOVal, 0, ResIOValLen, SRB_FIRST );
	SendResponseBytes( 0, (unsigned char *)&iProperty, sizeof(iProperty), 0 );
	SendResponseBytes( 0, (unsigned char *)&ADRESL, 1, 0 );
	SendResponseBytes( 0, (unsigned char *)&ADRESH, 1, SRB_LAST );
}

void GetMemCmd( void )
{
	// bytes 2 = storage type
	// bytes 3/4 = ADRESs
	// byte 5 = number of bytes required, up to 120 bytes (ascii) / 248 bytes (binary)
	unsigned char numBytes = BMTRxBuff[5];

	// error if not 6 bytes - no other error checking done
	if (BMTRxEnd!=6) { SetCmdErr;  return; }
	
	// error if more than 120 / 248 bytes requested
	if (Darc_ASCIIMessages && numBytes>120) { SetCmdErr;  return; }
	if (!Darc_ASCIIMessages && numBytes>248) { SetCmdErr;  return; }

	// get data 
	GetBytes( BMTRxBuff[2], *((ADD*)(&BMTRxBuff[3])), (unsigned char *)&BMTRxBuff[5], numBytes );

	// transmit result
	SendResponseBytes( pResGotBytes, 0, ResGotBytesLen, SRB_FIRST );
	SendResponseBytes( 0, (unsigned char *)&BMTRxBuff[2], 3, 0 );
	SendResponseBytes( 0, (unsigned char *)&BMTRxBuff[5], numBytes, SRB_LAST );
}


void SetMemCmd( void )
{
	// bytes 2 = storage type
	// bytes 3/4 = ADRESs
	// remainder = data to set

	// error if less than 5 bytes - no other error checking done
	if (BMTRxEnd<5) { SetCmdErr;  return; }

	// set data 
	SetBytes( BMTRxBuff[2], *((ADD*)(&BMTRxBuff[3])), 0, (unsigned char *)&BMTRxBuff[5], BMTRxBuff[0] - 5 );

	// acknowledge 
	SendOKResponse();
}


unsigned char SetHex( unsigned char * pBuff, unsigned char ChASCII )
{
	*pBuff = (ChASCII >> 4) + '0';
	if (*pBuff > '9') *pBuff += 0x07;
	pBuff++;
	*pBuff = (ChASCII & 0x0F) + '0';
	if (*pBuff > '9') *pBuff += 0x07;
}


void StreamFrameCmd( unsigned char bStream )
{
	DarcSF &= ~DarcSF_StreamError;
	DarcSF &= ~DarcSF_ProcessingStrFrm;

	// turn off real time clock
	PIE1bits.TMR1IE = 0;

	// initialize streaming / framing
	if (bStream)
	{
		if (BMTRxEnd!=0x02) { SetCmdErr;  return; }
		DarcSF |= DarcSF_Streaming ;
		DarcSF &= ~DarcSF_Framing;
	}
	else
	{
		if (BMTRxEnd!=0x04) { SetCmdErr;  return; }
		FLMemPtr = 0x0000;
		FLMemType = FLMemStart;
		DarcSF |= DarcSF_Framing ;
		DarcSF &= ~DarcSF_Streaming;
		NumFrameLeft = BMTRxBuff[3];
		NumFrameLeft = (NumFrameLeft<<8) + BMTRxBuff[2];
	}

	// start stream / frame timer interrupt
	Countdown = StreamRate;
	T4CON = 0;
	TMR4 = 0;
	IPR3bits.TMR4IP = 0;  // low pri
	PIR3bits.TMR4IF = 0;  // clear int
	PIE3bits.TMR4IE = 1;  // enable int

	// trigger every 1000 clock cycles = every 200us
	PR4 = 250;  // 250 * 4 = 1000 clock cycles
	T4CON = 0b00000101;  // pre scale = 4, postscale = 1:1
}

// find next place to store data
void FindNextRecordLoc( unsigned char nBytes )
{
	// try next location in this memory block
	unsigned char bGoneOffTheEnd = False;
	unsigned short long NewPtr = FLMemPtr + nBytes;
	if (FLMemType == STR_RAM && NewPtr >= 0x900 - nBytes) bGoneOffTheEnd = True;
	if (FLMemType == STR_EE  && NewPtr >= 0x400 - nBytes) bGoneOffTheEnd = True;
	if (FLMemType == STR_ROM01 && NewPtr >= 0x00F000 - nBytes) bGoneOffTheEnd = True;
	if (FLMemType >= STR_EXT0 && FLMemType <= STR_EXT7 && NewPtr >= 0x8000 - nBytes) bGoneOffTheEnd = True;

	if (!bGoneOffTheEnd) 
	{
		FLMemPtr = (ADD) NewPtr;
		return;
	}

	// if fail, move to next memory block
	FLMemPtr = 0x0000;
	if (FLMemType==FLMemEnd || FLMemType == STR_ROM01) 
	{
		FLMemType = FLMemStart;
		return;
	}
	if (FLMemType == STR_EXT7) FLMemType = STR_ROM01;
	if (FLMemType >= STR_EXT0 && FLMemType <= STR_EXT6) FLMemType += 1;
	if (FLMemType == STR_EE) FLMemType = STR_EXT0;
	if (FLMemType == STR_RAM) FLMemType = STR_EE;

}

// break code into sections to make life easy for linker
#pragma code DarcSECTION2


// get record data and return record length
unsigned char GetRecord( void )
{
	unsigned char nChar = 0;

	if ( LogPar & LPar_A )
	{
		switch (ParallelA)
		{
			case 0x17:  record[nChar] = GetPA7 ; break;
			case 0x16:  record[nChar] = GetPA6 ; break;
			case 0x15:  record[nChar] = GetPA5 ; break;
			case 0x14:  record[nChar] = GetPA4 ; break;
			case 0x13:  record[nChar] = GetPA3 ; break;
			case 0x12:  record[nChar] = GetPA2 ; break;
			default: 	record[nChar] = 0; 		 break;
		}
		nChar ++;  
	}
	if ( LogPar & LPar_B )
	{
		switch (ParallelB)
		{
			case 0x15:  record[nChar] = GetPB5 ; break;
			case 0x14:  record[nChar] = GetPB4 ; break;
			case 0x13:  record[nChar] = GetPB3 ; break;
			case 0x12:  record[nChar] = GetPB2 ; break;
			default: 	record[nChar] = 0; 		 break;
		}
		nChar ++;  
	}
	if ( LogPar & LPar_C )
	{
		switch (ParallelC)
		{
			case 0x15:  record[nChar] = GetPC5 ; break;
			case 0x14:  record[nChar] = GetPC4 ; break;
			case 0x13:  record[nChar] = GetPC3 ; break;
			case 0x12:  record[nChar] = GetPC2 ; break;
			default: 	record[nChar] = 0; 		 break;
		}
		nChar ++;  
	}

	if ( LogAN07 & LAN07_AN0 )
	{
		if (NumAnalog>0)
		{
			GetADVal( 0x00 );
			record[nChar] = ADRESL ;
			nChar ++;
			record[nChar] = ADRESH ;
			nChar ++;
		}
		else
		{
			record[nChar] = AN0Pin;
			nChar ++;
		}
	}

	if ( LogAN07 & LAN07_AN1 )
	{
		if (NumAnalog>1)
		{
			GetADVal( 0x01 );
			record[nChar] = ADRESL ;
			nChar ++;
			record[nChar] = ADRESH ;
			nChar ++;
		}
		else
		{
			record[nChar] = AN1Pin;
			nChar ++;
		}
	}

	if ( LogAN07 & LAN07_AN2 )
	{
		if (NumAnalog>2)
		{
			GetADVal( 0x02 );
			record[nChar] = ADRESL ;
			nChar ++;
			record[nChar] = ADRESH ;
			nChar ++;
		}
		else
		{
			record[nChar] = AN2Pin;
			nChar ++;
		}
	}

	if ( LogAN07 & LAN07_AN3 )
	{
		if (NumAnalog>3)
		{
			GetADVal( 0x03 );
			record[nChar] = ADRESL ;
			nChar ++;
			record[nChar] = ADRESH ;
			nChar ++;
		}
		else
		{
			record[nChar] = AN3Pin;
			nChar ++;
		}
	}

	if ( LogAN07 & LAN07_AN4 )
	{
		if (NumAnalog>4)
		{
			GetADVal( 0x04 );
			record[nChar] = ADRESL ;
			nChar ++;
			record[nChar] = ADRESH ;
			nChar ++;
		}
		else
		{
			record[nChar] = AN4Pin;
			nChar ++;
		}
	}

	if ( LogAN07 & LAN07_AN5 )
	{
		if (NumAnalog>5)
		{
			GetADVal( 0x05 );
			record[nChar] = ADRESL ;
			nChar ++;
			record[nChar] = ADRESH ;
			nChar ++;
		}
		else
		{
			record[nChar] = AN5Pin;
			nChar ++;
		}
	}

	if ( LogAN07 & LAN07_AN6 )
	{
		if (NumAnalog>6)
		{
			GetADVal( 0x06 );
			record[nChar] = ADRESL ;
			nChar ++;
			record[nChar] = ADRESH ;
			nChar ++;
		}
		else
		{
			record[nChar] = AN6Pin;
			nChar ++;
		}
	}

	if ( LogAN07 & LAN07_AN7 )
	{
		if (NumAnalog>7)
		{
			GetADVal( 0x07 );
			record[nChar] = ADRESL ;
			nChar ++;
			record[nChar] = ADRESH ;
			nChar ++;
		}
		else
		{
			record[nChar] = AN7Pin;
			nChar ++;
		}
	}

	if ( LogAN811 & LAN811_AN8 )
	{
		if (NumAnalog>8)
		{
			GetADVal( 0x08 );
			record[nChar] = ADRESL ;
			nChar ++;
			record[nChar] = ADRESH ;
			nChar ++;
		}
		else
		{
			record[nChar] = AN8Pin;
			nChar ++;
		}
	}

	if ( LogAN811 & LAN811_AN9 )
	{
		if (NumAnalog>9)
		{
			GetADVal( 0x09 );
			record[nChar] = ADRESL ;
			nChar ++;
			record[nChar] = ADRESH ;
			nChar ++;
		}
		else
		{
			record[nChar] = AN9Pin;
			nChar ++;
		}
	}

	if ( LogAN811 & LAN811_AN10 )
	{
		if (NumAnalog>10)
		{
			GetADVal( 0x0A );
			record[nChar] = ADRESL ;
			nChar ++;
			record[nChar] = ADRESH ;
			nChar ++;
		}
		else
		{
			record[nChar] = AN10Pin;
			nChar ++;
		}
	}

	if ( LogAN811 & LAN811_AN11 )
	{
		if (NumAnalog>11)
		{
			GetADVal( 0x0B );
			record[nChar] = ADRESL ;
			nChar ++;
			record[nChar] = ADRESH ;
			nChar ++;
		}
		else
		{
			record[nChar] = AN11Pin;
			nChar ++;
		}
	}

	if ( LogCCP & LCCP_CCP1 )
	{
		record[nChar] = CCP1Pin;
		nChar ++;
	}

	if ( LogCCP & LCCP_CCP2 )
	{
		record[nChar] = CCP2Pin;
		nChar ++;
	}

	if ( LogCCP & LCCP_CCP3 )
	{
		record[nChar] = CCP3Pin;
		nChar ++;
	}

	if ( LogCCP & LCCP_CCP4 )
	{
		record[nChar] = CCP4Pin;
		nChar ++;
	}

	if ( LogCCP & LCCP_CCP5 )
	{
		record[nChar] = CCP5Pin;
		nChar ++;
	}

	if ( LogAUX & LAUX_INT0 )
	{
		record[nChar] = INT0Pin;
		nChar ++;
	}

	if ( LogAUX & LAUX_INT1 )
	{
		record[nChar] = INT1Pin;
		nChar ++;
	}

	if ( LogAUX & LAUX_SCL )
	{
		record[nChar] = SCLPin;
		nChar ++;
	}

	if ( LogAUX & LAUX_SDA )
	{
		record[nChar] = SDAPin;
		nChar ++;
	}

	if ( LogAUX & LAUX_SDO )
	{
		record[nChar] = SDOPin;
		nChar ++;
	}

	if ( LogAUX & LAUX_TxD )
	{
		record[nChar] = TxDPin;
		nChar ++;
	}

	if ( LogAUX & LAUX_RxD )
	{
		record[nChar] = RxDPin;
		nChar ++;
	}

	return nChar;
}

