//------------------------------------------------------------------------------------------------------
//
//  OpenTooth Access Controller main program loop
//
//  (c) FlexiPanel Ltd
//
//-------------------------------------------------------------------------------------------------------

#include <p18f6720.h>
#include <string.h>
#include "..\ToothPIC.h"
#include "OpenToothRes.h"

//#define USE_BLUETOOTH_EVAL_BOARD  // sets outputs to demo board leds

#ifdef USE_BLUETOOTH_EVAL_BOARD
# define IndicatorLED  	AN11Pin
# define IndicatorLEDDir DirAN11
# define Data1Pin		CCP1Pin
# define Data1Dir		DirCCP1
# define Data0Pin		CCP2Pin
# define Data0Dir		DirCCP2
# define InitRelay() 	DirAN9 = DirOutput
# define RelayOn 		AN9Pin = 1;
# define RelayOff 		AN9Pin = 0;
#else
# define IndicatorLED  	AN6Pin
# define IndicatorLEDDir DirAN6
# define Data1Pin		CCP5Pin
# define Data1Dir		DirCCP5
# define Data0Pin		CCP3Pin
# define Data0Dir		DirCCP3
# define InitRelay() 	{ SetPB4( 0x00 ); DirPB4( DirParOutput ); }
# define RelayOn 		SetPB4( 0x0F )
# define RelayOff 		SetPB4( 0x00 )
#endif


#define MyPushbutton  	AN7Pin 
#define MyPushbuttonDir DirAN7

unsigned char AccConSemaphores = 0;   
#define ACSF_SETUPMODE			0x01  /* raised on button long press, low on disconnect or timeout */
#define ACSF_SETUPNOTCONNECTED	0x02  /* raised on button long press, low on connect */
#define ACSF_INQUIRYINPROCESS	0x04  /* raised when device search is in process */
#define ACSF_RUNDELETIONCHECK	0x08  /* raised when device search is in process */
#define ACSF_RUNACCESSLOG		0x10  /* access granted; run access log at end of cycle  */
#define ACSF_NEWUSERMODE		0x20  /* new user can access using pin */

// trigger conditions
#define TRIG_ONARRIVAL 0
#define TRIG_ONARRIVALORBUTTON 1
#define TRIG_EACHDETECT 2
#define TRIG_EACHDETECTORBUTTON 3
#define TRIG_ONPRESS 4
#define TRIG_WHILEPRESENT 5

// log conditions
#define LL_NONE 0
#define LL_ONACCESS 1
#define LL_ALLDETECTS 2

// weigand options list
#define WEIG_NONE 0
#define WEIG_50_GRANTED 1
#define WEIG_26_GRANTED 2
#define WEIG_50_REQUESTED 3
#define WEIG_26_REQUESTED 4
#define WEIG_50_ALLDETECTS 5
#define WEIG_26_ALLDETECTS 6

// You will wish to invert these if you are not going to drive common-collector outputs
#define WEIG_5V 0
#define WEIG_0V 1

// development version has DEVELOPMENT defined to enable error reporting 
#define DEVELOPMENT 1

// text strings
rom unsigned char pInquireSecs[] = "F";  // 15 seconds inquire cycle
#define NOUSERSLEN 22
rom unsigned char pNoUsers[] = "(No registered users)";
rom unsigned char pLogEntryTemplate[] = "00:00:00 00/00/00                           ";
rom unsigned char pAnonymous[] = "Anon 000000000000";  // template for unknown devices that register
extern rom unsigned char * pszPIN;  // Out pin code declared in ToothPIC.c
rom unsigned char TrueVal = 0xFF;
rom unsigned char FalseVal = 0x00;

// short user entry structure consists of just BT addr and name
#define USERNAMELEN 19 /*including zero terminator */

typedef struct 
{
	unsigned char BluetoothID[BTADDRLEN];
	unsigned char UserName[USERNAMELEN];
} UserEntry2;

// long user entry consists of all features of registered user
typedef struct 
{
	DateTimeU ExpiryDT;
	unsigned char BluetoothID[BTADDRLEN];
	unsigned char UserName[USERNAMELEN];
	unsigned char FromSc;
	unsigned char FromMn;
	unsigned char FromHr;
	unsigned char ToSc;
	unsigned char ToMn;
	unsigned char ToHr;
	unsigned char DayFlags;
} UserEntry10;

#define DF_MONDAY 		0x01
#define DF_TUESDAY		0x02
#define DF_WEDNESDAY	0x04
#define DF_THURSDAY		0x08
#define DF_FRIDAY		0x10
#define DF_SATURDAY		0x20
#define DF_SUNDAY		0x40


#define INQUIRYMAX 10 
#pragma udata InquiryTable
UserEntry2 pInquiryTable[INQUIRYMAX];   	// table of encountered devices collaed during inquiry
unsigned char nInquiry;						// number of valid entries in InquiryTable

#pragma udata VicinityTable
UserEntry2 pVicinityTable[INQUIRYMAX];  	// table of devices currently in the vicinity
unsigned char nVicinity;					// number of valid entries in VicinityTable

#pragma udata

// Weigand network control
unsigned char bTrustedDeviceInVicinity = 0;
unsigned char bSleepModeRequested = 0;
void WiegandBit( unsigned char iBit );
void WiegandEvent(const unsigned char * pBTAddr);

// temporary storage of modified user data
UserEntry10 ModUser;
unsigned char ModUserIndex;

// locate user table at known location in memory
#define MAXUSERS 255

#define USERTABLELOC 0xC000
#define USERTABLESTR STR_ROM01
#pragma romdata UserList=0x1C000
rom UserEntry10 pUserList[MAXUSERS];  // new users list BT IDs

#define NUMUSERLOC 0xBFFC
#define NUMUSERSTR STR_ROM01
#pragma romdata NumUsers=0x1BFFC
rom unsigned char nUsers = 0;

#define NEXTLOGLOC 0x1BFF8
#define NEXTLOGSTR STR_ROM01
#pragma romdata LogLoc=0x1BFF8
rom unsigned long NextLogLoc = 0;

#pragma romdata 

void ShowModUserList( void );
void UpdateRegUserList( void );
unsigned char TimeConditionSatisfied( unsigned char i );
void TriggerRelay( void );
void LogUser( rom unsigned char * pRUserName, unsigned char * pUserName );
void ModifyUser( int iUser );
unsigned char GetLowerHexChar( unsigned char iCh );
unsigned char GetUpperHexChar( unsigned char iCh );

unsigned char ShortPressCount = 0;
unsigned char ShortPressStartSecs ;

/*awaiting prev error on loss of trusted device
DST not implemented
on while present not yet implemented
*/

void main( void )
{
	SetAnalogNone;

	// initialize inquiry and vicinity tables
	nInquiry = 0;
	nVicinity = 0;

	// set up I/O
	IndicatorLED = 0;
	IndicatorLEDDir = DirOutput;
	MyPushbuttonDir = DirInput;
	Data1Pin = WEIG_5V;
	Data0Pin = WEIG_5V;
	Data1Dir = DirOutput;
	Data0Dir = DirOutput;

	// relay pins
	InitRelay();

	BMTCommand( BMTC_Security, 0, Sec_Auth );    	// enable authentication
	AwaitBMTOK;	

	// ensure pin is correct
	SetBytes( STR_ROM00, (ADD)&pszPIN, pKey_1C, 0, SZPINMAX );    		

	while ( 1 ) 
	{
		// reset short press counter if needed
		if (ShortPressCount > 0 && (60 + RealTimeClock.sec - ShortPressStartSecs) % 60 > 5 ) ShortPressCount = 0;

		// if pushbutton pressed and not in setup mode
		if (MyPushbutton && !(AccConSemaphores&ACSF_SETUPMODE) && !(AccConSemaphores&ACSF_NEWUSERMODE) )
		{
			// wait up to 5 seconds for button release to decide action
			unsigned char iDelCount;
			for (iDelCount = 250; iDelCount > 0; iDelCount--)
			{
				msDelay( 20 );
				if (!MyPushbutton) break;
			}

			if (iDelCount)
			{
				// short press of button

				// detect whether five short presses within 5 seconds
				if (ShortPressCount==0)  ShortPressStartSecs = RealTimeClock.sec;
				ShortPressCount++;

				// if five presses and space for new users and there is a new user pin specified, go into new user mode
				if (ShortPressCount==5 && nUsers < MAXUSERS && (*pUserPinKey_42)!=0 )
				{
					ShortPressCount = 0;
					
					// log state as in new user mode 
					AccConSemaphores |= ACSF_NEWUSERMODE;
	
					// if currently inquiring, stop
					if (AccConSemaphores&ACSF_INQUIRYINPROCESS)
					{
						BMTCommand( BMTC_Cancel , 0, 0 );
						AwaitBMTOK; 						
						
						// allow remote devices responde to change in mode 
						// this code inserted during development 
						// not sure whether this is still required 
						msDelay( 250 );
						msDelay( 250 );
						msDelay( 250 );
						msDelay( 250 );
						
						AccConSemaphores &= ~ACSF_INQUIRYINPROCESS;		
					}
	
					// set pin to user pin
					SetBytes( STR_ROM00, (ADD)&pszPIN, pUserPinKey_42, 0, SZPINMAX );    		
				
					BMTCommand( BMTC_Slave, 0, 0 );			// Start Bluematik service
					AwaitBMTOK; 							// Service operating, no connection
				}
				else
				{
					unsigned char j;

					// does not amount to five pressed of button

					// if open on press only or on arrival, open if trusted device around
					if ( ((*pcTriggerList_16 == TRIG_ONPRESS || *pcTriggerList_16 == TRIG_ONARRIVAL) && bTrustedDeviceInVicinity ) ||
						 (*pcTriggerList_16 == TRIG_ONARRIVALORBUTTON) || ( *pcTriggerList_16 == TRIG_EACHDETECTORBUTTON) )
					{
						TriggerRelay();
					}
	
					// if open on each detect, toggle sleep mode
					if ( (*pcTriggerList_16 == TRIG_EACHDETECT ||*pcTriggerList_16 == TRIG_EACHDETECTORBUTTON) && bTrustedDeviceInVicinity )
					{
						if (bSleepModeRequested)
						{
							bSleepModeRequested = 0;
							TriggerRelay();
						}
						else
						{
							bSleepModeRequested = 0xFF;
						}
					}

					// output all users in vicinity to Weigand 
					if (*pcWeigandStyle_4D==WEIG_50_REQUESTED || *pcWeigandStyle_4D==WEIG_26_REQUESTED) 
					{
					    for (j=0; j<nVicinity; j++)
					    {
							WiegandEvent( pVicinityTable[j].BluetoothID );
							msDelay(10); // 10ms extra delay
						}
			
				  	}
			
				}
			}
			else
			{
				// long press of button

				// log state as in setup mode but nothing connected
				AccConSemaphores |= ACSF_SETUPMODE + ACSF_SETUPNOTCONNECTED;

				// if currently inquiring, stop
				if (AccConSemaphores&ACSF_INQUIRYINPROCESS)
				{
					BMTCommand( BMTC_Cancel , 0, 0 );

					// allow remote devices responde to change in mode 
					// this code inserted during development 
					// not sure whether this is still required 
					msDelay( 250 );
					msDelay( 250 );
					msDelay( 250 );
					msDelay( 250 );
						
					AccConSemaphores &= ~ACSF_INQUIRYINPROCESS;		
				}

				FxPCommand( FxPC_Start, 0, 0 );		// Start FlexiPanel service
				AwaitBMTOK; 							// Service operating, no client
			}

		}

		// if unconnected but in setup mode, flash light until device connects, button is pressed gain or times out
		if (AccConSemaphores & ACSF_SETUPNOTCONNECTED)
		{
			unsigned short iTimeOutCount;
			unsigned char bButtonReleasedYet = 0x00;

			// timeout is 300 seconds
			for (iTimeOutCount = 1200; 1; iTimeOutCount--)
			{
				// if connection happened, break;
				if (!(AccConSemaphores & ACSF_SETUPNOTCONNECTED))
				{
					IndicatorLED = 1;
					break;
				}

				// flash LED
				IndicatorLED = ~IndicatorLED;

				// see if button release yet
				if (!MyPushbutton) bButtonReleasedYet = 0xFF;

				// if button pressed or timeout, reset
				if (!iTimeOutCount || (bButtonReleasedYet && MyPushbutton) ) 
				{
					AccConSemaphores = 0;
					Reset();
				}

				// wait 1/4 second and test all conditions again
				msDelay( 250 );			
			}

		}

		// if in new user mode, flash until timeout
		if (AccConSemaphores & ACSF_NEWUSERMODE)
		{
			unsigned short iTimeOutCount;
			unsigned char bButtonReleasedYet = 0x00;

			// Tell ToothPIC not to store the link key if a device pairs with it
			// This flag will reset automatically when the device resets
			ToothPICSemaphores2 |= TPSF2_DONTSTORELINK;

			// timeout is 300 seconds
			for (iTimeOutCount = 1200; 1; iTimeOutCount--)
			{
				// flash LED
				IndicatorLED = ~IndicatorLED;

				// see if button release yet
				if (!MyPushbutton) bButtonReleasedYet = 0xFF;

				// if button pressed or timeout, reset
				if (!iTimeOutCount || (bButtonReleasedYet && MyPushbutton) ) 
				{
					AccConSemaphores = 0;
					Reset();
				}

				// wait 1/4 second and test all conditions again
				msDelay( 250 );			
			}

		}
	
		// if no device search is in process, start one
		if ( !(AccConSemaphores&ACSF_INQUIRYINPROCESS) && !(AccConSemaphores & ACSF_SETUPMODE)  && !(AccConSemaphores & ACSF_NEWUSERMODE) )
		{
			if (bTrustedDeviceInVicinity)
			{
				// Occult LED to show inquiry starting
				IndicatorLED = 0;
				msDelay( 150 );
				IndicatorLED = 1;
			}
			else
			{
				// Blink LED to show inquiry starting
				IndicatorLED = 1;
				msDelay( 50 );
				IndicatorLED = 0;
			}

			nInquiry = 0;
			AccConSemaphores |= ACSF_INQUIRYINPROCESS;
			BMTCommand( BMTC_Inquire , 0, pInquireSecs );
		}

		// run deletion check if needed
		if (AccConSemaphores&ACSF_RUNDELETIONCHECK)
		{
			DateTimeU TmpClk;
			unsigned char i;

			DisableHighInts;
			memcpy( &TmpClk, (const void *) &RealTimeClock, sizeof(RealTimeClock) );
			EnableHighInts;

			for (i=0; i<nUsers; i++)
			{
				unsigned char DeleteMe = False;
				if (TmpClk.year < pUserList[i].ExpiryDT.year) continue;
				if (TmpClk.year > pUserList[i].ExpiryDT.year) DeleteMe = True;
				if (!DeleteMe && TmpClk.month < pUserList[i].ExpiryDT.month) continue;
				if (TmpClk.month > pUserList[i].ExpiryDT.month) DeleteMe = True;
				if (!DeleteMe && TmpClk.date < pUserList[i].ExpiryDT.date) continue;
				if (TmpClk.date > pUserList[i].ExpiryDT.date) DeleteMe = True;
				if (!DeleteMe && TmpClk.hour < pUserList[i].ExpiryDT.hour) continue;
				if (TmpClk.hour > pUserList[i].ExpiryDT.hour) DeleteMe = True;
				if (!DeleteMe && TmpClk.min < pUserList[i].ExpiryDT.min) continue;
				if (TmpClk.min > pUserList[i].ExpiryDT.min) DeleteMe = True;

				if (DeleteMe)
				{
					long lVal;

					// time to delete user
					for (; i<nUsers-1; i++)
					{
						unsigned char pTempData[ sizeof( UserEntry10 ) ];
						GetBytes( USERTABLESTR, USERTABLELOC + sizeof(UserEntry10) * (i+1), pTempData, sizeof(UserEntry10) );
						SetBytes( USERTABLESTR, USERTABLELOC + sizeof(UserEntry10) * i, 0, pTempData, sizeof(UserEntry10) );
					}
		
					// decrease nUsers
					lVal = nUsers - 1;
					SetBytes( NUMUSERSTR, NUMUSERLOC, 0, &lVal, sizeof(long) );
		
					// update registered users list box
					UpdateRegUserList();
				}
			}

			AccConSemaphores &= ~ACSF_RUNDELETIONCHECK;
		}


	}

}



// HighInterrupt handler
void HighInterrupt (void)
{
}

// LowInterrupt handler
void LowInterrupt (void)
{
	// Once-per-second tick
	if (IsSWI( SWI_Tick ) )
	{
		// update summertime dialog
		if (*pIn_DST_now_4E == 0)  // winter
		{
			if (DSTEvent & DST_SUMMER_FLAG)  SetUp_In_DST_now_4E( &TrueVal, 0 );
		}
		else
		{
			if ( (DSTEvent & DST_SUMMER_FLAG)==0 )  SetUp_In_DST_now_4E( &FalseVal, 0 );
		}
				
		// once a minute, raise deletion check flag
		if (!RealTimeClock.sec)
		{
			AccConSemaphores |= ACSF_RUNDELETIONCHECK;
		}

		ClearSWI( SWI_Tick );		// Clear interrupt flag
		return;
	}

	// incoming non-flexipanel data
	if (IsSWI( SWI_BMTData ) )
	{
		// A non-FlexiPanel device has connected  Reset immediately
		AccConSemaphores = 0;
		Reset();		
	}
}


// Error event (Operating System callback)
void ErrorStatus( unsigned char ErrNum )
{
	// controls may be updated while the flexipanel server is not operating; this is permitted
	// but generates a ERR_NOFLEXIPANELSERVICE warning which needs to be ignored
	if (ErrNum==ERR_NOFLEXIPANELSERVICE) return;

#ifdef DEVELOPMENT
	Breakpoint( ErrNum );
#else
	// unexpected error; reset
	AccConSemaphores = 0;
	Reset();
#endif
}


// BlueMatik radio events (Operating System callback)
void BMTEvent( unsigned char EventID, void *pData1, void *pData2 )
{
	if ( EventID==BMTE_Error || EventID==BMTE_Warning )
	{
#ifdef DEVELOPMENT
		Breakpoint( *((unsigned char*)pData1) );
#else
		// unexpected error; reset
		AccConSemaphores = 0;
		Reset();
#endif
		return;
	}

	// device found
	if (EventID==BMTE_Found)
	{
		unsigned char i;

		// add to inquiry table
		if (nInquiry<INQUIRYMAX)
		{
			memcpy( pInquiryTable[nInquiry].BluetoothID, ((const unsigned char*)pData1), BTADDRLEN );
			memcpy( pInquiryTable[nInquiry].UserName, ((const unsigned char*)pData2), USERNAMELEN );
			pInquiryTable[nInquiry].UserName[USERNAMELEN-1] = 0;
			nInquiry ++;
		}

		// if trusted device, raise trusted device flag
		for (i = 0; i< nUsers; i++)
		{
			if (memcmpram2pgm( pUserList[i].BluetoothID, ((const unsigned char*)pData1), BTADDRLEN) == 0 &&
					TimeConditionSatisfied( i ) )
			{
				IndicatorLED = 1;

				// activate relay if mode is on while present
				if (*pcTriggerList_16 == TRIG_WHILEPRESENT)  RelayOn;

				// activate relay if on arrival or (and first time) each detection condition
				if ( ( (*pcTriggerList_16 == TRIG_ONARRIVAL || *pcTriggerList_16 == TRIG_ONARRIVALORBUTTON) && !bTrustedDeviceInVicinity) || 
					 ( (*pcTriggerList_16 == TRIG_EACHDETECT || *pcTriggerList_16 == TRIG_EACHDETECTORBUTTON)  && !bSleepModeRequested ) ) 
				{
					TriggerRelay();
				}

				bTrustedDeviceInVicinity = 0xFF;
				break;
			}
		}

		// log weigand data for all passers by if required
		if (*pcWeigandStyle_4D==WEIG_50_ALLDETECTS || *pcWeigandStyle_4D==WEIG_26_ALLDETECTS) WiegandEvent( ((const unsigned char*)pData1) );

		return;

	}

	// inquiry completed
	if (EventID==BMTE_OK && (AccConSemaphores & ACSF_INQUIRYINPROCESS))
	{
		unsigned char i, j, bFoundTrustedThisCycle;

		// end of enquiry
		AccConSemaphores &= ~ACSF_INQUIRYINPROCESS;

		// store collated data in Vicinity table
		memcpy( pVicinityTable, (const void *) pInquiryTable, nInquiry * sizeof( UserEntry2 ) );
		nVicinity = nInquiry;

		// if logging all phones, or access triggered, log now
		if (*pcLogStyle_18 == LL_ALLDETECTS || 	(*pcLogStyle_18 == LL_ONACCESS && (AccConSemaphores & ACSF_RUNACCESSLOG) ) )
		{
		  unsigned char bIsRegistered = 0;

		  AccConSemaphores &= ~ACSF_RUNACCESSLOG;

		  for (j=0; j<nVicinity; j++)
		  {
			for (i = 0; i< nUsers && !bIsRegistered; i++)
			{
				// write to log
				if (memcmpram2pgm( pUserList[i].BluetoothID, (const void *) pVicinityTable[j].BluetoothID, BTADDRLEN) == 0 )
				{
					LogUser( pUserList[i].UserName, 0 );
					bIsRegistered = 0xFF;

					// output to weigand data
					if (*pcWeigandStyle_4D==WEIG_50_GRANTED || *pcWeigandStyle_4D==WEIG_26_GRANTED) 
					{
						WiegandEvent( pUserList[i].BluetoothID );
						msDelay(10); // 10ms extra delay
					}

				}
			}
			if (!bIsRegistered && *pcLogStyle_18 == LL_ALLDETECTS) LogUser( 0, pVicinityTable[j].UserName );
		  }
		}

		// test for any devices in vicinity
		bFoundTrustedThisCycle = 0;
		for (i = 0; i< nUsers && !bFoundTrustedThisCycle; i++)
		{
		  for (j=0; j<nVicinity; j++)
		  {
			if (memcmpram2pgm( pUserList[i].BluetoothID, (const void *) pVicinityTable[j].BluetoothID, BTADDRLEN) == 0 
					&& TimeConditionSatisfied( i ))
			{
				bFoundTrustedThisCycle = 0xFF;
				break;
			}
		  }
		}

		// if none found, clear trusted device flag and sleep mode
		if (!bFoundTrustedThisCycle)
		{
			bTrustedDeviceInVicinity = 0;
			bSleepModeRequested = 0;
			IndicatorLED = 0;

			// turn off relay if in on while present mode
			if (*pcTriggerList_16 == TRIG_WHILEPRESENT)  RelayOff;
		}

		return;

	}

	// connect in new user mode is not allowed - this mode is for pairing only, so reset
	if ((AccConSemaphores&ACSF_NEWUSERMODE) && EventID==BMTE_Connect)
	{
		AccConSemaphores = 0;
		Reset();
	}

	// if paired in new user mode, add new user
	if ((AccConSemaphores&ACSF_NEWUSERMODE) && EventID==BMTE_Paired)
	{
		unsigned char i;

		// set mod user to new entry, six month pass
		memcpy( &ModUser.ExpiryDT, (const void *) &RealTimeClock, sizeof( DateTimeU ) );
		ModUser.ExpiryDT.month += 6;
		if (ModUser.ExpiryDT.month>12)
		{
			ModUser.ExpiryDT.month -= 6;
			ModUser.ExpiryDT.year++;
		}
		memcpy( ModUser.BluetoothID, (const void *) pData1, BTADDRLEN );
		ModUser.FromHr = 0;
		ModUser.FromMn = 0;
		ModUser.FromSc = 0;
		ModUser.ToHr = 0;
		ModUser.ToMn = 0;
		ModUser.ToSc = 0;
		ModUser.DayFlags = DF_MONDAY | DF_TUESDAY | DF_WEDNESDAY | DF_THURSDAY | DF_FRIDAY | DF_SATURDAY | DF_SUNDAY;

		// remote device doesn't have to give its name at this stage, but with lick it's in the vicinity table
		// set name to "anonymous"
		memcpypgm2ram( ModUser.UserName, (rom const void *) pAnonymous, USERNAMELEN );
		ModUser.UserName[5] = GetUpperHexChar( ModUser.BluetoothID[0] );
		ModUser.UserName[6] = GetLowerHexChar( ModUser.BluetoothID[0] );
		ModUser.UserName[7] = GetUpperHexChar( ModUser.BluetoothID[1] );
		ModUser.UserName[8] = GetLowerHexChar( ModUser.BluetoothID[1] );
		ModUser.UserName[9] = GetUpperHexChar( ModUser.BluetoothID[2] );
		ModUser.UserName[10] = GetLowerHexChar( ModUser.BluetoothID[2] );
		ModUser.UserName[11] = GetUpperHexChar( ModUser.BluetoothID[3] );
		ModUser.UserName[12] = GetLowerHexChar( ModUser.BluetoothID[3] );
		ModUser.UserName[13] = GetUpperHexChar( ModUser.BluetoothID[4] );
		ModUser.UserName[14] = GetLowerHexChar( ModUser.BluetoothID[4] );
		ModUser.UserName[15] = GetUpperHexChar( ModUser.BluetoothID[5] );
		ModUser.UserName[16] = GetLowerHexChar( ModUser.BluetoothID[5] );

		// search for device in vicinity table - if it's there, use than name instead
		for (i=0; i<nVicinity; i++)
		{
			if (memcmp( (const void *) ModUser.BluetoothID, (const void *) pVicinityTable[i].BluetoothID, BTADDRLEN) == 0 )
			{
				memcpy( ModUser.UserName, (const void *) pVicinityTable[i].UserName, USERNAMELEN );
				break;
			}
		}

		// insert in user table
		ModifyUser( 0xFF );

		// reset to return to known state and old pin code
		AccConSemaphores = 0;
		msDelay( 250 );
		msDelay( 250 );
		msDelay( 250 );
		msDelay( 250 );
		msDelay( 250 );
		msDelay( 250 );
		msDelay( 250 );
		msDelay( 250 );
		Reset();
	}

	 
}

// User interface events (Operating System callback)
void FxPEvent( unsigned char EventID, void *pData )
{
	// remote device connected
	if ( EventID==FxPE_Connect ) 
	{
		// log state as in setup mode and remote device connected;
		AccConSemaphores &= ~ACSF_SETUPNOTCONNECTED ;

		// LED permanently on during connection
		IndicatorLED = 1;
		return;
	}

	if ( EventID==FxPE_Disco ) 
	{
		// Settings may have changed, so simplest thing is to Reset; wait for 1s to ensure remote radio has disconnected 
		msDelay(250);
		msDelay(250);
		msDelay(250);
		msDelay(250);
		AccConSemaphores = 0;
		Reset();
	}

	// respond to user actions
	if ( EventID==FxPE_ClntUpdate ) 
	{
		// add button pressed
		if (*((unsigned short*) pData) == ID_Add____3)
		{
			if (nUsers>=MAXUSERS)
			{
				// display message than no new users can be added
				ShowMessage_AddNoMore_3F;
			}
			else if (nVicinity == 0)
			{
				// display message than no new users were found
				ShowMessage_NoNewUsers_4A;
			}
			else
			{
				unsigned char i;
				unsigned long lVal;
				ADD pListItems = ListTextLoc_NewUserList_8;
	
				// set up new user list from vicinity table
				for (i = 0; i<nVicinity; i++)
				{
					// note we are am deliberately not removing users who are already in the list because multiple users may need multiple timed entries
	
					// get string length
					unsigned char iLen;
					for (iLen = 0; iLen < pVicinityTable[i].UserName[iLen]!=0; iLen++) {;}
					iLen++;
	
					// add string to list box
					SetBytes( ListTextStr_NewUserList_8, pListItems, 0, pVicinityTable[i].UserName, iLen );
					pListItems += iLen;

				}
				lVal = nVicinity;
				SetListNum_NewUserList_8( 0, &lVal );
	
				// set selected item to zero
				*pcNewUserList_8 = 0;
				
				// show add users dialog
				FxPCommand( FxPC_SetDialog, DlgID_AddUsersText_3B, 0 );		
			}
			return;
		}

		// modify button pressed
		if (*((unsigned short*) pData) == ID_Modify____11)
		{
			// check there is at least one user since otherwise may try to modify the "(no users)" entry
			if (nUsers)
			{
				unsigned char iSelItem = *pcRegUserList_2;
	
				// set mod user to table entry 
				memcpypgm2ram( &ModUser, (const rom void *) &pUserList[iSelItem], sizeof( ModUser ) );
				ModUserIndex = iSelItem;  
	
				// show modify users list
				ShowModUserList();
			}
		}

		// delete button pressed
		if (*((unsigned short*) pData) == ID_Delete_4)
		{
			// check there is at least one user since otherwise may try to delete the "(no users)" entry
			if (nUsers)
			{
				// set text in confirmation dialog
				unsigned char iSelItem = *pcRegUserList_2;
				memcpypgm2ram( pUserName_47, (const rom void *) &pUserList[iSelItem].UserName, MaxBytes_UserName_47 );

				// show confirmation dialog
				FxPCommand( FxPC_SetDialog, DlgID_WarnText_46, 0 );		
			}
			return;
		}

		// confirm button pressed, confirmation dialog
		if (*((unsigned short*) pData) == ID_Confirm_48)
		{
			unsigned char i;
			unsigned long lVal;

			// ModUserIndex now points to insertion point, move remaining items up the table
			for (i = *pcRegUserList_2; i<nUsers-1; i++)
			{
				unsigned char pTempData[ sizeof( UserEntry10 )];
				GetBytes( USERTABLESTR, USERTABLELOC + sizeof(UserEntry10) * (i+1), pTempData, sizeof(UserEntry10) );
				SetBytes( USERTABLESTR, USERTABLELOC + sizeof(UserEntry10) * i, 0, pTempData, sizeof(UserEntry10) );
			}

			// decrease nUsers
			lVal = nUsers - 1;
			SetBytes( NUMUSERSTR, NUMUSERLOC, 0, &lVal, sizeof(long) );

			// update registered users list box
			UpdateRegUserList();

			// show registered users dialog
			FxPCommand( FxPC_SetDialog, DlgID_RegUsersText_1, 0 );		
			return;
		}

		// cancel button, confirmation dialog
		if (*((unsigned short*) pData) == ID_Cancel_49)  
		{
			// return to registered users list
			FxPCommand( FxPC_SetDialog, DlgID_RegUsersText_1, 0 );		
			return;
		}

		// access log button pressed
		if (*((unsigned short*) pData) == ID_Access_Log____3A)
		{
			// show access log dialog
			FxPCommand( FxPC_SetDialog, DlgID_AccessLogText_C, 0 );		
			return;
		}

		// settings button pressed
		if (*((unsigned short*) pData) == ID_Setup____6)
		{
			// show settings dialog
			FxPCommand( FxPC_SetDialog, DlgID_Settings_12, 0 );		
			return;
		}

		// close button, log dialog
		if (*((unsigned short*) pData) == ID_Close_3C)  // log dialog
		{
			// return to registered users list
			FxPCommand( FxPC_SetDialog, DlgID_RegUsersText_1, 0 );		
			return;
		}

		// cancel button, add user dialog
		if (*((unsigned short*) pData) == ID_Cancel_A)  // add user dialog
		{
			// return to regustered users list
			FxPCommand( FxPC_SetDialog, DlgID_RegUsersText_1, 0 );		
			return;
		}

		// select button, add user dialog
		if (*((unsigned short*) pData) == ID_Select_9)  // add user dialog
		{
			char iSelItem = *pcNewUserList_8;

			// set mod user to new entry 
			memcpy( &ModUser.ExpiryDT, (const void *) &RealTimeClock, sizeof( DateTimeU ) );
			ModUser.ExpiryDT.year += 50;
			memcpy( ModUser.BluetoothID, (const void *) pVicinityTable[iSelItem].BluetoothID, BTADDRLEN );
			memcpy( ModUser.UserName, (const void *) pVicinityTable[iSelItem].UserName, USERNAMELEN );
			ModUser.FromHr = 0;
			ModUser.FromMn = 0;
			ModUser.FromSc = 0;
			ModUser.ToHr = 0;
			ModUser.ToMn = 0;
			ModUser.ToSc = 0;
			ModUser.DayFlags = DF_MONDAY | DF_TUESDAY | DF_WEDNESDAY | DF_THURSDAY | DF_FRIDAY | DF_SATURDAY | DF_SUNDAY;
			ModUserIndex = 0xFF;  // index 0xFF indicates that this is an added user not an existing user in the list

			// move on to modify users list
			ShowModUserList();
			return;
		}

		// username changed
		if (*((unsigned short*) pData) == ID_UserName_E)  // log dialog
		{
			memcpy( ModUser.UserName, (const void *) pUserName_E, USERNAMELEN );
			return;
		}

		// expiry settings button pressed
		if (*((unsigned short*) pData) == ID_Expiry_Settings____26)
		{
			// copy current expiry settings
			memcpy( pDeletionDate_37, (const void *) &ModUser.ExpiryDT, sizeof( DateTimeU ) );

			// show expiry settings dialog
			FxPCommand( FxPC_SetDialog, DlgID_ExpiryText_27, 0 );		
			return;
		}

		// Close button, expiry dialog
		if (*((unsigned short*) pData) == ID_Close_38)  // expiry dialog
		{
			// copy new expiry settings
			memcpy( &ModUser.ExpiryDT, (const void *) pDeletionDate_37, sizeof( DateTimeU ) );

			// return to registered users list
			FxPCommand( FxPC_SetDialog, DlgID_SetUsersText_10, 0 );		
			return;
		}

		// timed entry settings button pressed
		if (*((unsigned short*) pData) == ID_Timed_Entry_Settings____34)
		{
			// copy current timed entry settings
			pFromTime_29->hour = ModUser.FromHr;
			pFromTime_29->min = ModUser.FromMn;
			pFromTime_29->sec = ModUser.FromSc;
			pToTime_2B->hour = ModUser.ToHr;
			pToTime_2B->min = ModUser.ToMn;
			pToTime_2B->sec = ModUser.ToSc;
			*pMonday_2C = (ModUser.DayFlags & DF_MONDAY) ? 0xFF : 0x00;
			*pTuesday_2D = (ModUser.DayFlags & DF_TUESDAY) ? 0xFF : 0x00;
			*pWednesday_2E = (ModUser.DayFlags & DF_WEDNESDAY) ? 0xFF : 0x00;
			*pThursday_2F = (ModUser.DayFlags & DF_THURSDAY) ? 0xFF : 0x00;
			*pFriday_30 = (ModUser.DayFlags & DF_FRIDAY) ? 0xFF : 0x00;
			*pSaturday_31 = (ModUser.DayFlags & DF_SATURDAY) ? 0xFF : 0x00;
			*pSunday_32 = (ModUser.DayFlags & DF_SUNDAY) ? 0xFF : 0x00;

			// show timed entry settings dialog
			FxPCommand( FxPC_SetDialog, DlgID_TimedEntryText_35, 0 );		
			return;
		}

		// Close button, timed entry dialog
		if (*((unsigned short*) pData) == ID_Close_33)  // timed entry date dialog
		{
			// copy new timed entry settings
			ModUser.FromHr = pFromTime_29->hour ;
			ModUser.FromMn = pFromTime_29->min ;
			ModUser.FromSc = pFromTime_29->sec ;
			ModUser.ToHr = pToTime_2B->hour ;
			ModUser.ToMn = pToTime_2B->min ;
			ModUser.ToSc = pToTime_2B->sec ;
			ModUser.DayFlags =  ( *pMonday_2C ? DF_MONDAY : 0 ) | ( *pTuesday_2D ? DF_TUESDAY : 0 ) | ( *pWednesday_2E ? DF_WEDNESDAY : 0 ) | ( *pThursday_2F ? DF_THURSDAY : 0 ) |
								 ( *pFriday_30 ? DF_FRIDAY : 0 ) | ( *pSaturday_31 ? DF_SATURDAY : 0 ) | ( *pSunday_32 ? DF_SUNDAY : 0 ) ;

			// return to modify users list
			FxPCommand( FxPC_SetDialog, DlgID_SetUsersText_10, 0 );		
			return;
		}

		// OK button, modify user dialog
		if (*((unsigned short*) pData) == ID_OK_B)  // modify user dialog
		{
			ModifyUser( ModUserIndex );

			// return to registered users list
			FxPCommand( FxPC_SetDialog, DlgID_RegUsersText_1, 0 );		
			return;
		}

		// Cancel button, modify user dialog
		if (*((unsigned short*) pData) == ID_Cancel_F)  // modify user dialog
		{
			// return to registered users list
			FxPCommand( FxPC_SetDialog, DlgID_RegUsersText_1, 0 );		
			return;
		}

		// OK button, settings dialog
		if (*((unsigned short*) pData) == ID_Close_21)  // settings dialog
		{
			// return to registered users list
			FxPCommand( FxPC_SetDialog, DlgID_RegUsersText_1, 0 );		
			return;
		}

		// Close button, user list dialog
		if (*((unsigned short*) pData) == ID_Close_7)  // main screen
		{
			// disconnect from remote device; give time for disconnection message to be sent then reset
			FxPCommand( FxPC_Disco, 0, 0 );		
			msDelay(250);
			msDelay(250);
			msDelay(250);
			msDelay(250);
			AccConSemaphores = 0;
			Reset();
		}

		// clock modified
		if (*((unsigned short*) pData) == ID_TimeNow_23 )
		{
			// calculate day of week based on date entered
			CalcDayOfWeek( RealTimeClock.dow, RealTimeClock.year, RealTimeClock.month, RealTimeClock.date );
			return;
		}	

		// DST check box modified
		if (*((unsigned short*) pData) == ID_DTSStyle_24 || *((unsigned short*) pData) == ID_In_DST_now_4E )
		{
			// update daylight savings time rules
			DSTEvent = *pcDTSStyle_24;
			if (*pIn_DST_now_4E) DSTEvent |= DST_SUMMER_FLAG;
			return;
		}	
	}

	if (EventID==FxPE_PingFail)
	{
		// remote device went out of range.  Reset immediately
		AccConSemaphores = 0;
		Reset();
	}

}


// GetLowerHexChar returns least significant character byte value in ASCII hexadecimal
unsigned char GetLowerHexChar( unsigned char iCh )
{
	iCh = iCh & 0x0F;
	if (iCh > 0x09) 
		return iCh - 0x0A + 'A';
	else
		return iCh + '0';
}

// GetLowerHexChar returns most significant character byte value in ASCII hexadecimal
unsigned char GetUpperHexChar( unsigned char iCh )
{
	iCh = iCh >> 4;
	return GetLowerHexChar( iCh );
}

void ShowModUserList( void )
{
	// fill controls with new data

	// user name
	memcpy( pUserName_E, (const void *) ModUser.UserName, USERNAMELEN );

	// bluetooth ID
	pBTID_40[15] = GetUpperHexChar( ModUser.BluetoothID[0] );
	pBTID_40[16] = GetLowerHexChar( ModUser.BluetoothID[0] );
	pBTID_40[18] = GetUpperHexChar( ModUser.BluetoothID[1] );
	pBTID_40[19] = GetLowerHexChar( ModUser.BluetoothID[1] );
	pBTID_40[21] = GetUpperHexChar( ModUser.BluetoothID[2] );
	pBTID_40[22] = GetLowerHexChar( ModUser.BluetoothID[2] );
	pBTID_40[24] = GetUpperHexChar( ModUser.BluetoothID[3] );
	pBTID_40[25] = GetLowerHexChar( ModUser.BluetoothID[3] );
	pBTID_40[27] = GetUpperHexChar( ModUser.BluetoothID[4] );
	pBTID_40[28] = GetLowerHexChar( ModUser.BluetoothID[4] );
	pBTID_40[30] = GetUpperHexChar( ModUser.BluetoothID[5] );
	pBTID_40[31] = GetLowerHexChar( ModUser.BluetoothID[5] );
	
	FxPCommand( FxPC_SetDialog, DlgID_SetUsersText_10, 0 );		
}

void UpdateRegUserList( void )
{
	if (nUsers==0)
	{
		// "(No users)" message
		unsigned long lVal = 1;
		SetListNum_RegUserList_2( 0, &lVal );
		SetBytes( ListTextStr_RegUserList_2, ListTextLoc_RegUserList_2, pNoUsers, 0, NOUSERSLEN );
	}
	else
	{
		unsigned char i;
		unsigned long lVal;
		ADD pListItems = ListTextLoc_RegUserList_2;

		// set up user list from user list
		for (i = 0; i<nUsers; i++)
		{
			// get string length
			unsigned char iLen;
			for (iLen = 0; iLen < pUserList[i].UserName[iLen]!=0; iLen++) {;}
			iLen++;

			// add string to list box
			SetBytes( ListTextStr_RegUserList_2, pListItems, pUserList[i].UserName, 0, iLen );
			pListItems += iLen;
		}
		lVal = nUsers;
		SetListNum_RegUserList_2( 0, &lVal );

		// set selected item to zero
		*pcNewUserList_8 = 0;
		
	}

}


// check the time conditions for user i
unsigned char TimeConditionSatisfied( unsigned char i )
{
	// check OK by timed entry rules; cast sec / min / hour triplets as unsigned short longs
	unsigned long short FromTime = *((rom unsigned long short *) &pUserList[i].FromSc);	
	unsigned long short ToTime = *((rom unsigned long short *) &pUserList[i].ToSc);	
	unsigned long short TimeNow = 	*((unsigned long short *) &RealTimeClock.sec);

	// from and to times do not span midnight
	if ( FromTime <= ToTime)
	{
		// test weekday condition
		if ( RealTimeClock.dow == 0 && ( pUserList[i].DayFlags & DF_SUNDAY )==0 ) return 0;
		if ( RealTimeClock.dow == 1 && ( pUserList[i].DayFlags & DF_MONDAY )==0 ) return 0;
		if ( RealTimeClock.dow == 2 && ( pUserList[i].DayFlags & DF_TUESDAY )==0 ) return 0;
		if ( RealTimeClock.dow == 3 && ( pUserList[i].DayFlags & DF_WEDNESDAY )==0 ) return 0;
		if ( RealTimeClock.dow == 4 && ( pUserList[i].DayFlags & DF_THURSDAY )==0 ) return 0;
		if ( RealTimeClock.dow == 5 && ( pUserList[i].DayFlags & DF_FRIDAY )==0 ) return 0;
		if ( RealTimeClock.dow == 6 && ( pUserList[i].DayFlags & DF_SATURDAY )==0 ) return 0;

		// if from and two times equal, don't test time condition,
		if ( ToTime != FromTime && ( TimeNow < FromTime || TimeNow > ToTime ) ) return 0;
	}
	else
	{
		// from and to times span midnight

		// if it's after midnightthe day of week condition should actually be for previous day
		unsigned char EffectiveDow = RealTimeClock.dow;
		if ( TimeNow < FromTime ) EffectiveDow = (EffectiveDow==0) ? 6 : EffectiveDow - 1;

		// test weekday condition
		if ( EffectiveDow == 0 && ( pUserList[i].DayFlags & DF_SUNDAY )==0 ) return 0;
		if ( EffectiveDow == 1 && ( pUserList[i].DayFlags & DF_MONDAY )==0 ) return 0;
		if ( EffectiveDow == 2 && ( pUserList[i].DayFlags & DF_TUESDAY )==0 ) return 0;
		if ( EffectiveDow == 3 && ( pUserList[i].DayFlags & DF_WEDNESDAY )==0 ) return 0;
		if ( EffectiveDow == 4 && ( pUserList[i].DayFlags & DF_THURSDAY )==0 ) return 0;
		if ( EffectiveDow == 5 && ( pUserList[i].DayFlags & DF_FRIDAY )==0 ) return 0;
		if ( EffectiveDow == 6 && ( pUserList[i].DayFlags & DF_SATURDAY )==0 ) return 0;

		// test time condition
		if ( TimeNow < FromTime || TimeNow > ToTime ) return 0;
	}

	return 0xFF;
}


void TriggerRelay( void )
{
	unsigned char i, j;

	RelayOn;
	msDelay( 250 );
	RelayOff;

	// if logging accesses granted, log all users in vicinity
	AccConSemaphores |= ACSF_RUNACCESSLOG;
}


#define LOGLISTMAX 100
#define LOGLISTENTRYLEN 37
#if (ItemTextMaxBytes_Access_Log_3E != LOGLISTENTRYLEN * LOGLISTMAX)
#error Access_Log_3E incorrectly sized
#endif

void LogUser( rom unsigned char * pRUserName, unsigned char * pUserName )
{
	unsigned char LogEntry[LOGLISTENTRYLEN];
	DateTimeU ClockTmp;
	unsigned char i;
	unsigned long iNextLoc;
	unsigned long iListCount;

	// get default string
	memcpypgm2ram( LogEntry, (rom void *) pLogEntryTemplate, LOGLISTENTRYLEN-1 );
	LogEntry[LOGLISTENTRYLEN-1] = 0;

	// get time
	DisableHighInts;
	memcpy( &ClockTmp, (const void *) &RealTimeClock, sizeof(RealTimeClock) );
	EnableHighInts;

	// time
	LogEntry[0] = ClockTmp.hour/10 + '0';
	LogEntry[1] = ClockTmp.hour%10 + '0';
	LogEntry[3] = ClockTmp.min/10 + '0';
	LogEntry[4] = ClockTmp.min%10 + '0';
	LogEntry[6] = ClockTmp.sec/10 + '0';
	LogEntry[7] = ClockTmp.sec%10 + '0';

	// date - european style
	LogEntry[9] = ClockTmp.date/10 + '0';
	LogEntry[10] = ClockTmp.date%10 + '0';
	LogEntry[12] = ClockTmp.month/10 + '0';
	LogEntry[13] = ClockTmp.month%10 + '0';
	LogEntry[15] = (ClockTmp.year/10)%10 + '0';
	LogEntry[16] = ClockTmp.year%10 + '0';

	// name without zero terminator
	for (i = 0; pRUserName ? pRUserName[i] : pUserName[i]; i++)
	{
		LogEntry[18 + i] = pRUserName ? pRUserName[i] : pUserName[i];
	}

	// write to list box
	SetBytes( ListTextStr_Access_Log_3E, ListTextLoc_Access_Log_3E + LOGLISTENTRYLEN * NextLogLoc, 0, LogEntry, LOGLISTENTRYLEN );

	// compute next append location
	iNextLoc = NextLogLoc + 1;
	if (iNextLoc >= LOGLISTMAX) iNextLoc = 0;
	SetBytes( NEXTLOGSTR, NEXTLOGLOC, 0, &iNextLoc, sizeof( long ) );

	// update list item count
	GetBytes( ListNumStr_Access_Log_3E, ListNumLoc_Access_Log_3E, &iListCount, sizeof( long ) );
	if (iListCount < LOGLISTMAX ) 
	{
		iListCount += 1;
		SetBytes( ListNumStr_Access_Log_3E, ListNumLoc_Access_Log_3E, 0, &iListCount, sizeof( long ) );
	}
	
}

// ModifyUser replaces user iUser with ModUser; if iUser is 0xFF, inserts in alphabetical orer
void ModifyUser( int iUser )
{
	if (iUser==0xFF)
	{
		unsigned char i;
		unsigned long lVal;

		// find the insertion point in the list in alphabetical order
		for (iUser = 0; iUser<nUsers; iUser++)
		{
			unsigned char chNew, chTable, bIsBefore, bIsAfter;
			bIsBefore = False;
			bIsAfter = False;
			for (i=0; i<USERNAMELEN && !bIsBefore && !bIsAfter; i++)
			{
				chNew = ModUser.UserName[i];
				if (chNew >= 'A' && chNew <= 'Z')  chNew |= 0x20;  // convert to lower case
				chTable = pUserList[iUser].UserName[i] ;
				if (chTable >= 'A' && chTable <= 'Z')  chTable |= 0x20;  // convert to lower case
				if (chTable > chNew)  bIsBefore = True;
				if (chTable < chNew)  bIsAfter = True;
			}
			if (bIsBefore) break;
		}

		// ModUserIndex now points to insertion point, move existing item and all after it down in the table
		for (i=nUsers; i>iUser; i--)
		{
			unsigned char pTempData[ sizeof( UserEntry10 )];
			GetBytes( USERTABLESTR, USERTABLELOC + sizeof(UserEntry10) * (i-1), pTempData, sizeof(UserEntry10) );
			SetBytes( USERTABLESTR, USERTABLELOC + sizeof(UserEntry10) * i, 0, pTempData, sizeof(UserEntry10) );
		}

		// increase nUsers
		lVal = nUsers + 1;
		SetBytes( NUMUSERSTR, NUMUSERLOC, 0, &lVal, sizeof(long) );
	}

	// overwrite existing table entry
	SetBytes( USERTABLESTR, USERTABLELOC + sizeof(UserEntry10) * iUser, 0, &ModUser, sizeof(UserEntry10) );

	// update registered users list box
	UpdateRegUserList();
}



void WiegandEvent(const unsigned char * pBTAddr)
{
	unsigned char iBit;
	unsigned char iByte;
	unsigned char Parity1;
	unsigned char Parity2;
	unsigned char CoreData[6];
	unsigned char W26bit;

	// get format
	W26bit = (*pcWeigandStyle_4D==WEIG_26_GRANTED || *pcWeigandStyle_4D==WEIG_26_REQUESTED || *pcWeigandStyle_4D==WEIG_26_ALLDETECTS) ? 0xFF:0;
	// get core data
	if ( W26bit!=0 )
	{
		// copy Core Data from pRecentBTAddr, exclusive oring 
		CoreData[0] = pBTAddr[0] ^ pBTAddr[3];
		CoreData[1] = pBTAddr[1] ^ pBTAddr[4];
		CoreData[2] = pBTAddr[2] ^ pBTAddr[5];
	}
	else
	{
		// copy Core Data from pRecentBTAddr
		CoreData[0] = pBTAddr[0];
		CoreData[1] = pBTAddr[1];
		CoreData[2] = pBTAddr[2];
		CoreData[3] = pBTAddr[3];
		CoreData[4] = pBTAddr[4];
		CoreData[5] = pBTAddr[5];
	}

	// calculate parity1
	Parity1 = 0;
	for (iByte = 0; iByte < 3; iByte++)
	{
		for (iBit = 0; iBit < 8; iBit++)
		{
			if ( W26bit!=0 && iByte==1 && iBit==4)  break;
			if ( (CoreData[iByte] >> iBit ) & 0x01 ) Parity1 = 1 - Parity1;
		}
		if ( W26bit!=0 && iByte==1 && iBit==4)  break;
	}

	// calculate parity2
	Parity2 = 1;
	for (iByte = (W26bit!=0 ? 1 : 3); iByte < (W26bit ? 3 : 6); iByte++)
	{
		for (iBit = (iByte==1? 4 : 0); iBit < 8; iBit++)
		{
			if ( (CoreData[iByte] >> iBit ) & 0x01 ) Parity2 =  1 - Parity2;
		}
	}

	// write data out
	WiegandBit( Parity1 );
	for (iByte = 0; iByte < (W26bit ? 3 : 6); iByte++)
	{
		for (iBit = 0; iBit < 8; iBit++)
		{
			WiegandBit( (CoreData[iByte] >> iBit ) & 0x01 );
		}
	}
	WiegandBit( Parity2 );

}

void WiegandBit( unsigned char iBit )
{
	if (iBit!=0)
	{
		Data1Pin = WEIG_0V;
		CyclesDelay3p2plus3p2times( 15 ); // approx 50us		
		Data1Pin = WEIG_5V;
	}
	else
	{
		Data0Pin = WEIG_0V;
		CyclesDelay3p2plus3p2times( 15 ); // approx 50us		
		Data0Pin = WEIG_5V;
	}

	msDelay(1); //1ms	

}

