//------------------------------------------------------------------------------------------------------
//
//  Slave.c  
//
//  Application for customizing ToothPIC from an external host processor
//
//  (c) FlexiPanel Ltd
//
//-------------------------------------------------------------------------------------------------------

#include "..\ToothPIC.h"
#include <p18f6720.h>  

// header file for control information flags
#include "..\hopcodes.h"

// Serial port definitions
#define SlvB_1220		0x01  
#define SlvB_2400		0x02  
#define SlvB_4800		0x03  
#define SlvB_9600		0x04  /* default */
#define SlvB_19200		0x05
#define SlvB_38400		0x06  
#define SlvB_57600		0x07  
#define SlvB_115200		0x08  
#define SlvS_None		0x00  /* default */
#define SlvS_Auth		0x01  
#define SlvS_AuthEncr	0x02  
#define SlvF_None		0x00  /* default */
#define SlvF_RTSisINT0	0x01  
#define SlvF_RTSisSDO	0x02  
#define SlvF_RTSisCCP1	0x03  
#define SlvF_RTSisAN11	0x04  
#define SlvF_RTSisSCL	0x05    
#define SlvF_NoRTS		0x10  
#define SlvE_OnErrFlash	0x01  
#define SlvE_OnErrReset	0x02  
#define SlvE_OnErrResp	0x03  
#define SlvI_None		0x00
#define SlvI_I2C100kHz	0x01
#define SlvI_I2C400kHz	0x02

#define SHORTMESSLEN	0x16 					/* max short message length as stored in queue */
#define LONGMESSLEN		0x30 					/* max long message length short message as ascii hex preceeded and followed by <CR><LF> */
#define MESSAGE_TOP		0x900

#define CTSPin	INT1Pin
#define DirCTS	DirINT1
#define CTSStop		1			  			
#define CTSClear	0			  			
#define RTSStop		1			  			
#define RTSClear	0
#define CTSIntFlag		INTCON3bits.INT1IF

#define DirData DirINT0
#define DataPin INT0Pin

// data in ToothPIC.c
extern rom unsigned short int DeviceClass; 
extern rom unsigned char pLocalName[];    
extern rom char pszPIN[];

// receive and transmit buffers
unsigned char RxBuff[ LONGMESSLEN ];
unsigned char RxBuffLen;		
unsigned char TxBuff[ LONGMESSLEN ];
unsigned char TxBuffLen;		
unsigned char TxBuffPos;	

// reserve RAM used by message queue and FlexiPanel user interface to ensure linker doesn't allocate it to anything else
// this is only necessary because the user interface is not included at compile time
// if you customize the application, remove the lines between the following 17 lines of code
// and program the User Interface by selecting ToothPIC MPLAB as the Designer target and include the
// files generated in the project in place of SlaveDefaultRes.c and SlaveDefaultRes.h
// but remember that the linker may allocate rom immediately after ram used by FlexiPanel; you
// must make sure this does not overlap your message queue
#pragma udata TP_UIRAM000			// located at 0x100 - 0x1FF
unsigned char pFxPRAM000B[0xFF];  	// one byte allocated in SlaveDefaultRes.c
#pragma udata TP_UIRAM100			// located at 0x200 - 0x1FF
unsigned char pFxPRAM100[0x100];
#pragma udata TP_UIRAM200			// located at 0x300 - 0x1FF
unsigned char pFxPRAM200[0x100];
#pragma udata TP_UIRAM300			// located at 0x400 - 0x8FF
unsigned char pFxPRAM300[0x100];
#pragma udata TP_UIRAM400			// located at 0x500 - 0x8FF
unsigned char pFxPRAM400[0x100];
#pragma udata TP_UIRAM500			// located at 0x600 - 0x8FF
unsigned char pFxPRAM500[0x100];
#pragma udata TP_UIRAM600			// located at 0x700 - 0x8FF
unsigned char pFxPRAM600[0x100];
#pragma udata TP_UIRAM700			// located at 0x800 - 0x8FF
unsigned char pFxPRAM700[0x100];
#pragma udata

// reserve rom used by Flexipanel user interface - maximum availavle is 0x00E000 bytes
// this is only necessary because the user interface is not included at compile time; DARC-II doesn't need to 
// because it doesn't require so much rom memory
// for you customize the application and you need more ROM memory, reduce the size of this reserved space but
// leave a contiguous block from 0x0101D7 upwards big enough for your User Interface (if any)
#pragma romdata Section_010000 = 0x0101D7
rom unsigned char p010000[0x3E29]; // first 0x052B Bytes in SlaveDefaultRes.h
#pragma romdata Section_014000 = 0x014000
rom unsigned char p014000[0x4000];
#pragma romdata Section_018000 = 0x018000
rom unsigned char p018000[0x4000];
#pragma romdata Section_01C000 = 0x01C000
rom unsigned char p01C000[0x2000];
#pragma romdata 


// Parallel, analog and CCP I/O functions
unsigned char NumAnalog;
unsigned char ParallelA;
unsigned char ParallelB;
unsigned char CCPFunc;
#define CCP1Func	0x01
#define CCP2Func	0x02
#define CCP3Func	0x04
#define CCP4Func	0x08
#define CCP5Func	0x10

// message queue
unsigned char *TxMessageQueue;  // message queue will be located immediately below 0xA00

// permanent settings stored in EE rom; Alternate_Defaults is used by FlexiPanel for customizations
#ifndef Alternate_Defaults
#pragma romdata EEBaudLoc=0xF003E0
rom unsigned char EESlv_Security = SlvS_None;				// Security level, see SlvS_ defines
rom unsigned char EESlv_Baud = SlvB_9600;					// Baud rate, see SlvB_ defines
rom unsigned char EESlv_FlowControl = SlvF_None;			// Flow control, see SlvF_ defines
rom unsigned char EESlv_HostHasRxBuffer = True;				// Host has receive buffer and can cope with extra byte being sent after RTS goes high
rom unsigned char EESlv_MessQueueLen = 0x20;				// Number of messages that can be stored in the queue; sum of this times MESSLEN 
															// plus FlexiPanel RAM requirement not to exceed 0x800
															// 2 is lowest allowable value 
rom unsigned char EESlv_InitMsg = True;						// Send initialization response
rom unsigned char EESlv_MsgsAnytime = True;					// By default, can send reply messages at any time subject to flow control; if false, waits for send message command
rom unsigned char EESlv_ASCIIMessages = True;				// reply messages are in ascii		
rom unsigned char EESlv_OnInternaError = SlvE_OnErrResp;	// what to do if an internal error occurs
rom unsigned char EESlv_I2CSetup = SlvI_None;				// I2C memory type
rom unsigned char EESlv_Int0IsData = 0x00;					// Int0 is data pin
rom unsigned char EESlv_Reserved0B = 0xFF;					// Reserved
rom unsigned char EESlv_Reserved0C = 0xFF;					// Reserved
rom unsigned char EESlv_Reserved0D = 0xFF;					// Reserved
rom unsigned char EESlv_Reserved0E = 0xFF;					// Reserved
rom unsigned char EESlv_Reserved0F = 0xFF;					// Reserved
rom unsigned char EESlv_Reserved1F[0x10] ;					// Reserved
#pragma romdata 
#else
extern rom unsigned char EESlv_Security ;
extern rom unsigned char EESlv_Baud ;
extern rom unsigned char EESlv_FlowControl ;
extern rom unsigned char EESlv_HostHasRxBuffer ;
extern rom unsigned char EESlv_MessQueueLen ;
extern rom unsigned char EESlv_InitMsg ;
extern rom unsigned char EESlv_MsgsAnytime ;
extern rom unsigned char EESlv_ASCIIMessages ;	
extern rom unsigned char EESlv_OnInternaError ;
extern rom unsigned char EESlv_I2CSetup ;
extern rom unsigned char EESlv_Int0IsData;	
extern rom unsigned char EESlv_Reserved0B;	
extern rom unsigned char EESlv_Reserved0C;	
extern rom unsigned char EESlv_Reserved0D;	
extern rom unsigned char EESlv_Reserved0E;	
extern rom unsigned char EESlv_Reserved0F;	
extern rom unsigned char EESlv_Reserved1F[];	
#endif


// temporary copy of permanent settings 
unsigned char Slv_Security ;			
unsigned char Slv_Baud ;			
unsigned char Slv_FlowControl ;		
unsigned char Slv_HostHasRxBuffer ;		
unsigned char Slv_MessQueueLen ;			
unsigned char Slv_InitMsg ;				
unsigned char Slv_MsgsAnytime ;				
unsigned char Slv_ASCIIMessages ;				
unsigned char Slv_OnInternaError ;
unsigned char Slv_I2CSetup ;				
unsigned char Slv_Int0IsData ;				

// Message queue
unsigned char * pNextMessageToSend ;
unsigned char * pNextSpareMessagePos ;

#define SetSWISendMessage SetSWI( SWI_SWI2 )

// text strings
#define ResOKLen 1
rom unsigned char pResOK[] = { 0x01 };
#ifdef BASIC_Stamp_Edition
#define ResInitLen 12
rom unsigned char pResInit[] = "ToothPIC SE ";
#else
#define ResInitLen 6
rom unsigned char pResInit[] = "Slave ";
#endif
extern rom unsigned char pVersionStr[];  // defined in ToothPIC.c
#define ResErrQueueFullLen 3
rom unsigned char pResErrQueueFull[] = { 0x03, 0x02, 0xF0 };
#define ResErrLen 2
rom unsigned char pResErrNotUnderstood[] = { 0x02, 0xF1 };
#define ResErrLen 2
rom unsigned char pResErrPingFail[] = { 0x02, 0xF2 };
#define ResErrInternalLen 1
rom unsigned char pResErrInternal[] = { 0x02 };
#define ResDateTimeLen 1
rom unsigned char pResDateTime[] = { 0x03 };
#define ResIOValLen 1
rom unsigned char pResIOVal[] = { 0x04 };
#define ResBMTLen 1
rom unsigned char pResBMT[] = { 0x05 };
#define pZeroLen 1
rom unsigned char pZero[] = { 0x00 };
#define ResReceiveLen 2
rom unsigned char pResReceive[] = { 0x05, 0x10 };
#define ResFxPLen 1
rom unsigned char pResFxP[] = { 0x06 };
#define ResFxPInfoLen 1
rom unsigned char pResFxPInfo[] = { 0x07 };
#define ResCtlDataLen 1
rom unsigned char pResCtlData[] = { 0x08 };
#define ResGotBytesLen 1
rom unsigned char pResGotBytes[] = { 0x09 };


// function prototypes
void AppendToMessage( rom unsigned char * pROMStr, unsigned char * pRAMStr, unsigned char nChar );
void AddMessageToQueue(void);
char SendNextMessage( void );
unsigned char ExtractBinaryVal( unsigned char * pASCII );
void ConfigSlvCmd( void );
void ConfigIOCmd( void );
void GetIOCmd( void );
void SetIOCmd( void );
void BMTCmd( void );
void FxPCmd( void );
void GetFxPInfoCmd( void );
void GetCtlCmd( void );
void SetCtlPropsCmd( void );
void SetCtlCmd( void );
void SetRowCmd( unsigned char RowType );
void GetMemCmd( void );
void SetMemCmd( void );
void SendOKResponse(void);
void SendDateTimeResponse( void );
void InternalError( unsigned char ErrorID );
void SendIOValResponse( unsigned char iProperty, unsigned char iValue );
void SendADResponse( unsigned char iProperty );
unsigned short GetCtlIndex( unsigned char * pCtlIDBigEndian );
void GetMatrixStats( rom bqFxPCtlData* pCtl );
void FxPHostProgram( void );

// GetMatrixStats calculates matrix vital statistics on the fly
// results held in static variables since inefficient to pass so much as arguments
unsigned char Cellsize;  	// bytes per matrix cell int
unsigned char NumCols; 		// number of columns
signed long RowCounter;		// current value of 'number of valid rows' counter (can be -ve if appending)
ADD RowCountAddress;		// Row counter address in pCtl->VarStr memory
signed long RowToUpdate;			// if apending, row index of row where new data will be placed
signed long NextRowCounter;		// if apending, new value of 'number of valid rows' counter after appending
unsigned char RowSize;		// bytes per row of data
unsigned char XValSize;  	// bytes per X-axis value
signed short NumRows;		// number of rows of data available
signed short BytesInXCol;	// total number of bytes in the X axis column
signed short BytesInCol;	// total number of bytes in one matrix column


// Semaphores
unsigned char SlvSF = 0;
#define SlvSF_CommandError 			0x01
#define SlvSF_AwaitingBMTOK 		0x02
#define SlvSF_OKToSendNextMessage 	0x04  // Messages anytime is not set but a request for the next message has been made o rthis is the initialization message
#define SlvSF_UARTInit	 			0x08  // UART has been initialized and CTS interrupts can now be processed
#define SlvSF_ErrorLastCycle		0x10	// Uart error occurred
#define SlvSF_FxPServerRunning		0x40  	// new for V3.0.00005 - reset BlueMatik on general error
#define SlvSF_ResetMePlease			0x80  	// new for V3.0.00005 - reset BlueMatik on general error
#define IsUARTInit	( SlvSF & SlvSF_UARTInit )
#define SetUARTInit	{ SlvSF |= SlvSF_UARTInit; }
#define IsCmdErr	( SlvSF & SlvSF_CommandError )
#define SetCmdErr	{ SlvSF |= SlvSF_CommandError; }
#define ClearCmdErr	{ SlvSF &= ~SlvSF_CommandError; }
#define ReportNextBMTOK	{ SlvSF |= SlvSF_AwaitingBMTOK; }
#define IsDataPin 		(Slv_Int0IsData)


void main( void )
{
	unsigned char BaudClock;
	unsigned char BaudClock_High = 1;

	// if no BlueMatik, flash red led rapidly
	while ((ToothPICSemaphores&TPSF_BMTEXISTS)==0)
	{
		LedRed = ~LedRed;
		msDelay(50);
	}

	// initialize static variables
	ParallelA = 0;
	ParallelB = 0;
	CCPFunc = 0;
	NumAnalog = 0;

	// read permanently stored values 
	GetBytes( STR_EE, (ADD)&EESlv_Security, &Slv_Security, 1 );
	GetBytes( STR_EE, (ADD)&EESlv_Baud, &Slv_Baud, 1 );
	GetBytes( STR_EE, (ADD)&EESlv_FlowControl, &Slv_FlowControl, 1 );
	GetBytes( STR_EE, (ADD)&EESlv_HostHasRxBuffer, &Slv_HostHasRxBuffer, 1 );
	GetBytes( STR_EE, (ADD)&EESlv_MessQueueLen, &Slv_MessQueueLen, 1 );
	GetBytes( STR_EE, (ADD)&EESlv_InitMsg, &Slv_InitMsg, 1 );
	GetBytes( STR_EE, (ADD)&EESlv_MsgsAnytime, &Slv_MsgsAnytime, 1 );
	GetBytes( STR_EE, (ADD)&EESlv_ASCIIMessages, &Slv_ASCIIMessages, 1 );
	GetBytes( STR_EE, (ADD)&EESlv_OnInternaError, &Slv_OnInternaError, 1 );
	GetBytes( STR_EE, (ADD)&EESlv_I2CSetup, &Slv_I2CSetup, 1 );
	GetBytes( STR_EE, (ADD)&EESlv_Int0IsData, &Slv_Int0IsData, 1 );

	// send data pin high to show initializing
	// data pin			
	if (Slv_Int0IsData) 
	{
		DataPin = 1;
		DirData = 0;
	}

	// message queue
	TxMessageQueue = ((unsigned char *)MESSAGE_TOP) - (Slv_MessQueueLen * SHORTMESSLEN);
	pNextMessageToSend = TxMessageQueue;
	pNextSpareMessagePos = TxMessageQueue;
	*pNextSpareMessagePos = 1; // message length marker for first message

	// initialize serial port
	switch (Slv_Baud)
	{
		case SlvB_1220:
			BaudClock_High = 0;
			BaudClock = SPBaudBrghZero1220;
			break;
		case SlvB_2400:
			BaudClock_High = 0;
			BaudClock = SPBaudBrghZero2400;
			break;
		case SlvB_4800:
			BaudClock_High = 0;
			BaudClock = SPBaudBrghZero4800;
			break;
		case SlvB_9600:
			BaudClock = SPBaudBrghOne9600;
			break;
		case SlvB_19200:
			BaudClock = SPBaudBrghOne19200;
			break;
		case SlvB_38400:
			BaudClock = SPBaudBrghOne38400;
			break;
		case SlvB_57600:
			BaudClock = SPBaudBrghOne57600;
			break;
		case SlvB_115200:
			BaudClock = SPBaudBrghOne115200;
			break;
	}
	InitUART( BaudClock, BaudClock_High );
	RxBuffLen = 0;
	TxBuffLen = 1;
	TxBuffPos = 1; 

	// if flow control, on setup flow control
	if (Slv_FlowControl)
	{
		// flow control output
		switch (Slv_FlowControl)
		{
			case SlvF_NoRTS:
				break;				
			case SlvF_RTSisINT0:
				INT0Pin = RTSClear;
				DirINT0 = DirOutput;
				break;				
			case SlvF_RTSisSDO:
				SDOPin = RTSClear;
				DirSDO = DirOutput;
				break;				
			case SlvF_RTSisCCP1:
				CCP1Pin = RTSClear;
				DirCCP1 = DirOutput;
				break;				
			case SlvF_RTSisAN11:
				AN1Pin = RTSClear;
				DirAN1 = DirOutput;
				break;				
			case SlvF_RTSisSCL:
				SCLPin = RTSClear;
				DirSCL = DirOutput;
				break;				
		}

		// enable flow control input interrupt	
		DirCTS = DirInput;

		CTSIntFlag = 0;
		INTCON3bits.INT1IE = 1;  
		INTCON2bits.INTEDG1 = 0;  // trigger on clear
		INTCON3bits.INT1IP = 0;    // low priority interrupt
	}
	SetUARTInit;

	// initialize I2C memory
	switch (Slv_I2CSetup)
	{
		case SlvI_None:   break;
		case SlvI_I2C100kHz:  I2CMemInit100kHz;   break;
		case SlvI_I2C400kHz:  I2CMemInit400kHz;   break;
	}	

	// initialize security
	switch (Slv_Security)
	{
		case SlvS_None:  BMTCommand( BMTC_Security, 0, Sec_None );   break;
		case SlvS_Auth:  BMTCommand( BMTC_Security, 0, Sec_Auth );   break;
		case SlvS_AuthEncr:  BMTCommand( BMTC_Security, 0, Sec_AuthEncr );   break;
	}	
	AwaitBMTOK;			

	// finally, enable UART receive
	UARTEnable = 1; 

	// LED test completes initialization
	LedGreen = LedGreenOn;
	msDelay( 250 );
	LedGreen = LedGreenOff;
	LedRed = LedRedOn;
	msDelay( 250 );
	LedRed = LedRedOff;

	// transmit greetings message
	if ( Slv_InitMsg )
	{
		SlvSF |= SlvSF_OKToSendNextMessage;
		AppendToMessage( pResInit, 0, ResInitLen );
		AppendToMessage( pVersionStr, 0, VERSTRLEN-1 );
		AddMessageToQueue();
	}

	// clear data pin to show finished initializing 
	if (Slv_Int0IsData) 
	{
		DataPin = 0;
	}

	// main loop
	while ( 1 ) 
	{
		// wait half a second to ensure error led glows for at least that long before resetting error
		msDelay( 250 );
		msDelay( 250 );

		// processs Rx errors
		if (UARTEnable == 0)
		{
			// an error must have occurred
			if ((SlvSF & SlvSF_ErrorLastCycle) == 0)
			{
				// Error occurred this time through the loop; wait one cycle so Red LED is clearly visible
				SlvSF |= SlvSF_ErrorLastCycle;
			}
			else
			{
				// Error occured during previous loop; can reset UART Rx now
				SlvSF &= ~SlvSF_ErrorLastCycle;
				LedRed = LedRedOff;
				UARTEnable = 1;
			}
		}

		// new for V3.0.00005 - reset BlueMatik on general error
		if (SlvSF & SlvSF_ResetMePlease)
		{
			BMTCommand( BMTC_Reset, 0, 0 );  
			SlvSF &= ~SlvSF_ResetMePlease;

			// re-initialize flexipanel service
			if (SlvSF & SlvSF_FxPServerRunning)
			{
				FxPCommand( FxPC_Start, 0, 0 );		// Start FlexiPanel service
				AwaitBMTOK; 						// Service operating, no client
			}
		}

	}  // end of while ( 1 ) 
}

// code rather large for linker so break into small sections
#pragma code MemorySectionG

// HighInterrupt handler 
void HighInterrupt (void)
{
	// process Rx byte
	if (UARTRxIntFlag) 
	{
		unsigned char CommandLen;
		unsigned char bIsASCII;

		// note UARTRxIntFlag cleared by hardware
		if (UARTOverrunErr) 
		{
			Breakpoint( 111 );  // indicate serious error
		}
		if (UARTFrameErr) 
		{
			unsigned char ch;
			// turn on red led to indicate an error - probably Rx not connected
			LedRed = LedRedOn;
			UARTEnable = 0;  // reset uart
			ch = UARTChIn;  // clears UARTRxIntFlag
			return;
		}
		RxBuff[RxBuffLen] = UARTChIn;  // clears UARTRxIntFlag
		RxBuffLen++;

		// decide if ASCII
		bIsASCII = False;
		if ( (RxBuff[0] >='0' && RxBuff[0] <='9') || (RxBuff[0] >='a' && RxBuff[0] <='f') || (RxBuff[0] >='A' && RxBuff[0] <='F') ) bIsASCII = True;

		// if looks like ASCII command and this is the first byte of command length byte, do no further processing
		if (RxBuffLen==1 && bIsASCII ) 
		{
			return;
		}

		// determine command length
		if (bIsASCII) // calculate length for text commands
		{
			CommandLen = ExtractBinaryVal( RxBuff );
			CommandLen = CommandLen * 2 + 2;  // correct number of characetrs including <CR><LF>
			if ( CommandLen > LONGMESSLEN ) SetCmdErr;
		}
		else
		{
			CommandLen = RxBuff[0];  // length for binary commands
			if ( CommandLen > SHORTMESSLEN ) SetCmdErr;
		}

		if ( RxBuffLen >= CommandLen || IsCmdErr) // complete command has been read, or error occured
		{
			// set RTS to not clear
			if (Slv_FlowControl)
			{
				// flow control output
				switch (Slv_FlowControl)
				{
					case SlvF_NoRTS:
						break;				
					case SlvF_RTSisINT0:
						INT0Pin = RTSStop;
						break;				
					case SlvF_RTSisSDO:
						SDOPin = RTSStop;
						break;				
					case SlvF_RTSisCCP1:
						CCP1Pin = RTSStop;
						break;				
					case SlvF_RTSisAN11:
						AN1Pin = RTSStop;
						break;				
					case SlvF_RTSisSCL:
						SCLPin = RTSStop;
						break;				
				}
			}

			SetSWI( SWI_SWI1 ); // tell low priority interrrupt to process command
		}
		return;
	}
}

// code rather large for linker so break into small sections
#pragma code MemorySectionF

// LowInterrupt handler 
void LowInterrupt (void)
{

	if (IsSWI( SWI_Tick ) )
	{
		ClearSWI( SWI_Tick );		// Clear clock interrupt flag
		return;
	}

	// process Tx byte complete / CTS went low
	if ( (UARTTxIntEnable && UARTTxIntFlag) || (IsUARTInit && Slv_FlowControl && CTSIntFlag) ) 
	{
		if (!Slv_HostHasRxBuffer)  
		{
			while (!UARTNotTxmitting) {;} 	// wait until current byte has been transmitted
			msDelay( 1 );					// allow unbuffered devices time to process data
		}
		if ( (Slv_FlowControl==0 || CTSPin == CTSClear) && TxBuffPos < TxBuffLen)  // if CTS is clear and not currently transmitting and there is more data to transmit
		{
			UARTChOut = TxBuff[TxBuffPos];
			if (TxBuffPos<TxBuffLen) UARTIntAfterTx;  		// interrupt me when complete if more to transmit
			TxBuffPos++;
			if (IsDataPin && pNextSpareMessagePos==pNextMessageToSend) DataPin = 0;
		}
		else
		{
			// disable interrupt-on-complete
			UARTCnclIntAfterTx;
			if (IsDataPin && pNextSpareMessagePos==pNextMessageToSend) DataPin = 0;
		}

		// if end of message, load next message into buffer
		if (TxBuffPos==TxBuffLen) 
		{
			if (IsDataPin && pNextSpareMessagePos==pNextMessageToSend) 
			{
				DataPin = 0;
			}
			else
			{
				SetSWISendMessage;
			}
		}

		// clear cts
		if ( IsUARTInit && Slv_FlowControl )  CTSIntFlag = 0;
		return;
	}

	if (IsSWI( SWI_SWI1 ) )			// command to process
	{
		// length byte is zero, error
		if (!IsCmdErr && !RxBuff[0]) SetCmdErr;

		// translate to binary if required
		if (!IsCmdErr && RxBuffLen>1 && RxBuff[0] >='0') 
		{
			unsigned char iCh;

			// convert length byte to binary
			RxBuff[0] = ExtractBinaryVal( RxBuff );
			
			// if length byte is zero, reset
			if (!IsCmdErr && !RxBuff[0])  
			{
				Reset();
			}


			// generate error if final characters are not <CR><LF> or if too long
			if (!IsCmdErr && (RxBuff[RxBuffLen-2] != '\r' || RxBuff[RxBuffLen-1] != '\n' || RxBuff[0] > LONGMESSLEN )) SetCmdErr;
	
			// convert remaining bytes to binary
			for (iCh = 1; !IsCmdErr && iCh < RxBuff[0]; iCh++)
			{
				RxBuff[iCh] = ExtractBinaryVal( RxBuff+(iCh<<1) );
			}
			RxBuffLen = RxBuff[0];
		}

		// process command
		if (!IsCmdErr)
		{ 
			switch ( RxBuff[1] )
			{
				case 0x00: // reset
					 Reset();

				case 0x01:	// Config Peripheral Mode command
					ConfigSlvCmd();
					break;

				case 0x02:	// Config IO command
					ConfigIOCmd();
					break;

				case 0x03:	// Set IO command
					SetIOCmd();
					break;

				case 0x04:	// Get IO command

					GetIOCmd();
					break;

				case 0x05:	// bluematik command
					BMTCmd();
					break;

				case 0x06:	// flexipanel command
					FxPCmd();
					break;

				case 0x07:	// get flexipanel info
					GetFxPInfoCmd();
					break;

				case 0x08:	// get control
					GetCtlCmd();
					break;

				case 0x09:  // Set Control properties
					SetCtlPropsCmd();
					break;

				case 0x0A:  // Set Control value
					SetCtlCmd();
					break;

				case 0x0B:  // set row
				case 0x0C:  // append row
				case 0x0D:  // log row
					SetRowCmd(RxBuff[1]);
					break;

				case 0x0E:  // get memory
					GetMemCmd();
					break;

				case 0x0F:  // set memory
					SetMemCmd();
					break;

				case 0x10:  // breakpoint
					Breakpoint(RxBuff[2]);
					break;

				case 0x40:  // get message
					if (pNextSpareMessagePos==pNextMessageToSend)
					{
						// send OK message
						AppendToMessage( pResOK, 0, ResOKLen );
						AddMessageToQueue();
					}
					SlvSF |= SlvSF_OKToSendNextMessage;
					SetSWISendMessage; 
					break;

				default:  // not understood
					SetCmdErr;
					break;
			}
		}

		if (IsCmdErr)
		{
			// didn't understand command - error
			AppendToMessage( pResErrNotUnderstood, 0, ResErrLen );
			AddMessageToQueue();
			ClearCmdErr;
		}

		// reset incoming command buffer
		RxBuffLen = 0;

		// set RTS to clear
		if (Slv_FlowControl)
		{
			// flow control output
			switch (Slv_FlowControl)
			{
				case SlvF_NoRTS:
					break;				
				case SlvF_RTSisINT0:
					INT0Pin = RTSClear;
					break;				
				case SlvF_RTSisSDO:
					SDOPin = RTSClear;
					break;				
				case SlvF_RTSisCCP1:
					CCP1Pin = RTSClear;
					break;				
				case SlvF_RTSisAN11:
					AN1Pin = RTSClear;
					break;				
				case SlvF_RTSisSCL:
					SCLPin = RTSClear;
					break;				
			}
		}

		ClearSWI( SWI_SWI1 );		// Clear software interrupt
		SetSWISendMessage;
		return;
	}

	if (IsSWI(SWI_SWI2))  // at least one response has been placed on the queue
	{
		if (Slv_MsgsAnytime || (SlvSF&SlvSF_OKToSendNextMessage) ) SendNextMessage();
		SlvSF &= ~SlvSF_OKToSendNextMessage;

		// say clear to send - may have been not clear if an incoming message was being processed
		if (Slv_FlowControl)
		{
			// flow control output
			switch (Slv_FlowControl)
			{
				case SlvF_NoRTS:
					break;				
				case SlvF_RTSisINT0:
					INT0Pin = RTSClear;
					break;				
				case SlvF_RTSisSDO:
					SDOPin = RTSClear;
					break;				
				case SlvF_RTSisCCP1:
					CCP1Pin = RTSClear;
					break;				
				case SlvF_RTSisAN11:
					AN1Pin = RTSClear;
					break;				
				case SlvF_RTSisSCL:
					SCLPin = RTSClear;
					break;				
			}
		}
		ClearSWI( SWI_SWI2 );		// Clear software interrupt
		return;
	}

	if (IsSWI(SWI_BMTData))  // raw data received
	{
		// gather as much data as possible
		unsigned NumChar;
		do 
		{
			// wait 1ms to see if more data arrives
			NumChar = BMTRxNChar;
			msDelay(1);
		}
		while (	BMTRxNChar<19 && NumChar!=BMTRxNChar );

		// send up to 19 bytes
		NumChar = BMTRxNChar;
		if (NumChar > 19) NumChar = 19;

		// due to possibility that 		
		AppendToMessage( pResReceive, 0, ResReceiveLen );
		for ( ; NumChar!=0; NumChar--)
		{
			AppendToMessage( 0, (unsigned char*)BMTRxCh, 1 );
			BMTRxAdvanceCh;
		}
		AddMessageToQueue();

		if (!BMTRxNChar) ClearSWI( SWI_BMTData );		// Clear software interrupt
		return;
	}


}

// code rather large for linker so break into small sections
#pragma code MemorySectionH


// Error event handler 
void ErrorStatus( unsigned char ErrNum )
{
	InternalError( ErrNum );
}

// BlueMatik Event handler 
void BMTEvent( unsigned char EventID, void *pData1, void *pData2 )
{
	switch (EventID)
	{
		case BMTE_Error:
			if (((unsigned char*)pData1)[0] == 0x0F) // general error
			{
				SlvSF |= SlvSF_ResetMePlease;
				return;
			}
			InternalError( ((unsigned char*)pData1)[0] );
			return;
		case BMTE_Warning:
			InternalError( ((unsigned char*)pData1)[0] );
			return;
		case BMTE_LoPower:
		case BMTE_Disco:
			AppendToMessage( pResBMT, 0, ResBMTLen );
			AppendToMessage( 0, &EventID, 1 );
			AddMessageToQueue();
			return;
		case BMTE_Connect:
		case BMTE_Paired:
			AppendToMessage( pResBMT, 0, ResBMTLen );
			AppendToMessage( 0, &EventID, 1 );
			AppendToMessage( 0, pData1, BTADDRLEN );
			AddMessageToQueue();
			return;
		case BMTE_Found:
			AppendToMessage( pResBMT, 0, ResBMTLen );
			AppendToMessage( 0, &EventID, 1 );
			AppendToMessage( 0, pData1, BTADDRLEN );
			AppendToMessage( 0, pData2, 12 );
			AppendToMessage( pZero, 0, 1 );
			AddMessageToQueue();
			return;
		case BMTE_LinkQ:
		case BMTE_Signal:
			AppendToMessage( pResBMT, 0, ResBMTLen );
			AppendToMessage( 0, &EventID, 1 );
			AppendToMessage( 0, pData1, 1 );
			AddMessageToQueue();
			return;
		case BMTE_OK:
			if (SlvSF&SlvSF_AwaitingBMTOK)
			{
				SlvSF &= ~SlvSF_AwaitingBMTOK;
				SendOKResponse();
				SetSWISendMessage;
			}
			return;
// BMTE_HostProgram message handling is reserved for FlexiPanel Ltd use
#ifdef FxP_HostProgram
		case BMTE_HostProgram:
			FxPHostProgram();  // does not return
#endif
	}
}

// code rather large for linker so break into small sections
#pragma code MemorySectionI


// FxPEvent handler 
void FxPEvent( unsigned char EventID, void *pData )
{
	switch (EventID)
	{
		case FxPE_PingFail:
			AppendToMessage( pResErrPingFail, 0, ResErrLen );
			AddMessageToQueue();
			return;
		case FxPE_Connect:
		case FxPE_Disco:
			AppendToMessage( pResFxP, 0, ResFxPLen );
			AppendToMessage( 0, &EventID, 1 );
			AddMessageToQueue();
			return;
		case FxPE_ClntUpdate:
			AppendToMessage( pResFxP, 0, ResFxPLen );
			AppendToMessage( 0, &EventID, 1 );
			AppendToMessage( 0, pData+1, 1 );  // big endian...
			AppendToMessage( 0, pData, 1 );
			AddMessageToQueue();
			return;
		case FxPE_MsgResp:
			AppendToMessage( pResFxP, 0, ResFxPLen );
			AppendToMessage( 0, &EventID, 1 );
			AppendToMessage( 0, pData, 1 );
			AddMessageToQueue();
			return;
	}
}

// code rather large for linker so break into small sections
#pragma code MemorySectionE

void InternalError( unsigned char ErrorID )
{
	switch (Slv_OnInternaError)
	{
		case SlvE_OnErrFlash:
			Breakpoint( ErrorID );
			Reset();

		case SlvE_OnErrReset:
			Reset();

		case SlvE_OnErrResp:
			AppendToMessage( pResErrInternal, 0, ResErrInternalLen );
			AppendToMessage( 0, &ErrorID, 1 );
			AddMessageToQueue(); 
	}
}

// code rather large for linker so break into small sections
#pragma code MemorySectionJ

void AppendToMessage( rom unsigned char * pROMStr, unsigned char * pRAMStr, unsigned char nChar )
{
	// note - first byte of message, *NextSpareMessagePos, is current length of message
	for ( ; nChar!=0; nChar--)
	{
		if (pROMStr)
		{
			pNextSpareMessagePos[ *pNextSpareMessagePos ] = *pROMStr;
			pROMStr++;
		}
		else
		{
			pNextSpareMessagePos[ *pNextSpareMessagePos ] = *pRAMStr;
			pRAMStr++;
		}
		(*pNextSpareMessagePos)++;
	}
}

// send message completes adds this message to the queue and moves on to the next
// bIsSolicited does three things.  First, it puts the message at the front of the queue.
// second, if bIsSolicited is not set and bProcessingCommand is set, the send message interrupt is not raised
// third, if bIsSolicited is set, bProcessingCommand is cleared
// thus the response after a message will always be the response to the message
void AddMessageToQueue(void)
{
	unsigned char * pNextMessageSlotAfterThis;

	// calculate what will be the next spare message slot after this message 
	pNextMessageSlotAfterThis = pNextSpareMessagePos + SHORTMESSLEN;
	if (pNextMessageSlotAfterThis >= ((unsigned char *)MESSAGE_TOP)) pNextMessageSlotAfterThis = TxMessageQueue;

	// if this will fill the message queue, replace next message to be transmitted with message queue full and
	// ensure next message pointer is not incremented at end of this function
	if (pNextMessageSlotAfterThis==pNextMessageToSend)
	{
		unsigned char iCh;

		// set next transmitted message to full
		unsigned char LIS = LowIntState;
		DisableLowInts;
		DisableHighInts;  // stop the message being pulled off the queue
		for ( iCh = 0; iCh < ResErrQueueFullLen; iCh++)
		{
			pNextMessageToSend[iCh] = pResErrQueueFull[iCh];
		}
		EnableHighInts;
		if (LIS) EnableLowInts;

		// do not increment message slot position
		pNextMessageSlotAfterThis = pNextSpareMessagePos;
	}
	else
	{
		// initialize length of next message marker
		*pNextMessageSlotAfterThis = 1;
	}

	pNextSpareMessagePos = pNextMessageSlotAfterThis;

	// raise SWI to send message
	SetSWISendMessage;

	// update data pin
	if (IsDataPin) DataPin = 1;
}

// code rather large for linker so break into small sections
#pragma code MemorySectionK

// SendNextMessage pulls a message off the queue and transmits it
// returns zero if no message sent
char SendNextMessage( void )
{
	unsigned char iCh;
	unsigned char LIS;

	// if previous transmit not finished, do nothing
	if (TxBuffLen != TxBuffPos)  return 0xFF;

	// if no messages return
	if (pNextSpareMessagePos==pNextMessageToSend) 
	{
		if (IsDataPin) DataPin = 0;
		return 0x00;
	}

	// fill buffer either with binary or hex as required
	TxBuffLen = 0;
	if (Slv_ASCIIMessages)
	{
		// <CR>
		TxBuff[TxBuffLen] = '\r';
		TxBuffLen++;
		// <LF>
		TxBuff[TxBuffLen] = '\n';
		TxBuffLen++;				
		for (iCh = 0; iCh < *pNextMessageToSend; iCh++)
		{
			// high nibble
			TxBuff[TxBuffLen] = '0' + (pNextMessageToSend[iCh] >> 4);
			if (TxBuff[TxBuffLen] > '9') TxBuff[TxBuffLen] += 7;  // letters 'A' to 'F'
			TxBuffLen++;
			// low nibble
			TxBuff[TxBuffLen] = '0' + (pNextMessageToSend[iCh] & 0x0F);
			if (TxBuff[TxBuffLen] > '9') TxBuff[TxBuffLen] += 7;  // letters 'A' to 'F'
			TxBuffLen++;
		}
		// <CR>
		TxBuff[TxBuffLen] = '\r';
		TxBuffLen++;
		// <LF>
		TxBuff[TxBuffLen] = '\n';
		TxBuffLen++;				
	}
	else
	{
		for (iCh = 0; iCh < *pNextMessageToSend; iCh++)
		{
			TxBuff[TxBuffLen] = pNextMessageToSend[iCh];
			TxBuffLen++;
		}
	}

	// increment message pointer
	pNextMessageToSend = pNextMessageToSend + SHORTMESSLEN;
	if (pNextMessageToSend >= ((unsigned char *)MESSAGE_TOP)) pNextMessageToSend = TxMessageQueue;

	// set data pin
	if (IsDataPin) DataPin = 1;

	// if CTS is clear, start to transmit message
	LIS = LowIntState;
	DisableLowInts;
	DisableHighInts;
	// if no CTS=stop and empty, start send now, otherwise wait for interrupt
	if ((Slv_FlowControl==0 || CTSPin == CTSClear) && UARTTxIntFlag)
	{
		UARTChOut = *TxBuff;
		UARTIntAfterTx;  // interrupt me when complete
		TxBuffPos = 1;
	}
	else
	{
		// otherwise, setting TxBuffPos to zero indicates to interrupt secvice routine that it can start transmitting the message
		TxBuffPos = 0;
		// if currently transmitting, request an interrupt when done
		if (!UARTTxIntFlag) 
		{
			UARTIntAfterTx;
		}
	}
	EnableHighInts;
	if (LIS) EnableLowInts;

	return 0xFF;
}

// code rather large for linker so break into small sections
#pragma code MemorySectionL

// convert hexadecimal ascii to binary
unsigned char ExtractBinaryVal( unsigned char * pASCII )
{

	unsigned char ch, RetVal;

	// convert to upper case
	if (pASCII[0]>='a' && pASCII[0]<='f') pASCII[0]-=0x20;
	if (pASCII[1]>='a' && pASCII[1]<='f') pASCII[1]-=0x20;

	// check valid
	if (pASCII[0] < '0' || (pASCII[0]>'9' && pASCII[0]<'A') || pASCII[0]>'F' || 
		pASCII[1] < '0' || (pASCII[1]>'9' && pASCII[1]<'A') || pASCII[1]>'F' ) 
	{
		SetCmdErr;
		return 0xFF;
	}

	// convert to binary
	ch = pASCII[0] - '0';
	if (ch>9) ch -=7;
	RetVal = ch << 4;
	ch = pASCII[1] - '0';
	if (ch>9) ch -=7;
	RetVal += ch;

	return RetVal;
}

void SendOKResponse( void )
{
	// if no messages anytime, do nothing
	if (!Slv_MsgsAnytime) return;

	// send OK message
	AppendToMessage( pResOK, 0, ResOKLen );
	AddMessageToQueue();
}

// code rather large for linker so break into small sections
#pragma code MemorySectionM

void SendDateTimeResponse( void )
{
	// if no messages anytime, do nothing
	if (!Slv_MsgsAnytime) return;

	// send OK message
	AppendToMessage( pResDateTime, 0, ResDateTimeLen );
	AppendToMessage( 0, (unsigned char *)&RealTimeClock, sizeof(RealTimeClock) );
	AddMessageToQueue();
}

#ifdef SBD
extern unsigned char SetBytesDebug;
#endif

void ConfigSlvCmd( void )
{
	// error if no 3rd byte
	if (RxBuffLen<3) { SetCmdErr;  return; }

	// read 3rd byte to identify the property being set
	switch (RxBuff[2])
	{
		case 0x01:  // security
			if (RxBuffLen<4 || RxBuff[3] > 0x02) { SetCmdErr;  return; }
			SetBytes( STR_EE, (ADD)&EESlv_Security, 0, &RxBuff[3], 1 );
			break;
		case 0x02:  // baud
			if (RxBuffLen<4 || RxBuff[3] < 0x01 || RxBuff[3] > 0x08) { SetCmdErr;  return; }
			SetBytes( STR_EE, (ADD)&EESlv_Baud, 0, &RxBuff[3], 1 );
			break;
		case 0x03:  // PIN
			if (RxBuffLen<4 || RxBuff[3] == 0x00 ) { SetCmdErr;  return; }
			// ensure some kind of zero terminator exists
			RxBuff[19] = 0x00;
#ifdef SBD
SetBytesDebug = 10;
#endif
			SetBytes( STR_ROM00, (ADD)pszPIN, 0, &RxBuff[3], 17 );
#ifdef SBD
SetBytesDebug = 11;
#endif
			break;
		case 0x04:  // Name
			if (RxBuffLen<4 || RxBuff[3] == 0x00 ) { SetCmdErr;  return; }
			// ensure some kind of zero terminator exists
			RxBuff[19] = 0x00;
#ifdef SBD
SetBytesDebug = 12;
#endif
			SetBytes( STR_ROM00, (ADD)pLocalName, 0, &RxBuff[3], 17 );
#ifdef SBD
SetBytesDebug = 13;
#endif
			break;
		case 0x05:  // flow control
			if (RxBuffLen<4 || (RxBuff[3] > 0x05 && RxBuff[3]!=0x10) ) { SetCmdErr;  return; }
			SetBytes( STR_EE, (ADD)&EESlv_FlowControl, 0, &RxBuff[3], 1 );
			break;
		case 0x06:  // host has rx buffer
			if (RxBuffLen<4) { SetCmdErr;  return; }  // any nonzero value is assumed to mean 'true'
			SetBytes( STR_EE, (ADD)&EESlv_HostHasRxBuffer, 0, &RxBuff[3], 1 );
			break;
		case 0x07:  // message queue length
			if (RxBuffLen<4) { SetCmdErr;  return; }
			SetBytes( STR_EE, (ADD)&EESlv_MessQueueLen, 0, &RxBuff[3], 1 );
			break;
		case 0x08:  // initialization response
			if (RxBuffLen<4) { SetCmdErr;  return; }  // any nonzero value is assumed to mean 'true'
			SetBytes( STR_EE, (ADD)&EESlv_InitMsg, 0, &RxBuff[3], 1 );
			break;
		case 0x09:  // messages anytime
			if (RxBuffLen<4) { SetCmdErr;  return; }  // any nonzero value is assumed to mean 'true'
			SetBytes( STR_EE, (ADD)&EESlv_MsgsAnytime, 0, &RxBuff[3], 1 );
			break;
		case 0x0A:  // ASCII responses
			if (RxBuffLen<4) { SetCmdErr;  return; }  // any nonzero value is assumed to mean 'true'
			SetBytes( STR_EE, (ADD)&EESlv_ASCIIMessages, 0, &RxBuff[3], 1 );
			break;
		case 0x0B:  // On internal error
			if (RxBuffLen<4  || RxBuff[3] == 0x00 || RxBuff[3] > 0x03) { SetCmdErr;  return; }
			SetBytes( STR_EE, (ADD)&EESlv_OnInternaError, 0, &RxBuff[3], 1 );
			break;
		case 0x0C:  // I2C Setup
			if (RxBuffLen<4 || RxBuff[3] > 0x02) { SetCmdErr;  return; }
			SetBytes( STR_EE, (ADD)&EESlv_I2CSetup, 0, &RxBuff[3], 1 );
			break;
		case 0x0D:  // Device class
			if (RxBuffLen<6) { SetCmdErr;  return; }
			// swap first and third byte
			RxBuff[0] = RxBuff[3];
			RxBuff[3] = RxBuff[5];
			RxBuff[5] = RxBuff[0];
#ifdef SBD
SetBytesDebug = 14;
#endif
			SetBytes( STR_ROM00, (ADD)&DeviceClass, 0, &RxBuff[3], 3 );
#ifdef SBD
SetBytesDebug = 15;
#endif
			break;
		case 0x0E:  // INT0 is data
			if (RxBuffLen<4) { SetCmdErr;  return; }
			// note new value
			SetBytes( STR_EE, (ADD)&EESlv_Int0IsData, 0, &RxBuff[3], 1 );
			break;
		case 0x41:  // daylight savings state
			if (RxBuffLen<4) { SetCmdErr;  return; }
			DSTEvent = RxBuff[3];
			break;
		case 0x42:  // set time
			if (RxBuffLen<10) { SetCmdErr;  return; }
			RealTimeClock.sec = RxBuff[3];
			RealTimeClock.min = RxBuff[4];
			RealTimeClock.hour = RxBuff[5];
			RealTimeClock.date = RxBuff[6];
			RealTimeClock.month = RxBuff[8];
			RealTimeClock.year = *(( unsigned short *) (&RxBuff[9]) );
			RealTimeClock.dow = RxBuff[8];
			if (RealTimeClock.dow>=0x07)
			{
				CalcDayOfWeek( RealTimeClock.dow, RealTimeClock.year, RealTimeClock.month, RealTimeClock.date ) 
			}
			break;
		case 0x43:  // request data / time
			SendDateTimeResponse();
			return;
		default:
			SetCmdErr;  
			return; 
	}

	// generate OK Response
	SendOKResponse();
}

// code rather large for linker so break into small sections
#pragma code MemorySectionN

void ConfigIOCmd( void )
{
	// error if no 3rd byte
	if (RxBuffLen<3) { SetCmdErr;  return; }

	// read 3rd byte to identify the property being set
	switch (RxBuff[2])
	{
		case 0x01:  // a to d channels
			NumAnalog = RxBuff[3];
			if (RxBuffLen<4 || NumAnalog > 0x0C) { SetCmdErr;  return; }
			// set number of channels
			ADCON1 = (ADCON1 & 0xF0) | (0x0F-NumAnalog);
			// tun a to d on or off according to number of channels
			if (NumAnalog)
				ADConverterOn10bit
			else
				ADConverterOff
			break;
		case 0x02:  // -ve volt ref
			if (RxBuffLen<4) { SetCmdErr;  return; }  // any nonzero value is assumed to mean 'true'
			if (RxBuff[3])
				VRefNegIsAN2
			else
				VRefNegIsVss
			break;
		case 0x03:  // -ve volt ref
			if (RxBuffLen<4) { SetCmdErr;  return; }  // any nonzero value is assumed to mean 'true'
			if (RxBuff[3])
				VRefPosIsAN3
			else
				VRefNegIsVss
			break;
		case 0x04:  // PWM time base units
			if (RxBuffLen<4 || RxBuff[3] == 0 || RxBuff[3] > 0x03) { SetCmdErr;  return; }
			if (!RxBuff[3])
			{  // PWM off
				T4CON = 0b00000000;  // timer 4 off
			}
			else
			{  // PWM on, set time base
				T3CONbits.T3CCP1 = 1;  // set PWM to timer 4
				T3CONbits.T3CCP2 = 1;  // set PWM to timer 4
				T4CON = 0b00000100;  // 0.2us timebase
				if (RxBuff[3]==2) T4CON |= 0b00000001;  // 0.8us timebase
				if (RxBuff[3]==3) T4CON |= 0b00000010;  // 3.2us timebase	
			}		
			break;
		case 0x05:  // PWM period
			if (RxBuffLen<4) { SetCmdErr;  return; }
			PR4 = RxBuff[3];  // timer period = (PR4 + 1) x (PWM time base units)
			break;
		case 0x06:  // Parallel I/O
			if (RxBuffLen<4 || RxBuff[3] == 0x01 || (RxBuff[3]>0x07 && RxBuff[3]<0x12) || RxBuff[3] > 0x17 ) { SetCmdErr;  return; }		
			ParallelA = RxBuff[3]; 
			switch (ParallelA)
			{
				case 0x00:  break;

				case 0x07:  DirAN5 = DirOutput; // note how case statements fall through
				case 0x06:	DirAN6 = DirOutput;
				case 0x05:	DirAN7 = DirOutput;
				case 0x04:	DirAN8 = DirOutput;
				case 0x03:	DirAN9 = DirOutput;
				case 0x02:	DirAN10 = DirOutput;
							DirAN11 = DirOutput;
							break;

				case 0x17:  DirAN5 = DirInput; // note how case statements fall through
				case 0x16:	DirAN6 = DirInput;
				case 0x15:	DirAN7 = DirInput;
				case 0x14:	DirAN8 = DirInput;
				case 0x13:	DirAN9 = DirInput;
				case 0x12:	DirAN10 = DirInput;
							DirAN11 = DirInput;
							break;
			}
			break;
		case 0x07:  // Parallel I/O
			if (RxBuffLen<4 || RxBuff[3] == 0x01 || (RxBuff[3]>0x05 && RxBuff[3]<0x12) || RxBuff[3] > 0x15 ) { SetCmdErr;  return; }		
			ParallelB = RxBuff[3]; 
			switch (ParallelB)
			{
				case 0x00:  break;

				case 0x05:	DirAN0 = DirOutput;
				case 0x04:	DirAN1 = DirOutput;
				case 0x03:	DirAN2 = DirOutput;
				case 0x02:	DirAN3 = DirOutput;
							DirAN4 = DirOutput;
							break;

				case 0x15:	DirAN0 = DirInput;
				case 0x14:	DirAN1 = DirInput;
				case 0x13:	DirAN2 = DirInput;
				case 0x12:	DirAN3 = DirInput;
							DirAN4 = DirInput;
							break;
			}
			break;
		case 0x10:  // AN0 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN0 = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x11:  // AN1 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN1 = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x12:  // AN2 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN2 = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x13:  // AN3 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN3 = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x14:  // AN4 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN4 = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x15:  // AN5 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN5 = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x16:  // AN6 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN6 = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x17:  // AN7 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN7 = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x18:  // AN8 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN8 = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x19:  // AN9 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN9 = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x1A:  // AN10 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN10 = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x1B:  // AN11 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirAN11 = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x1C:  // CCP1 digital pin
			if (RxBuffLen<4 || RxBuff[3]>0x02 ) { SetCmdErr;  return; }	  
			if (RxBuff[3]==0x02)
			{
				CCP1CON = 0x0F;   // turn PWM on, zero output
				CCPR1L = 0; 
				CCPFunc |= CCP1Func;
				DirCCP1 = DirOutput;
			}
			else
			{
				DirCCP1 = (RxBuff[3] ? DirOutput : DirInput); 
				CCPFunc &= ~CCP1Func;
			}
			break;
		case 0x1D:  // CCP2 digital pin
			if (RxBuffLen<4 || RxBuff[3]>0x02 ) { SetCmdErr;  return; }	  
			if (RxBuff[3]==0x02)
			{
				CCP2CON = 0x0F;   // turn PWM on, zero output
				CCPR2L = 0; 
				CCPFunc |= CCP2Func;
				DirCCP2 = DirOutput;
			}
			else
			{
				DirCCP2 = (RxBuff[3] ? DirOutput : DirInput); 
				CCPFunc &= ~CCP2Func;
			}
			break;
		case 0x1E:  // CCP3 digital pin
			if (RxBuffLen<4 || RxBuff[3]>0x02 ) { SetCmdErr;  return; }	  
			if (RxBuff[3]==0x02)
			{
				CCP3CON = 0x0F;   // turn PWM on, zero output
				CCPR3L = 0; 
				CCPFunc |= CCP3Func;
				DirCCP3 = DirOutput;
			}
			else
			{
				DirCCP3 = (RxBuff[3] ? DirOutput : DirInput); 
				CCPFunc &= ~CCP3Func;
			}
			break;
		case 0x1F:  // CCP4 digital pin
			if (RxBuffLen<4 || RxBuff[3]>0x02 ) { SetCmdErr;  return; }	  
			if (RxBuff[3]==0x02)
			{
				CCP4CON = 0x0F;   // turn PWM on, zero output
				CCPR4L = 0; 
				CCPFunc |= CCP4Func;
				DirCCP4 = DirOutput;
			}
			else
			{
				DirCCP4 = (RxBuff[3] ? DirOutput : DirInput); 
				CCPFunc &= ~CCP4Func;
			}
			break;
		case 0x20:  // CCP5 digital pin
			if (RxBuffLen<4 || RxBuff[3]>0x02 ) { SetCmdErr;  return; }	  
			if (RxBuff[3]==0x02)
			{
				CCP5CON = 0x0F;   // turn PWM on, zero output
				CCPR5L = 0; 
				CCPFunc |= CCP5Func;
				DirCCP5 = DirOutput;
			}
			else
			{
				DirCCP5 = (RxBuff[3] ? DirOutput : DirInput); 
				CCPFunc &= ~CCP5Func;
			}
			break;
		case 0x21:  // INT0 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirINT0 = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x22:  // INT1 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirINT1 = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x23:  // SCL digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirSCL = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x24:  // SDA digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirSDA = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		case 0x25:  // SDO digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'output'	
			DirSDO = (RxBuff[3] ? DirOutput : DirInput); 
			break;
		default:
			SetCmdErr;  
			return; 
	}

	// generate OK Response
	SendOKResponse();
}

// code rather large for linker so break into small sections
#pragma code MemorySectionA

void SetIOCmd( void )
{
	// error if no 3rd byte
	if (RxBuffLen<3) { SetCmdErr;  return; }

	// read 3rd byte to identify the property being set
	switch (RxBuff[2])
	{
		case 0x06:  // Parallel I/O A
			if ( RxBuffLen<4 ) { SetCmdErr;  return; }		
			switch (ParallelA)
			{
				case 0x07:  SetPA7(  RxBuff[3] ); break;
				case 0x06:  SetPA6(  RxBuff[3] ); break;
				case 0x05:  SetPA5(  RxBuff[3] ); break;
				case 0x04:  SetPA4(  RxBuff[3] ); break;
				case 0x03:  SetPA3(  RxBuff[3] ); break;
				case 0x02:  SetPA2(  RxBuff[3] ); break;
				default: break;
			}
			break;
		case 0x07:  // Parallel I/O B
			if ( RxBuffLen<4 ) { SetCmdErr;  return; }		
			switch (ParallelB)
			{
				case 0x05:  SetPB5(  RxBuff[3] ); break;
				case 0x04:  SetPB4(  RxBuff[3] ); break;
				case 0x03:  SetPB3(  RxBuff[3] ); break;
				case 0x02:  SetPB2(  RxBuff[3] ); break;
				default: break;
			}
			break;
		case 0x10:  // AN0 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN0Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x11:  // AN1 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN1Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x12:  // AN2 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN2Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x13:  // AN3 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN3Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x14:  // AN4 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN4Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x15:  // AN5 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN5Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x16:  // AN6 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN6Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x17:  // AN7 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN7Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x18:  // AN8 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN8Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x19:  // AN9 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN9Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x1A:  // AN10 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN10Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x1B:  // AN11 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			AN11Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x1C:  // CCP1 PWM / digital pin
			if (RxBuffLen<4) { SetCmdErr;  return; }	  
			if (CCPFunc & CCP1Func)  // if is PWM
			{
				unsigned char iVal = RxBuff[4];  		  // Note, high byte sent first since more intuitive for ASCII commands; FlexiPanel Ltd is usually little-endian
				if (RxBuffLen<5) { SetCmdErr;  return; }	  
				// lower 2 bits
				CCP1CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
				// upper 8 bits
				iVal = (iVal >> 2) + (RxBuff[3] << 6);
				CCPR1L = (unsigned char)iVal;
			}
			else
				CCP1Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x1D:  // CCP2 PWM / digital pin
			if (RxBuffLen<4) { SetCmdErr;  return; }	  
			if (CCPFunc & CCP2Func)  // if is PWM
			{
				unsigned char iVal = RxBuff[4];
				if (RxBuffLen<5) { SetCmdErr;  return; }	  
				// lower 2 bits
				CCP2CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
				// upper 8 bits
				iVal = (iVal >> 2) + (RxBuff[3] << 6);
				CCPR2L = (unsigned char)iVal;
			}
			else
				CCP2Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x1E:  // CCP3 PWM / digital pin
			if (RxBuffLen<4) { SetCmdErr;  return; }	  
			if (CCPFunc & CCP3Func)  // if is PWM
			{
				unsigned char iVal = RxBuff[4];
				if (RxBuffLen<5) { SetCmdErr;  return; }	  
				// lower 2 bits
				CCP3CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
				// upper 8 bits
				iVal = (iVal >> 2) + (RxBuff[3] << 6);
				CCPR3L = (unsigned char)iVal;
			}
			else
				CCP3Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x1F:  // CCP4 PWM / digital pin
			if (RxBuffLen<4) { SetCmdErr;  return; }	  
			if (CCPFunc & CCP4Func)  // if is PWM
			{
				unsigned char iVal = RxBuff[4];
				if (RxBuffLen<5) { SetCmdErr;  return; }	  
				// lower 2 bits
				CCP4CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
				// upper 8 bits
				iVal = (iVal >> 2) + (RxBuff[3] << 6);
				CCPR4L = (unsigned char)iVal;
			}
			else
				CCP4Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x20:  // CCP5 PWM / digital pin
			if (RxBuffLen<4) { SetCmdErr;  return; }	  
			if (CCPFunc & CCP5Func)  // if is PWM
			{
				unsigned char iVal = RxBuff[4];
				if (RxBuffLen<5) { SetCmdErr;  return; }	  
				// lower 2 bits
				CCP5CON = (( ((unsigned char)iVal)&0b00000011) << 4) | 0b00001100;
				// upper 8 bits
				iVal = (iVal >> 2) + (RxBuff[3] << 6);
				CCPR5L = (unsigned char)iVal;
			}
			else
				CCP5Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x21:  // INT0 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			INT0Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x22:  // INT1 digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			INT1Pin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x23:  // SCL digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			SCLPin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x24:  // SDA digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			SDAPin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x25:  // SDO digital pin
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			SDOPin = (RxBuff[3] ? 1 : 0); 
			break;
		case 0x30:  // Green LED
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			LedGreen = (RxBuff[3] ? LedGreenOn : LedGreenOff); 
			break;
		case 0x31:  // Red LED
			if (RxBuffLen<4 ) { SetCmdErr;  return; }	  // any nonzero value is assumed to mean 'high'	
			LedRed = (RxBuff[3] ? LedRedOn : LedRedOff); 
			break;
		default:
			SetCmdErr;  
			return; 
	}

	// generate OK Response
	SendOKResponse();
}

// code rather large for linker so break into small sections
#pragma code MemorySectionP

void GetIOCmd( void )
{
	// error if no 3rd byte
	if (RxBuffLen<3) { SetCmdErr;  return; }

	// read 3rd byte to identify the property being requested
	switch (RxBuff[2])
	{
		case 0x06:  // Parallel I/O A
			switch (ParallelA)
			{
				case 0x17:  SendIOValResponse( RxBuff[2], GetPA7 ); return;
				case 0x16:  SendIOValResponse( RxBuff[2], GetPA6 ); return;
				case 0x15:  SendIOValResponse( RxBuff[2], GetPA5 ); return;
				case 0x14:  SendIOValResponse( RxBuff[2], GetPA4 ); return;
				case 0x13:  SendIOValResponse( RxBuff[2], GetPA3 ); return;
				case 0x12:  SendIOValResponse( RxBuff[2], GetPA2 ); return;
				default: 
					SetCmdErr;
					return;
			}
		case 0x07:  // Parallel I/O B
			switch (ParallelB)
			{
				case 0x15:  SendIOValResponse( RxBuff[2], GetPB5 ); return;
				case 0x14:  SendIOValResponse( RxBuff[2], GetPB4 ); return;
				case 0x13:  SendIOValResponse( RxBuff[2], GetPB3 ); return;
				case 0x12:  SendIOValResponse( RxBuff[2], GetPB2 ); return;
				default: 
					SetCmdErr;
					return;
			}
		case 0x10:  // AN0 analog / digital pin
			if (NumAnalog>0)
				SendADResponse( RxBuff[2] );
			else
				SendIOValResponse( RxBuff[2], AN0Pin ); 
			return;
		case 0x11:  // AN1 analog / digital pin
			if (NumAnalog>1)
				SendADResponse( RxBuff[2] );
			else
				SendIOValResponse( RxBuff[2], AN1Pin ); 
			return;
		case 0x12:  // AN2 analog / digital pin
			if (NumAnalog>2)
				SendADResponse( RxBuff[2] );
			else
				SendIOValResponse( RxBuff[2], AN2Pin ); 
			return;
		case 0x13:  // AN3 analog / digital pin
			if (NumAnalog>3)
				SendADResponse( RxBuff[2] );
			else
				SendIOValResponse( RxBuff[2], AN3Pin ); 
			return;
		case 0x14:  // AN4 analog / digital pin
			if (NumAnalog>4)
				SendADResponse( RxBuff[2] );
			else
				SendIOValResponse( RxBuff[2], AN4Pin ); 
			return;
		case 0x15:  // AN5 analog / digital pin
			if (NumAnalog>5)
				SendADResponse( RxBuff[2] );
			else
				SendIOValResponse( RxBuff[2], AN5Pin ); 
			return;
		case 0x16:  // AN6 analog / digital pin
			if (NumAnalog>6)
				SendADResponse( RxBuff[2] );
			else
				SendIOValResponse( RxBuff[2], AN6Pin ); 
			return;
		case 0x17:  // AN7 analog / digital pin
			if (NumAnalog>7)
				SendADResponse( RxBuff[2] );
			else
				SendIOValResponse( RxBuff[2], AN7Pin ); 
			return;
		case 0x18:  // AN8 analog / digital pin
			if (NumAnalog>8)
				SendADResponse( RxBuff[2] );
			else
				SendIOValResponse( RxBuff[2], AN8Pin ); 
			return;
		case 0x19:  // AN9 analog / digital pin
			if (NumAnalog>9)
				SendADResponse( RxBuff[2] );
			else
				SendIOValResponse( RxBuff[2], AN9Pin ); 
			return;
		case 0x1A:  // AN10 analog / digital pin
			if (NumAnalog>10)
				SendADResponse( RxBuff[2] );
			else
				SendIOValResponse( RxBuff[2], AN10Pin ); 
			return;
		case 0x1B:  // AN11 analog / digital pin
			if (NumAnalog>11)
				SendADResponse( RxBuff[2] );
			else
				SendIOValResponse( RxBuff[2], AN11Pin ); 
			return;
		case 0x1C:  // CCP1 digital pin
			SendIOValResponse( RxBuff[2], CCP1Pin ); 
			return;
		case 0x1D:  // CCP2 digital pin
			SendIOValResponse( RxBuff[2], CCP2Pin ); 
			return;
		case 0x1E:  // CCP3 digital pin
			SendIOValResponse( RxBuff[2], CCP3Pin ); 
			return;
		case 0x1F:  // CCP4 digital pin
			SendIOValResponse( RxBuff[2], CCP4Pin ); 
			return;
		case 0x20:  // CCP5 digital pin
			SendIOValResponse( RxBuff[2], CCP5Pin ); 
			return;
		case 0x21:  // INT0 digital pin
			SendIOValResponse( RxBuff[2], INT0Pin );
			return;
		case 0x22:  // INT1 digital pin
			SendIOValResponse( RxBuff[2], INT1Pin );
			return;
		case 0x23:  // SCL digital pin
			SendIOValResponse( RxBuff[2], SCLPin );
			return;
		case 0x24:  // SDA digital pin
			SendIOValResponse( RxBuff[2], SDAPin );
			return;
		case 0x25:  // SDO digital pin
			SendIOValResponse( RxBuff[2], SDOPin );
			return;
		case 0x30:  // Green LED
			SendIOValResponse( RxBuff[2], (LedGreen==LedGreenOn) ? 0x01 : 0x00 );
			break;
		case 0x31:  // Red LED
			SendIOValResponse( RxBuff[2], (LedRed==LedRedOn) ? 0x01 : 0x00 );
			break;
		case 0x32:  // Pushbutton
			SendIOValResponse( RxBuff[2], (PushButton==PushButtonOn) ? 0x01 : 0x00 );
			break;
		default:
			SetCmdErr;  
			return; 
	}
}

void BMTCmd( void )
{
	// error if no 3rd byte
	if (RxBuffLen<3) { SetCmdErr;  return; }

	// read 3rd byte to identify the property being requested
	switch (RxBuff[2])
	{
		case 0x01:  // commands not requiring additional data
		case 0x03:
		case 0x06:
		case 0x07:
		case 0x08:
		case 0x09:
		case 0x0A:
		case 0x0B:
			BMTCommand( RxBuff[2], 0, 0 );
			ReportNextBMTOK;
			break;
		case 0x02:  // master
			if (RxBuffLen<9) { SetCmdErr;  return; }
			BMTCommand( RxBuff[2], &RxBuff[3], 0 );
			ReportNextBMTOK;
			break;
		case 0x04:  // inquire
			if (RxBuffLen<4) { SetCmdErr;  return; }
			// convert duration to ascii string
			RxBuff[4] = '0' + (RxBuff[3]&0x0F);
			if (RxBuff[4]>'9') RxBuff[4] += 0x07;
			RxBuff[3] = '0' + (RxBuff[3]>>4);
			RxBuff[5] = 0;
			BMTCommand( RxBuff[2], &RxBuff[3], 0 );
			ReportNextBMTOK;
			break;
		case 0x10:  // transmit command
			BMTTransmit( 0, &RxBuff[3], RxBuffLen-3, 0 );
			SendOKResponse();
			break;
		default:
			SetCmdErr;  
			return; 
	}
}

void FxPCmd( void )
{
	// error if no 3rd byte
	if (RxBuffLen<3) { SetCmdErr;  return; }

	// read 3rd byte to identify the property being requested
	switch (RxBuff[2])
	{
		case 0x01:  // commands not requiring additional data
		case 0x02:
		case 0x04:
		case 0x09:
			FxPCommand( RxBuff[2], 0, 0 );
			SendOKResponse();

			// log whether flexipanel server is running just in casse it's necessary to reset BlueMatik
			if (RxBuff[2]==0x01) SlvSF |= SlvSF_FxPServerRunning;
			if (RxBuff[2]==0x02) SlvSF &= ~SlvSF_FxPServerRunning;

			break;
		case 0x03:  // set dialog
			if (RxBuffLen<4 || RxBuffLen>4) { SetCmdErr;  return; }
			FxPCommand( RxBuff[2], RxBuff[3], 0 );
			SendOKResponse();
			break;
		default:
			SetCmdErr;  
			return; 
	}
}

// code rather large for linker so break into small sections
#pragma code MemorySectionB

void GetFxPInfoCmd( void )
{
	// error if no 3rd byte
	if (RxBuffLen<3) { SetCmdErr;  return; }

	// read 3rd byte to identify the property being requested
	switch (RxBuff[2])
	{
		case 0x01:  // general info
			AppendToMessage( pResFxPInfo, 0, ResFxPInfoLen );
			AppendToMessage( 0, (unsigned char *)&RxBuff[2], 1 );
			AppendToMessage( (rom unsigned char *)pDevD, 0, 16 );
			AddMessageToQueue();
			return;
		case 0x02:  // dialog info
			if (RxBuffLen<4) { SetCmdErr;  return; }
			AppendToMessage( pResFxPInfo, 0, ResFxPInfoLen );
			AppendToMessage( 0, (unsigned char *)&RxBuff[2], 1 );
			AppendToMessage( (rom unsigned char *) &pDlgD[RxBuff[3]], 0, 4 );
			AddMessageToQueue();
			return;
		case 0x03:  // control info
			if (RxBuffLen<5) { SetCmdErr;  return; }
			{
				unsigned short CtlIdx = GetCtlIndex( &RxBuff[3] );
				if (IsCmdErr) return;
				AppendToMessage( pResFxPInfo, 0, ResFxPInfoLen );
				AppendToMessage( 0, (unsigned char *)&RxBuff[2], 1 );
				AppendToMessage( (rom unsigned char *)&pDlgD[ CtlIdx ], 0, 16 );
				AddMessageToQueue();
				RxBuff[2]++;
				AppendToMessage( pResFxPInfo, 0, ResFxPInfoLen );
				AppendToMessage( 0, (unsigned char *)&RxBuff[2], 1 );
				AppendToMessage( ((rom unsigned char *)&pDlgD[ CtlIdx ])+16, 0, 12 );
				AddMessageToQueue();
			}
			return;
		default:
			SetCmdErr;  
			return; 
	}
}

void GetCtlCmd( void )
{
	unsigned short iCtlOffset;
	unsigned short CtlIdx;

	// error if not 4 bytes
	if (RxBuffLen!=4) { SetCmdErr;  return; }

	// get control index
	CtlIdx = GetCtlIndex( &RxBuff[2] );
	if (IsCmdErr) return;

	// generate as many responses as required to send message
	for (iCtlOffset = 0; iCtlOffset < pCtlD[CtlIdx].VarLen; iCtlOffset+=16)
	{
		unsigned short nBytes = pCtlD[CtlIdx].VarLen - iCtlOffset;
		if (nBytes>16) nBytes = 16;

		AppendToMessage( pResCtlData, 0, ResCtlDataLen );
		AppendToMessage( 0, (unsigned char *)&RxBuff[2], 2 );  // ctl ID
		AppendToMessage( 0, (unsigned char *)&iCtlOffset, 2 );  // offset
		GetBytes( pCtlD[CtlIdx].VarStr, pCtlD[CtlIdx].VarLoc + iCtlOffset, &RxBuff[4], nBytes  );
		AppendToMessage( 0, &RxBuff[4], (unsigned char)nBytes );
		AddMessageToQueue();
	}
}

void SendIOValResponse( unsigned char iProperty, unsigned char iValue )
{
	AppendToMessage( pResIOVal, 0, ResIOValLen );
	AppendToMessage( 0, (unsigned char *)&iProperty, sizeof(iProperty) );
	AppendToMessage( 0, (unsigned char *)&iValue, sizeof(iValue) );
	AddMessageToQueue();
}

void SendADResponse( unsigned char iProperty )
{
	SetADChan( iProperty - 0x10 );  					// select a to d channel
	CyclesDelay3p2plus3p2times( 4 ); 		// 16us delay to charge S & H cap (refer to PIC18F6720 manual)
	StartAtoD;								// start a to d conversion
	AwaitAtoDComplete;						// await end of conversion

	// Note, high byte sent first since more intuitive for ASCII responses; FlexiPanel Ltd is usually little-endian
	AppendToMessage( pResIOVal, 0, ResIOValLen );
	AppendToMessage( 0, (unsigned char *)&iProperty, sizeof(iProperty) );
	AppendToMessage( 0, (unsigned char *)&ADRESH, 1 );
	AppendToMessage( 0, (unsigned char *)&ADRESL, 1 );
	AddMessageToQueue();
}

// code rather large for linker so break into small sections
#pragma code MemorySectionD

unsigned short GetCtlIndex( unsigned char * pCtlIDBigEndian )
{
	unsigned short Index;
	unsigned short CtlID = pCtlIDBigEndian[0] * 0x100 + pCtlIDBigEndian[1];
	for (Index = 0; Index < pDevD->nCtl; Index++)
	{
		if (CtlID == pCtlD[Index].ID) return Index;
	}

	// error
	SetCmdErr;  
	return 0;
}

void SetCtlPropsCmd( void )
{
	unsigned char temp;

	// error if not equal to 12 bytes
	if (RxBuffLen!=12) { SetCmdErr;  return; }

	// turn ctl id to little endian
	temp = RxBuff[2] ;
	RxBuff[2] = RxBuff[3] ;
	RxBuff[3] = temp ;

	// set properties
	FxPCommand( FxPC_SetProps , *((unsigned short*)(&RxBuff[2])), &RxBuff[4] );
	SendOKResponse();
}

void SetCtlCmd( void )
{
	unsigned short iDatLen;
	unsigned short CtlIdx;

	// bytes 2/3 = control ID
	// bytes 4/5 = offset

	// error if less than 6 bytes
	if (RxBuffLen<6) { SetCmdErr;  return; }

	// get control index
	CtlIdx = GetCtlIndex( &RxBuff[2] );
	if (IsCmdErr) return;

	// get safe data payload length
	iDatLen = pCtlD[CtlIdx].VarLen - *((unsigned short*)(&RxBuff[4]));  // length from offset to end of control...
	if (iDatLen > RxBuffLen - 6) iDatLen = RxBuffLen - 6;				// ...or number of data bytes in command

	// set data
#ifdef SBD
SetBytesDebug = 20;
#endif
	SetBytes( pCtlD[CtlIdx].VarStr, pCtlD[CtlIdx].VarLoc + *((unsigned short*)(&RxBuff[4])), 0, &RxBuff[6], (unsigned char)iDatLen  );
#ifdef SBD
SetBytesDebug = 21;
#endif

	// if offset is zero, update client
	if (FxPDeviceConnected && RxBuff[4]==0 && RxBuff[5]==0)
	{
		// turn ctl id to little endian
		unsigned char temp;
		temp = RxBuff[2] ;
		RxBuff[2] = RxBuff[3] ;
		RxBuff[3] = temp ;

		// send update message
		FxPCommand( FxPC_CtlUpdate , *((unsigned short*)(&RxBuff[2])), 0 );		
	}

	// acknowledge 
	SendOKResponse();
}

// code rather large for linker so break into small sections
#pragma code MemorySectionC

// set, append or log row
void SetRowCmd( unsigned char RowType )
{
	unsigned short iDatLen;
	unsigned short CtlIdx;
	rom bqFxPCtlData* pCtl;
	ADD FieldOffest; // dogsbody pointer
	unsigned char SourceOffset;  // address of next byte of data in command
	unsigned char temp;

	// bytes 2/3 = control ID
	// bytes 4/5 row number if non-appending row update
	// remainder = row data (x axis first if XY or datetime)
	// if no remainder, just update control on client

	// error if less than 6 bytes
	if (RxBuffLen<4) { SetCmdErr;  return; }

	// get control index
	CtlIdx = GetCtlIndex( &RxBuff[2] );
	if (IsCmdErr) return;

	// turn ctl id to little endian
	temp = RxBuff[2] ;
	RxBuff[2] = RxBuff[3] ;
	RxBuff[3] = temp ;

	// update client
	if (RxBuffLen==4)
	{
		// update client
		if (FxPDeviceConnected)
		{
			// send update message
			FxPCommand( FxPC_CtlUpdate , *((unsigned short*)(&RxBuff[2])), 0 );		
		}
		// acknowledge 
		SendOKResponse();
		return;
	}

	// get matrix info
	pCtl = &pCtlD[CtlIdx];
	GetMatrixStats( pCtl );

	// calculate row to update
	SourceOffset = 4;
	if (RxBuff[1]==0x0B) // regular set row command
	{
		// error if less than 6 bytes & set row row
		if (RxBuffLen<6) { SetCmdErr;  return; }

		// calculate row
		RowToUpdate = *((unsigned short*)(&RxBuff[SourceOffset]));
		if (RowToUpdate>=NumRows) { SetCmdErr;  return; }
		SourceOffset = 6;
	}

	// update X axis, if there is one
	FieldOffest = NumCols * NumRows * Cellsize + RowToUpdate * XValSize;
	if (RxBuff[1]==0x0D) // log row command
	{
		// error if not date-time matrix
		if (XValSize!=sizeof(DateTimeU)) { SetCmdErr;  return; }
		
		// error if not equal to (SourceOffset+RowSize-sizeof(DateTimeU)) bytes - note this limits matrices to 16 bytes of updateable data per row / 18 bytes appendable
		if (RxBuffLen!=SourceOffset+RowSize-sizeof(DateTimeU) ) { SetCmdErr;  return; }

		// set to current time
#ifdef SBD
SetBytesDebug = 22;
#endif
		SetBytes( pCtl->VarStr, pCtl->VarLoc + FieldOffest, 0, (unsigned char *) &RealTimeClock, XValSize  );
#ifdef SBD
SetBytesDebug = 23;
#endif
	}
	else
	{   // set row / append row command
		// error if not equal to (SourceOffset+RowSize) bytes - note this limits matrices to 16 bytes of updateable data per row / 18 bytes appendable
		if (RxBuffLen!=SourceOffset+RowSize) { SetCmdErr;  return; }

		// update x axis
		if (XValSize) 
		{
#ifdef SBD
SetBytesDebug = 24;
#endif
			SetBytes( pCtl->VarStr, pCtl->VarLoc + FieldOffest, 0, &RxBuff[SourceOffset], XValSize  );
#ifdef SBD
SetBytesDebug = 25;
#endif
			SourceOffset += XValSize;
		}
	}

	// set Y data
	FieldOffest = RowToUpdate * Cellsize;
	while (SourceOffset<RxBuffLen)
	{
#ifdef SBD
SetBytesDebug = 26;
#endif
		SetBytes( pCtl->VarStr, pCtl->VarLoc + FieldOffest, 0, &RxBuff[SourceOffset], Cellsize  );
#ifdef SBD
SetBytesDebug = 27;
#endif
		FieldOffest += NumRows * Cellsize;
		SourceOffset += Cellsize;
	}

	// update valid row counter
	if (RxBuff[1]==0x0B) // regular set row command ( if append, NextRowCounter will have been calculated by GetMatrixStats
	{
		// if regular set row, just make sure this row is visible
		if (RowCounter <= RowToUpdate) 
			NextRowCounter = RowToUpdate +1;
		else
			NextRowCounter = RowCounter;
	}
#ifdef SBD
SetBytesDebug = 28;
#endif
	SetBytes( pCtl->VarStr, RowCountAddress, 0, &NextRowCounter, sizeof(NextRowCounter)  );
#ifdef SBD
SetBytesDebug = 29;
#endif
	// update client if not set row cdommand
	if (FxPDeviceConnected && RxBuff[1]!=0x0B)
	{
		// send update message
		FxPCommand( FxPC_PartUpdate, pCtl->ID, (void *) RowToUpdate );
	}

	// acknowledge 
	SendOKResponse();
}

void GetMemCmd( void )
{
	// bytes 2 = storage type
	// bytes 3/4 = address
	// byte 5 = number of bytes required, up to 17 bytes
	unsigned char numBytes = RxBuff[5];

	// error if not 6 bytes - no other error checking done
	if (RxBuffLen!=6) { SetCmdErr;  return; }
	
	// error if more than 17 bytes requested
	if (numBytes>17) { SetCmdErr;  return; }

	// get data 
	GetBytes( RxBuff[2], *((ADD*)(&RxBuff[3])), &RxBuff[5], numBytes );

	// transmit result
	AppendToMessage( pResGotBytes, 0, ResGotBytesLen );
	AppendToMessage( 0, &RxBuff[2], 3 );
	AppendToMessage( 0, &RxBuff[5], numBytes );
	AddMessageToQueue();
}

void SetMemCmd( void )
{
	// bytes 2 = storage type
	// bytes 3/4 = address
	// remainder = data to set, up to 17 bytes

	// error if less than 5 bytes - no other error checking done
	if (RxBuffLen<5) { SetCmdErr;  return; }

	// do not permit writing to Flash 0x80 or (0x81 locs > 0xDFFF)
	if (RxBuff[2] == 0x01) RxBuff[2] = 0x81;
	if (RxBuff[2] == 0x80 || (RxBuff[2] == 0x81 && RxBuff[4] >= 0xE0)  ) { SetCmdErr;  return; }

	// set data 
#ifdef SBD
SetBytesDebug = 30;
#endif
	SetBytes( RxBuff[2], *((ADD*)(&RxBuff[3])), 0, &RxBuff[5], RxBuff[0] - 5 );
#ifdef SBD
SetBytesDebug = 31;
#endif

	// acknowledge 
	SendOKResponse();
}


// GetMatrixStats calculates matrix vital statistics on the fly
// results held in static variables since inefficient to pass so much as arguments
void GetMatrixStats( rom bqFxPCtlData* pCtl )
{
	// check it's a matrix
	if (pCtl->Type!=TP_CtlTypeMatrix) { SetCmdErr; return; }

	// calculate matrix cell dimensions
	Cellsize = 1;
	if ( (pCtl->dwflags & CTL_MTX_Y_BYTE_MASK) == CTL_MTX_Y_2BYTE ) Cellsize = 2;
	if ( (pCtl->dwflags & CTL_MTX_Y_BYTE_MASK) == CTL_MTX_Y_4BYTE ) Cellsize = 4;
	NumCols = pCtl->nCtlVal;

	// calculate matrix x-axis value dimensions
	XValSize = 0;
	if ( (pCtl->dwflags & CTL_MTX_DATA_MASK) == CTL_MTX_DATA_TY )   XValSize = sizeof ( DateTimeU );
	if ( (pCtl->dwflags & CTL_MTX_DATA_MASK) == CTL_MTX_DATA_XY )   
	{
		XValSize = 1;
		if ( (pCtl->dwflags & CTL_MTX_X_BYTE_MASK) == CTL_MTX_X_2BYTE ) XValSize = 2;
		if ( (pCtl->dwflags & CTL_MTX_X_BYTE_MASK) == CTL_MTX_X_4BYTE ) XValSize = 4;
	}

	// get current value of RowCounter - always last 4 bytes of matrix data
	RowCountAddress = pCtl->VarLoc + pCtl->VarLen - 4;
	GetBytes( pCtl->VarStr, RowCountAddress, &RowCounter, 4  );

	// Row size in bytes is CellSize * NumCols plus x axis data
	RowSize = Cellsize * NumCols + XValSize;
	NumRows = (pCtl->VarLen - 4) / RowSize;
	BytesInXCol = XValSize * NumRows;
	BytesInCol = Cellsize * NumRows;

	// calculate which row will be updated if appending
	RowToUpdate = ( (RowCounter>=0) ? ( (RowCounter< NumRows) ? RowCounter : 0 ) : ( (RowCounter> -NumRows) ? - RowCounter : 0 ) );

	// calculate next row counter value if appending
	NextRowCounter = ( (RowCounter>=0) ? ( (RowCounter< NumRows) ? RowCounter+1 : -1 ) : ( (RowCounter> -NumRows) ? RowCounter - 1 : -1 ) );
}
